# -------------------------------
# IMPORT PYTHON LIBRARIES
# -------------------------------
import os
import sys
import numpy as np
import pandas as pd
import pylab as pl
from mpl_toolkits.basemap import Basemap
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.figure import Figure
from gtk.gdk import screen_get_default
import pdb

# ----------------------------------------------
# IMPORT DEVELOPED LIBRARIES
# ----------------------------------------------
from ldasModule import getModelGrid, res2lonlat

# -------------------------------
# DEFAULT PARAMETERS
# -------------------------------
pl.ioff()                  # Turn interactive mode off
def_NUM_TICKS = 5          # Number of colorbar ticks
def_TO_DRAW = 'countries'  # Additional drawing
def_PTS = 72.27            # Points in 1 inch
def_MMS = 25.4             # mm in 1 inch
def_L_FS = 12/def_PTS      # Axis label font size (inches)
def_T_FS = 14.4/def_PTS    # Title label font size (inches)
def_PAD = pl.rcParams['xtick.major.pad']/def_PTS   # Axis label pad (inches)
def_IN_PAD = def_L_FS/2    # Padding between sub-plots
def_SCR_WH = pl.array([screen_get_default().get_width_mm(), screen_get_default().get_height_mm()])/def_MMS   # Screen width and height (inches)

# -------------------------------
# MODULE FUNCTIONS
# -------------------------------
def uniformTicks(data, num_ticks=def_NUM_TICKS, zero_centred=False, minim=0., maxim=0.):
    '''
    Compute the interval of values for scaling the colors of the maps.
    Arguments:
        data (pandas...): values to be plotted.
        num_ticks (int): approximate number of colorbar ticks.
    Returns:
        ticks (numpy.ndarray): colorbar ticks.
    ** For improvement: take Emiliano's code to minimize the number of ticks with decimal point **
    '''
    if minim is None: minim = 0.
    if maxim is None: maxim = 0.
    if minim==0. and maxim==0.:
        if zero_centred:
            minim = np.nanmean(data.values)-2*np.nanstd(data.values)
        else:
            minim = np.nanmax([0.,np.nanmean(data.values)-2*np.nanstd(data.values)])
        maxim = np.nanmean(data.values)+2*np.nanstd(data.values)
    if np.logical_and(zero_centred, np.logical_and(minim<=0.,maxim>=0.)):
        minim = min(minim, -1*maxim)
        maxim = max(maxim, -1*minim)
    ticks = np.linspace(minim, maxim, num_ticks)
    return ticks
#
def discreteColorscale(data, color_map, color_ticks, num_ticks, nb_colors, zero_centred=False, minim=0., maxim=0.):
    '''
    '''
    if minim is None: minim = 0.
    if maxim is None: maxim = 0.
    if not color_ticks: color_ticks = uniformTicks(data, num_ticks, zero_centred, minim, maxim)
    if not nb_colors: nb_colors = numberOfColors(color_ticks.size)
    color_scale = pl.cm.get_cmap(color_map,nb_colors)
    if 0 in color_ticks[1:-1] and zero_centred: color_scale = zeroCentredColorMap(color_scale, color_ticks, nb_colors)
    return color_ticks, color_scale
#
def zeroCentredColorMap(color_scale, color_ticks, nb_colors):
    '''
    Function to offset the "center" of a colormap. Useful for data with a negative min and positive max
    and you want the middle of the colormap's dynamic range to be at zero
    '''
    num_colors = nb_colors #numberOfColors(color_ticks.size)
    cdict = {'red': [], 'green': [], 'blue': [], 'alpha': []}
    reg_index = pl.linspace(0,1,num_colors) # regular index to compute the colors
    zero_ix = pl.find(color_ticks == 0)/float(color_ticks.size-1)
    shift_index = pl.hstack([pl.linspace(0,zero_ix,num_colors/2,False), pl.linspace(zero_ix,1,num_colors-num_colors/2)])
    for ri, si in zip(reg_index, shift_index):
        r, g, b, a = color_scale(ri)
        cdict['red'].append((si, r, r))
        cdict['green'].append((si, g, g))
        cdict['blue'].append((si, b, b))
        cdict['alpha'].append((si, a, a))
    new_cmap = LinearSegmentedColormap('',cdict,num_colors)
    return new_cmap
#
def numberOfColors(num_ticks):
    '''Compute the number of colors based on the number of colorbar ticks.'''
    return 4*(num_ticks-1)
#



# -------------------------------
# CLASS
# -------------------------------
class mapSet:
    '''
    Class mapSet for plotting several maps with the same domain and coordinates.
    '''
    def __init__(self, mod_grid, pgd=None, map_resolution='l'):
        '''
        Constructor method of the mapSet class. It can only be initialized with mod_grid defined with the following format :
        mod_grid = {'type': 'uniform', 'res': 0.5, 'zone': [[-11.5,75.5],[62.5,25.]], 'proj': }
           # Model grid type ([[ul_lon,ul_lat],[lr_lon,lr_lat]])
        '''
        if 'proj' in mod_grid.keys() and mod_grid['proj']: map_proj = mod_grid['proj']
        else: map_proj = 'cyl'
        if map_proj=='lcc_France_SAFRAN':
            self.projection = Basemap(projection='lcc',    llcrnrlat=41.3,urcrnrlat=50.9,llcrnrlon=-4.1,urcrnrlon=11.0, \
                                                           lat_0=46.8, lon_0=2+(20+14.025/60)/60, \
                                                           lat_1=45+(53+56.108/60)/60, lat_2=47+(41+45.652/60)/60, \
                                                           resolution=map_resolution)
        else:
            self.projection = Basemap(projection=map_proj, llcrnrlat=mod_grid['zone'][1][1], urcrnrlat=mod_grid['zone'][0][1], \
                                                           llcrnrlon=mod_grid['zone'][0][0], urcrnrlon=mod_grid['zone'][1][0], \
                                                           resolution=map_resolution)
        self.mod_grid_type = mod_grid['type']
        self.mod_lon,self.mod_lat = res2lonlat(mod_grid['res'],mod_grid['zone'])
        self.mod_lon,self.mod_lat = pl.meshgrid(self.mod_lon,self.mod_lat)
        self.data_shape = ((len(self.mod_lat)*len(self.mod_lon),))
        #self.row_ix = pl.arange(0,len(self.mod_lat))
        #self.col_ix = pl.arange(0,len(self.mod_lon))
        if mod_grid['type']!='uniform':
            mod_lon, mod_lat = getModelGrid(mod_grid,pgd)
            self.remap = pl.int_((mod_lon-self.mod_lon[0,0])/mod_grid['res'])+\
                         pl.int_((mod_lat-self.mod_lat[0,0])/mod_grid['res'])*self.mod_lon.shape[1]
        self.X, self.Y = self.mod_lon-mod_grid['res']/2.,self.mod_lat-mod_grid['res']/2.
        self.zone = None
        self.fGUI = False
        self.fig = None
    # 
    # PROPERTIES
    @property
    def map_width(self):
        '''Map width (m)'''
        return self.projection.urcrnrx - self.projection.llcrnrx
    #
    @property
    def map_height(self):
        '''Map height (m)'''
        return self.projection.urcrnry - self.projection.llcrnry
    #
    @property
    def land_mask(self):
        '''Model mesh land mask (True = land)'''
        mask = pl.ones(self.data_shape, bool)
        return mask
    #
    # INTERNE FUNCTIONS
    def figureLayout(self, title, n_rows=1, n_cols=1): # use pl.rcParams['font.size'] to add space for title and axis ticks
        '''
        '''
        bar_rel_h = .05+.01*(n_rows-1)
        title_h   = 4*def_PAD+def_T_FS if title else 0
        dims_fig  = pl.figaspect((n_rows+bar_rel_h)*self.map_height/(n_cols*self.map_width))+pl.array([0,title_h+def_IN_PAD]) # width, height (in)
        if n_rows == n_cols == 1:
            dims_fig += def_PAD +def_IN_PAD +def_L_FS*pl.array([6,2])
        else:
            if n_cols > 1:
                dims_fig += pl.array([(n_cols-1)*def_IN_PAD, def_PAD+2*def_L_FS])
            if n_rows > 1:
                dims_fig += pl.array([def_PAD+2*def_L_FS, (n_rows-1)*def_IN_PAD])
        grid_spec = pl.GridSpec(n_rows+1,n_cols,height_ratios=n_rows*[1]+[bar_rel_h],width_ratios=n_cols*[1]) # hspace,wspace????
        # check and re-scale figure if its size exceeds any of the screen dimensions
        if any(dims_fig > def_SCR_WH): dims_fig /= (dims_fig/def_SCR_WH).max()
        if self.fGUI: self.fig = Figure(figsize=dims_fig)
        else        : self.fig = pl.figure(figsize=dims_fig)
        return grid_spec, self.fig, title_h/dims_fig[1], bar_rel_h*dims_fig[1]/((n_rows+bar_rel_h)*dims_fig[0])
    #
    def drawMap(self,data_values,color_ticks,color_scale,ax,to_draw,line_widths):
        '''
        '''
        data_matrix = pl.empty(self.X.shape)*pl.ma.masked
        #data_matrix[self.row_ix,self.col_ix] = data_values
        if self.mod_grid_type=='uniform':
            data_matrix[:,:] = data_values.reshape(self.X.shape)
        else:
            tmp = data_matrix.ravel()
            tmp[self.remap] = data_values
            data_matrix[:,:] = tmp.reshape(self.X.shape)
        data_matrix.mask[pl.isnan(data_matrix)] = True
        self.projection.drawcoastlines(linewidth=line_widths[0],ax=ax)
        if 'countries' in to_draw: self.projection.drawcountries(linewidth=line_widths[1],ax=ax)
        if 'rivers'    in to_draw: self.projection.drawrivers(linewidth=line_widths[1],ax=ax)
        hp = self.projection.pcolormesh(self.X,self.Y,data_matrix,vmin=color_ticks[0],vmax=color_ticks[-1],cmap=color_scale,ax=ax,latlon=True)
        if self.zone is not None: ax.axis([self.zone[0][0], self.zone[1][0], self.zone[1][1], self.zone[0][1]])
        return hp
    #
    def matrixMaps(self, data, color_map='jet', title=None, color_ticks=None, row_order=None, col_order=None, \
                   row_labels=None, col_labels=None, to_draw=def_TO_DRAW, line_widths=[.2,.1], num_ticks=def_NUM_TICKS, \
                   zero_centred=False, obs_mod_same_unit=True, minim=0., maxim=0., minim_diff=0., maxim_diff=0.):
        '''
        Draw array of raster maps.
        Arguments:
            self: instance of the mapSet class.
            data (pandas.Panel): maps to to be drawn.
            color_map (str): colormap. Default: 'jet'.
            color_ticks (numpy.ndarray): colorbar ticks. Default: None.
            row_order/col_order (list/numpy.ndarray): order of rows/columns for the multi-map figure (int).
            row_labels/col_labels (list): list of row/column labels for multi-map figure (str); Default: None.
            to_draw (str): additional map elements to be drawn among 'countries' or 'rivers'.
            line_widths (list): line widths (float) of coastlines and other line elements ('countries' or 'rivers').
            num_ticks (int): approximate number of colorbar ticks.
            minim,maxim (float): minimum and maximum of colorbar.
        '''
        # Check input data
        if not isinstance(data,pd.core.panel.Panel): raise Exception('The input data must be an instance of pandas.core.panel.Panel')
        # Get data dimensions
        n_cols, n_rows = data.shape[:2]
        #
        if not col_labels: col_labels = data.items
        if len(col_labels) != data.shape[0]: raise Exception('Column labels do not match data size')
        #
        if not row_labels: row_labels = data.major_axis
        if len(row_labels) != data.shape[1]: raise Exception('Row labels do not match data size')
        #
        if not col_order: col_order = range(n_cols)
        if len(col_order) != data.shape[0]: raise Exception('Column order does not match data size')
        col_labels = [col_labels[i] for i in col_order]
        #
        if not row_order: row_order = range(n_rows)
        if len(row_order) != data.shape[1]: raise Exception('Row order does not match data size')
        row_labels = [row_labels[i] for i in row_order]
        # Figure layout
        grid_spec, fig, title_rel_h, cbar_aspect = self.figureLayout(title,n_rows,n_cols)
        #
        # If the last column is a difference, then it should be handled differently
        bool_lastcol_diff = True if '-' in col_labels[-1] else False
        #
        # Fill removed points (with nan values)
        data = data.reindex(minor=pl.arange(self.X.size),copy=False)
        #
        # Draw maps of variables
        if obs_mod_same_unit:
            # Discrete colorscale for variables
            color_ticks, color_scale = discreteColorscale(data, color_map, color_ticks, num_ticks, nb_colors=None, zero_centred=False, minim=minim, maxim=maxim)
            for c in range(n_cols-bool_lastcol_diff):
                item = data.items[col_order[c]]
                for r in range(n_rows):
                    ax = fig.add_subplot(grid_spec[r,c])
                    pc = self.drawMap(data.ix[item,r].values,color_ticks,color_scale,ax,to_draw,line_widths)
                    if r == 0:
                        ax.set_xlabel(col_labels[c],fontsize=def_L_FS*def_PTS)
                        ax.xaxis.set_label_position('top')
                    if c == 0:
                        ax.set_ylabel(row_labels[r],fontsize=def_L_FS*def_PTS)
                # Draw colorbar of variables
                if bool_lastcol_diff: cax = fig.add_subplot(grid_spec[n_rows,0:-1], aspect=cbar_aspect)
                else:                 cax = fig.add_subplot(grid_spec[n_rows,:],    aspect=cbar_aspect)
                clb_var = fig.colorbar(pc, cax=cax, orientation='horizontal', ticks=color_ticks, format='%.1g')
                clb_var.ax.tick_params(labelsize=10) 
        else:
            # If the observations do not have the same unit than the model simulations, then they should have a different colorbar
            color_ticks_mod, color_scale_mod = discreteColorscale(data.loc[['Model'],:,:], color_map, color_ticks, num_ticks, None, False, 0.   , 0.   )
            color_ticks_obs, color_scale_obs = discreteColorscale(data.loc[['Obs'],:,:]  , color_map, color_ticks, num_ticks, None, False, minim, maxim)
            for c in range(n_cols-bool_lastcol_diff):
                item = data.items[col_order[c]]
                if item=='Obs':                       color_ticks,color_scale=color_ticks_obs, color_scale_obs
                if item=='Model' or item=='Analysis': color_ticks,color_scale=color_ticks_mod, color_scale_mod
                #print('item={0}, color_ticks={1}'.format(item,color_ticks))
                for r in range(n_rows):
                    ax = fig.add_subplot(grid_spec[r,c])
                    pc = self.drawMap(data.ix[item,r].values,color_ticks,color_scale,ax,to_draw,line_widths)
                    if r == 0:
                        ax.set_xlabel(col_labels[c],fontsize=def_L_FS*def_PTS)
                        ax.xaxis.set_label_position('top')
                    if c == 0:
                        ax.set_ylabel(row_labels[r],fontsize=def_L_FS*def_PTS)
                # Draw colorbar of the whole column
                cax = fig.add_subplot(grid_spec[n_rows,c],aspect=cbar_aspect*3.45)
                clb_var = fig.colorbar(pc, cax = cax, orientation='horizontal', ticks=color_ticks, format='%.2g')
                clb_var.ax.tick_params(labelsize=10) 
        # Draw map of difference
        if bool_lastcol_diff:
            if minim_diff==0. and maxim_diff==0.: 
                minim_diff = np.nanmean(data[data.items[-1]].values)-2*np.nanstd(data[data.items[-1]].values)
                maxim_diff = np.nanmean(data[data.items[-1]].values)+2*np.nanstd(data[data.items[-1]].values)
            if minim_diff != -1.*maxim_diff: minim_diff,maxim_diff = min(minim_diff,-1*maxim_diff),max(maxim_diff,-1*minim_diff)
            color_ticks_diff, color_scale_diff = discreteColorscale(data[data.items[-1]], color_map='bwr', color_ticks=None, num_ticks=3, nb_colors=None, zero_centred=True, minim=minim_diff, maxim=maxim_diff)
            c = n_cols-bool_lastcol_diff
            item = data.items[c]
            for r in range(n_rows):
                ax = fig.add_subplot(grid_spec[r,c])
                pc = self.drawMap(data.ix[item,r].values, color_ticks_diff, color_scale_diff, ax,to_draw,line_widths)
                if r == 0:
                    ax.set_xlabel(col_labels[c],fontsize=def_L_FS*def_PTS)
                    ax.xaxis.set_label_position('top')
                if c == 0:
                    ax.set_ylabel(row_labels[r],fontsize=def_L_FS*def_PTS)
            cax = fig.add_subplot(grid_spec[n_rows,-1],aspect=cbar_aspect*3.45)
            clb_diff = fig.colorbar(pc, cax=cax, orientation='horizontal', ticks=color_ticks_diff, format='%.2g')
            clb_diff.ax.tick_params(labelsize=10)
        # Tight layout
        fig.tight_layout(pad=0,h_pad=def_IN_PAD/def_L_FS,w_pad=def_IN_PAD/def_L_FS,rect=[0,0,1,1-title_rel_h])
        # Title
        if title: fig.suptitle(title,fontsize=def_T_FS*def_PTS,y=1)
        #
        return fig
#
    def matrixIncrMaps(self, incr, color_map='jet', title=None, color_ticks=None, row_order=None, col_order=None,\
                       row_labels=None, col_labels=None, to_draw=def_TO_DRAW, line_widths=[.2,.1], num_ticks=def_NUM_TICKS, \
                       zero_centred=True, minim=0., maxim=0.):
        '''
        Draw array of increment maps.
        Arguments:
            self: instance of the mapSet class.
            data (pandas.Panel): maps to to be drawn.
            color_map (str): colormap. Default: 'jet'.             
            color_ticks (numpy.ndarray): colorbar ticks. Default: None.
            row_order/col_order (list/numpy.ndarray): order of rows/columns for the multi-map figure (int).
            row_labels/col_labels (list): list of row/column labels for multi-map figure (str); Default: None.
            to_draw (str): additional map elements to be drawn among 'countries' or 'rivers'.
            line_widths (list): line widths (float) of coastlines and other line elements ('countries' or 'rivers').
            num_ticks (int): approximate number of colorbar ticks.
            minim,maxim (float): minimum and maximum of colorbar.
        '''
        # Check input data
        if not isinstance(incr,pd.core.panel.Panel): raise Exception('The input data must be an instance of pandas.core.panel.Panel')
        # Get data dimensions
        n_cols, n_rows = incr.shape[:2]
        #
        if not col_labels: col_labels = incr.items
        if len(col_labels) != incr.shape[0]: raise Exception('Column labels do not match data size')
        #
        if not row_labels: row_labels = incr.major_axis
        if len(row_labels) != incr.shape[1]: raise Exception('Row labels do not match data size')
        #
        if not col_order: col_order = range(n_cols)
        if len(col_order) != incr.shape[0]: raise Exception('Column order does not match data size')
        col_labels = [col_labels[i] for i in col_order]
        #
        if not row_order: row_order = range(n_rows)
        if len(row_order) != incr.shape[1]: raise Exception('Row order does not match data size')
        row_labels = [row_labels[i] for i in row_order]
        #
        # Figure layout
        grid_spec, fig, title_rel_h, cbar_aspect = self.figureLayout(title,n_rows,n_cols)
        #
        # Fill removed points (with nan values)
        incr = incr.reindex(minor=pl.arange(self.X.size),copy=False)
        #
        # If minim/maxim are scalars, then repeat
        if not isinstance(minim, pl.ndarray): minim = pl.ones((n_cols,))*minim
        if not isinstance(maxim, pl.ndarray): maxim = pl.ones((n_cols,))*maxim
        if len(minim)!=n_cols or len(maxim)!=n_cols: raise Exception('The number of minim/maxim given is not equal to the number of columns')
        # Draw maps of variables
        for c in range(n_cols):
            item = incr.items[col_order[c]]
            # Discrete colorscales for each increment type
            color_ticks2 = color_ticks
            color_ticks2, color_scale = discreteColorscale(pd.Panel(incr.values[None,col_order[c],:,:], [incr.items[col_order[c]]], incr.major_axis, incr.minor_axis),\
                                                           color_map, color_ticks2, 3, 21, zero_centred, minim[c], maxim[c])
            # Draw single maps
            for r in range(n_rows):
                ax = pl.subplot(grid_spec[r,c])
                pc = self.drawMap(incr.ix[item,r].values,color_ticks2,color_scale,ax,to_draw,line_widths)
                if r == 0:
                    ax.set_xlabel(col_labels[c],fontsize=def_L_FS*def_PTS)
                    ax.xaxis.set_label_position('top')
                if c == 0:
                    ax.set_ylabel(row_labels[r],fontsize=def_L_FS*def_PTS)
            # Draw colorbar of variables
            ax = pl.subplot(grid_spec[-1,c])
            clb_var = pl.colorbar(pc, cax=ax, orientation='horizontal', ticks=color_ticks2, format='%.2g')
            clb_var.ax.tick_params(labelsize=10)
        # Tight layout
        pl.tight_layout(pad=0,h_pad=def_IN_PAD/def_L_FS,w_pad=def_IN_PAD/def_L_FS,rect=[0,0,1,1-title_rel_h])
        # Title
        if title: fig.suptitle(title,fontsize=def_T_FS*def_PTS,y=1)
        #
        return fig

















