# ----------------------------------------------
# IMPORT LIBRARIES
# ----------------------------------------------
# Python modules
import pdb
from warnings import filterwarnings; filterwarnings('ignore')
import os
import pandas as pd
import pylab as pl
import pickletools as pkt
from gzip import open as gzOpen
from h5py import File as h5File
from netCDF4 import Dataset, num2date
from sys import stdout, exit
from inspect import currentframe, getargvalues
from tqdm import tqdm, trange
import f90nml
# Unused Python modules (should be fully checked)
#import datetime
#from imp import load_dynamic
#from numpy import load, save
#from glob import glob
#from re import match



# ----------------------------------------------
# SET CONSTANTS
# ----------------------------------------------
# Names of option variables that have to be set
OPT_ASSIM_NAMES = ('analysis_period', 'assim_hour')
OPT_MOD_NAMES = ()
OPT_FORC_NAMES = ('get_raw_forcing', 'freq_forcing', 'nc_version', 'raw_forcing_dir', 'raw_forcing_format', 'FORC_VARS')
OPT_OBS_NAMES = ('obs_names', 'obs_datasets', 'to_assim', 'obs_from_pickle', 'obs_grids',\
    'grid_transform_methods', 'grid_transform_parameters',\
    'surf_screen_features', 'surf_screen_thresholds', 'screen_mod_names', 'screen_mod_thres', 'rescale_methods',\
    'rescale_mod_names', 'rescale_mod_steps', 'rescale_max_miss', 'rescale_calibs', 'rescale_calib_periods')
OPT_POSTPROC_NAMES = ('post_mod_from_pickle', 'post_ana_from_pickle', 'post_obs_from_pickle', 'post_zones', \
    'list_time_windows', 'var_time_series_out', 'var_spatial_maps_out', 'stats_time_series_out', 'stats_spatial_maps_out', 'stats_table_out',\
    'modexp_names', 'patch_out')
OPT_ALL_NAMES = OPT_ASSIM_NAMES + OPT_MOD_NAMES + OPT_FORC_NAMES + OPT_OBS_NAMES + OPT_POSTPROC_NAMES + \
    ('exp_name', 'options_path', 'mod_pgd_path',\
     'obs_root_dir', 'openloop_dir', 'analysis_dir', 'forcing_dir', \
     'out_obs_dir', 'out_mod_dir', 'out_ana_dir', 'graphics_dir',\
     'fig_dpi')

# Fill value for missing observations (for CANARI files & F90 subroutines)
FILL = 999

# Long names of patches
PATCH_NAMES = ['no_veg','bare_rock','snow','broadleaves','coniferes','evergreen','c3','c4','irrigated','grassland','tropical','wetland']


# --------------------------------------------------------------
# PUBLIC FUNCTIONS: LDAS OPTIONS FILE
# --------------------------------------------------------------

def getOptionsFile(shell_args):
    '''
    Gets the options file name from the arguments passed to the Python interpreter.
    Argument:
        shell_args (list): shell arguments (str).
    '''
    if not shell_args[1] or shell_args[1][-3:] != '.py':
        raise Exception('Options file missing!\n\
        The path to the options file should be passed as argument when executing the program, i.e.:\n\
        python ldasPre.py <options_file_path>')
    return shell_args[1]

def checkOptions(main_var_dict,options_file=None):
    '''
    Checks that all necessary options are defined and are consistent.
    Arguments:
        main_var_dict (dict): dict of variables names (str) defined in the main program.
        options_file (str): options file name.
    '''
    # check that all lists specifying observation options have the same length
    obs_opt_len = pl.array([len(main_var_dict[opt]) for opt in OPT_OBS_NAMES])
    has_bad_len = pl.find(obs_opt_len != obs_opt_len[OPT_OBS_NAMES.index('obs_names')])
    if has_bad_len.size:
        raise Exception('The following observation options do not have the same length as \'obs_names\'):\n\n' +\
        ', '.join([OPT_OBS_NAMES[i] for i in has_bad_len]) + '\n')
    # check that all necessary options are defined
    opt_chk = pl.array([n in main_var_dict.keys() for n in OPT_ALL_NAMES])
    if options_file is not None:
        if opt_chk.all():
            print('\nAll necessary options read from {0}.'.format(options_file))
        else:
            raise Exception('The following options are missing in {0}:\n {1}'.format(options_file,[OPT_ALL_NAMES[i] for i in pl.find(~opt_chk)]))

def parseOptionsNam(options_path):
    '''
    Parse model informations from OPTIONS.nam and TRIP_OPTIONS.nam
    Arguments:
        options_path (str): path to directory where OPTIONS.nam and TRIP_OPTIONS.nam are stored.
    Returns:
        parsed information (dict).
    '''
    if not os.path.isfile(options_path+'OPTIONS.nam'):
        raise Exception('OPTIONS.nam was not found in {0}'.format(options_path))
    nml = f90nml.read(options_path+'OPTIONS.nam')
    if nml['nam_pgd_grid']['cgrid'] == 'LONLAT REG':
        mod_grid = dict()
        mod_grid['type'] = 'uniform'
        nam_grid = nml['nam_lonlat_reg']
        mod_grid['res']  = float(nam_grid['xlonmax']-nam_grid['xlonmin'])/nam_grid['nlon']
        if mod_grid['res']!=float(nam_grid['xlatmax']-nam_grid['xlatmin'])/nam_grid['nlat']:
            raise Exception('Latitudinal resolution must match longitudinal resolution.')
        mod_grid['zone'] = [[nam_grid['xlonmin'],nam_grid['xlatmax']],[nam_grid['xlonmax'],nam_grid['xlatmin']]]
    elif nml['nam_pgd_grid']['cgrid'] == 'LONLATVAL':
        try:
            if os.path.isfile(options_path+'mapping.nc'):
                nc = Dataset(options_path+'mapping.nc', 'r')
                mod_grid = {'type':'points', 'res':nc.DX, 'zone':[[nc.X_beg,nc.Y_end],[nc.X_end,nc.Y_beg]], 'G2P':nc['G2P'][:]} #  zone = [[ul_lon,ul_lat],[lr_lon,lr_lat]]
                nc.close()
            else:
                mod_grid = currentframe(1).f_locals['mod_grid']
                mod_grid['type'],mod_grid['res'],mod_grid['zone']
        except:
            raise Exception('When model grid is defined as "LONLATVAL", the options file should define the mod_grid dictionary with keys "type", "res" and "zone"')
    else:
        raise Exception('Grid definition in OPTIONS.nam should be "LONLAT REG" or "LONLATVL"')
    if 'nam_oasis' in nml.keys() and 'loasis' in nml['nam_oasis'].keys() and nml['nam_oasis']['loasis'] is True:
        nml_trip = f90nml.read(options_path+'TRIP_OPTIONS.nam')
        trip_grid = {'type': 'uniform', 'res': nml_trip['nam_trip_grid']['tres'],
                     'zone': [[nml_trip['nam_trip_grid']['tlonmin'],nml_trip['nam_trip_grid']['tlatmax']],
                              [nml_trip['nam_trip_grid']['tlonmax'],nml_trip['nam_trip_grid']['tlatmin']]]}
    else: trip_grid = dict()
    if nml['nam_io_offline']['ctimeseries_filetype'] != 'NETCDF':
        raise Exception('SURFEX outputs must be stored in netCDF format.')
    return mod_grid,trip_grid

def parseGridTransformInfo(obs_names,obs_grids,methods,parameters,obs_root_dir,obs_datasets,mod_grid):
    '''
    Parse grid transformation information for all observation variables to be assimilated.
    Arguments:
        obs_names (list): names of observation variables (str).
        obs_grids (list): types and characteristics of observation grids (dict)
        methods (list): names of methods to project each observation variable onto model grid (str).
        parameters (list): parameters of methods (dict).
        obs_root_dir (str): root directory of observations
        obs_datasets (str): observation dataset
        mod_grid (dict): types and characteristics of model grid
    Returns:
        parsed information (dict).
    '''
    # check model grid zone
    if 'zone' in mod_grid.keys() and \
      (mod_grid['zone'][1][0] < mod_grid['zone'][0][0] or mod_grid['zone'][0][1] < mod_grid['zone'][1][1]):
        raise Exception('Model grid zone shoud be defined as [[ul_lon,ul_lat],[lr_lon,lr_lat]]')
    # obs grid info for up/down-scaling
    regions = []
    for n in range(len(obs_names)):
        regions.append(getObsRegions(obs_root_dir,obs_datasets[n],obs_names[n]))
        if isinstance(obs_grids[n],str) and 'def' in obs_grids[n].lower():
            obs_grids[n] = {'type': 'uniform', 'res': regions[n][regions[n].keys()[0]]['reslon']}
        if 'res' not in mod_grid.keys(): mod_grid['res'] = obs_grids[n]['res']
        if isinstance(methods[n],str) and 'def' in methods[n].lower():
            if obs_grids[n]['res'] <= mod_grid['res']: methods[n] = 'upscale_average'
            else: methods[n] = 'downscale_radius'
        if isinstance(parameters[n],str) and 'def' in parameters[n].lower():
            parameters[n] = dict()
            if methods[n] == 'upscale_average':
                parameters[n]['half_pix'] = mod_grid['res']/2.
                parameters[n]['min_obs']  = int((mod_grid['res']/obs_grids[n]['res'])**2/2)
            elif methods[n] == 'downscale_radius':
                parameters[n]['radius']   = 1.15*mod_grid['res']
                parameters[n]['min_obs']  = 1
            elif methods[n] == 'nearest':
                parameters[n]['d_max']    = obs_grids[n]['res']/2.
                parameters[n]['min_obs']  = 1
        if methods[n] == 'upscale_average':
            if parameters[n]['half_pix'] == 0:
                raise Exception('Half pixel value cannot be null with upscale_average method.')
        elif methods[n] == 'downscale_radius':
            if parameters[n]['radius'] == 0:
                raise Exception('Radius value cannot be null with downscale_radius method.')
        elif methods[n] == 'nearest':
            if parameters[n]['d_max'] == 0:
                raise Exception('Maximum distance for nearest neighbor method cannot be null.')
        # check grid type for observation
        if not obs_grids[n]['type']=='uniform':
            raise Exception('Observation grid must be uniform.')
    return dict([(v, {'grid': obs_grids[n], 'regions': regions[n], 'method': methods[n], 'parameters': parameters[n]})\
        for v, n in zip(obs_names,range(len(obs_names)))])

def parseSurfaceScreenInfo(obs_names,features,thresholds,mod_pgd_path):
    '''
    Parse information to screen observation variables according to static surface characteristics.
    Arguments:
        obs_names (list): names of observation variables (str).
        features (list): surface characteristics to check for each observation variable (list).
        thresholds (list): max values of surface characteristics for observation point to be valid (list).
        mod_pgd_path (str): directory of PGD_fractown.nc
    Returns:
        parsed information (dict).
    '''
    features_name_in_nc = {'altitude': 'ZS', 'urban_fraction': 'FRAC_TOWN'}
    surf_chk_info = dict()
    for obs_var, v in zip(obs_names,range(len(obs_names))):
        if not all([f in features_name_in_nc.keys() for f in features[v]]):
            raise Exception('surf_screen_features must be in ['+', '.join(features_name_in_nc.keys())+']')
        surf_chk_info[obs_var] = dict([(f, {'threshold': t, 'path': mod_pgd_path+'PGD_fractown.nc', 'ncname': features_name_in_nc[f]})\
        for f, t in zip(features[v],thresholds[v])])
    return surf_chk_info

def parseModelScreenInfo(obs_names,mod_names,thres):
    '''
    Parse information about the use of model output to screen observations.
    Arguments:
        obs_names (list): names of observation variables (str).
        mod_names (list): names of model output variables (str).
        thres (list): model output thresholds above which observations are discarded (float).
    Returns:
        parsed information (dict).
    '''
    screen_mod_info = dict()
    for obs_var, n in zip(obs_names,range(len(obs_names))):
        screen_mod_info[obs_var] = dict()
        if mod_names[n]:
            screen_mod_info[obs_var]['mod_out'] = mod_names[n]
            screen_mod_info[obs_var]['thres'] = thres[n]
    return screen_mod_info

def parseRescaleInfo(obs_names,methods,calibs,calib_periods,analysis_period,mod_names,steps,max_miss,exp_name,out_files_dir):
    '''
    Parse information about rescaling of observations.
    Arguments:
        obs_names (list): names of observation variables (str).
        methods (list): names of rescaling methods (str).
        calibs (list): whether parameter calibration is required or not (bool).
        calib_periods (list): start and end years for calibrating the rescale parameters (list of 2 int).
        analysis_period (list): start and end dates for pre-processing (str: yyyy-mm-dd).
        mod_names (list): model output variables used for rescaling (str).
        steps (list): indices of model output time steps (int).
        max_miss (list): maximum allowed missing frequencies in observation time series (float).
        exp_name (str): experiment name ['France_SAFRAN'].
        out_files_dir (str): folder where rescale parameters are saved.
    Returns:
        parsed information (dict).
    '''
    parameters_dir = os.path.normpath(out_files_dir+'/rescale_parameters')
    makeDirIfNeeded(parameters_dir)
    rescale_correction_info = dict()
    for obs_var, n in zip(obs_names,range(len(obs_names))):
        rescale_correction_info[obs_var] = dict()
        if methods[n]:
            parameter_path = '{0}/{1}_{2}_parameters_{3}_{4}_{5}.npy'\
                .format(parameters_dir,exp_name,obs_var,methods[n],calib_periods[n][0],calib_periods[n][1])
            to_calib = calibs[n] or not os.path.isfile(parameter_path)
            rescale_correction_info[obs_var] = {'method': methods[n], 'parameter_path': parameter_path,'calibrate': to_calib,\
                'mod_out': mod_names[n], 'step': steps[n], 'max_miss': max_miss[n], 'period': calib_periods[n]}
            if to_calib:
                if pd.to_datetime(calib_periods[n][0]) < pd.to_datetime(analysis_period[0])\
                or pd.to_datetime(calib_periods[n][1]) > pd.to_datetime(analysis_period[1]):
                    raise Exception('The rescale calibration period of {0} exceeds the analysis period'.format(obs_var))
                if pl.mod(pd.to_datetime(calib_periods[n][1]).month,12) != pd.to_datetime(calib_periods[n][0]).month-1:
                    print('The rescale calibration period of {0} does not cover a whole number of years'.format(obs_var))
                    print('Rescaling method currently does not handle this case.')
                    exit()
                    #ans = raw_input('Are you sure you want to proceed? [y/n] ')
                    #if ans == 'n': exit()
    return rescale_correction_info

# --------------------------------------------------------------
# PUBLIC FUNCTIONS: OBSERVATIONS
# --------------------------------------------------------------

def getObsRegions(obs_root_dir,obs_dataset,obs_var):
    '''
    Read regions characteristics for variable in dataset
    Arguments:
        obs_root_dir (str): root directory of observations
        obs_dataset (str): observation dataset
        obs_var (str): observation variable name.
    Returns:
        regions (dict): regions characteristics
    '''
    # get regions from file
    regions_file = '{0}/{1}/{2}/{1}_{2}_REGIONS.csv'.format(obs_root_dir,obs_dataset,obs_var)
    if not os.path.isfile(regions_file):
        raise Exception('{0} does not exist.'.format(regions_file))
    regions = pd.read_csv(regions_file,sep=',',header=0,dtype={'#NAME':str}).T
    regions.columns = regions.ix['#NAME'].values
    regions.drop('#NAME',inplace=True)
    # check res consistency
    r = regions[regions.keys()[0]]
    res = float(r['lr_lon']-r['ul_lon'])/(r['nlon']-r['pix_center'])
    eps = res*1e-4
    for _,r in regions.iteritems():
        if abs(float(r['lr_lon']-r['ul_lon'])/(r['nlon']-r['pix_center'])-res)>eps or abs(r['reslon']-res)>eps or \
           abs(float(r['ul_lat']-r['lr_lat'])/(r['nlat']-r['pix_center'])-res)>eps or abs(r['reslat']-res)>eps:
            raise Exception('Inconsistencies in {0}.'.format(regions_file))
    return regions

# --------------------------------------------------------------
# PUBLIC FUNCTIONS: SURFEX MODEL
# --------------------------------------------------------------

def getModelGrid(mod_grid,mod_pgd_path=None):
    '''
    Read the coordinates of model grid points from file.
    Arguments:
        mod_grid (dict): types and characteristics of model grid
        mod_pgd_path (str): directory of PGD.nc
    Returns:
        mod_lon (numpy.ndarray): longitude values of model grid points.
        mod_lat (numpy.ndarray): latitude values of model grid points.
    '''
    if mod_grid['type']=='uniform':
        mod_lon,mod_lat = res2lonlat(mod_grid['res'],mod_grid['zone'])
        mod_lon,mod_lat = pl.meshgrid(mod_lon,mod_lat)
    elif mod_pgd_path is not None:
        pgd_file = mod_pgd_path+'PGD.nc'
        if not os.path.isfile(pgd_file):
            raise Exception(pgd_file+' does not exist.')
        if mod_grid['type']=='uniform':
            mod_lon,mod_lat = readFromFile(pgd_file,['REG_LON','REG_LAT'])
        elif mod_grid['type']=='points':
            mod_lon,mod_lat = readFromFile(pgd_file,['XX','XY'])
    else:
        raise Exception('Problem with model grid definition. If not uniform, please provide correct PGD.nc file.')
    mod_lon = mod_lon.ravel()
    mod_lat = mod_lat.ravel()
    return mod_lon,mod_lat

def getModelPatchFractions(mod_pgd_path):
    '''
    Read the patch fractions of model grid points from file.
    Arguments:
        mod_pgd_path (str): directory of PREP.nc
    Returns:
        mod_patch_frac (numpy.ndarray): patch fractions of each model grid point; shape = n_points,n_patches
    '''
    prep_file = mod_pgd_path+'PREP.nc'
    if not os.path.isfile(prep_file):
        raise Exception(prep_file+' does not exist.')
    frac = readFromFile(prep_file,['PATCH'])[0]
    return frac.reshape((frac.shape[0],-1))

def getModelZS(mod_pgd_path):
    '''
    Read the ZS of model grid points from file.
    Arguments:
        mod_pgd_path (str): directory of PGD.nc
    Returns:
        mod_zs (numpy.ndarray): ZS of each model grid point; length = n_points
    '''
    pgd_file = mod_pgd_path+'PGD.nc'
    if not os.path.isfile(pgd_file):
        raise Exception(pgd_file+' does not exist.')
    mod_zs = readFromFile(pgd_file,['ZS'])[0]
    mod_zs = mod_zs.ravel()
    return mod_zs

def readModOutDaily(mod_var,mod_patch_frac,mod_out_dir,curr_date,out_date=None,step=-1,patch=None,new_miss_val=None,NPNT=None):
    '''
    Read a model output variable from a daily output file  for 1 step only, all patches are averaged.
    Arguments:
        mod_var (str): model output variable name.
        mod_patch_frac (numpy.ndarray): patch fractions of each model grid point; shape = n_points,n_patches.
        mod_out_dir (str): directory where model output are stored (in netCDF format).
        curr_date (datetime.datetime): date of the output file name.
        out_date (datetime.datetime): date for which values are read.
        step (int): time step to read in the output file
        patch (array): patches to read (None if aggregated)
        new_miss_val (float): new value to be assigned to missing observations; if default (None), no substitution is performed.
        NPNT (int): number of points (by default: mod_patch_frac.shape[1])
    Returns:
        mod_values (numpy.ndarray): model output variable values for one day; shape = (n_points,)
    '''
    curr_date_str = curr_date.strftime('%Y%m%d')
    next_date_str = (curr_date+pd.DateOffset(days=1)).strftime('%Y%m%d%H')
    mod_values,var_name = None,False
    # Paths of isba/trip output files
    # !! Please do not change order, it is important for reading aggregated variables !!
    # !! Normalement PROGNOSTIC_next - DIAGNOSTICS - PROGNOSTIC_curr - DIAG_CUMUL - etc !!
    out_path = []
    out_path.append(mod_out_dir+'/ISBA_PROGNOSTIC_'+next_date_str+'_0.nc')
    out_path.append(mod_out_dir+'/ISBA_DIAGNOSTICS_'+curr_date_str+'_0.nc')
    out_path.append(mod_out_dir+'/ISBA_PROGNOSTIC_'+curr_date_str+'_0.nc')
    out_path.append(mod_out_dir+'/ISBA_DIAG_CUMUL_'+curr_date_str+'_0.nc')
    out_path.append(mod_out_dir+'/ISBA_ANALYSIS_'+next_date_str+'_0.nc')
    out_path.append(mod_out_dir+'/TRIP_DIAG_'+curr_date_str+'_0.nc')
    if out_date is None: Idat = [step]
    # Try to find file containing the required variable for the required date
    missing = 0
    for out_file in out_path:
        if os.path.isfile(out_file):
            nc = Dataset(out_file)
            if   patch is None and mod_var+'_ISBA' in nc.variables.keys(): var_name = mod_var+'_ISBA'
            elif                   mod_var         in nc.variables.keys(): var_name = mod_var
            elif                   mod_var+'_P'    in nc.variables.keys(): var_name = mod_var+'_P'
            if not var_name: nc.close()
            else:
                if out_date is not None:
                    file_date = num2date(nc.variables['time'][:],nc.variables['time'].units)+\
                                         pd.DateOffset(days=1 if 'TRIP' in os.path.basename(out_file) else 0)
                    Idat = [i for i in range(len(file_date)) if out_date==file_date[i]]
                if len(Idat)==0: nc.close(); var_name = False
                else           : step = Idat[0]; break
        else:
            missing += 1
    if missing == len(out_path):
        errmsg = 'Can not find any of the following files: \n'
        for out_file in out_path:
            errmsg += '%s\n' % out_file
        raise Exception(errmsg)
    # Missing dates in ISBA_ANALYSIS (normally, only the 1st date is missing)
    if not var_name:# and ('ANAL_INCR' in mod_var or 'HO_' in mod_var):
        # 1st day does not exist for analysis increment, so put NaN
        #print('Filling day {1} for {0} with NaN values'.format(mod_var,curr_date_str))
        miss_val   = pl.nan if new_miss_val is None else new_miss_val
        if NPNT is None: NPNT = mod_patch_frac.shape[1]
        mod_values = miss_val*pl.ones((NPNT,))
        return mod_values
    if not var_name:
        raise Exception('The model variable {0} cannot be found! Date {1}'.format(mod_var,curr_date_str))
    # Read variable values at the desired time step and reshape (aggregate if needed)
    v = nc.variables[var_name]
    #if patch is not None: mod_values = v[:][pl.ix_([step],[patch])].reshape(-1)
    if patch is not None: mod_values = v[step,patch,...].reshape(-1)
    elif v._getdims()[1]=='Number_of_Patches': mod_values = pl.nansum(mod_patch_frac*v[step,...].reshape(v.shape[1],-1),0)
    else: mod_values = v[step,...].reshape(-1)
    miss_val = v.getncattr('_FillValue')
    nc.close()
    # Handle missing values
    if pl.ma.is_masked(mod_values): mod_values = mod_values.filled()
    #if pl.all(mod_values == miss_val): print('ALL NANS ON '+curr_date_str)
    if new_miss_val is not None: mod_values[mod_values==miss_val] = new_miss_val
    return mod_values

def readModOut(mod_var,patch_frac,mod_out_dir,mod_dates=None,period=None,assim_hour=None,hour=None,patch=None,NPNT=None):
    '''
    Reads model output (only averaged values over the whole pixel to compare them with observations)
    Arguments:
        mod_var (str): model variable to read
        patch_frac (numpy.ndarray): patch fractions of each model grid point; shape = n_patches,n_points
        mod_out_dir (str): directory where model output are stored (in netCDF format).
        mod_dates (datetime array): dates of the model files to read
        period (list): start and end dates (str: yyyy-mm-dd)
        assim_hour (int): hour of assimilation (hour of the start of each daily simulation)
        hour (int in range(24)): hour to read
        patch (array): patches to read (None if aggregated)
        NPNT (int): number of points (required if it differs from patch_frac.shape[1], as with TRIP variables)
    Returns:
        mod_values (pandas.DataFrame): model values for the whole period 
    '''
    if mod_dates is None and (period is None or assim_hour is None):
        raise Exception('Either mod_dates or period/assim_hour must be specified.')
    # Model dates
    if mod_dates is None: # Read all daily data within period
        mod_dates = pd.date_range(pd.to_datetime(period[0]),pd.to_datetime(period[1]),freq='D')+pd.DateOffset(hours=assim_hour)
    else:
        mod_dates = pd.DatetimeIndex(mod_dates)
    # Output dates
    if assim_hour is None: assim_hour = mod_dates[0].hour
    if hour is None: hour = assim_hour
    out_dates = mod_dates+pd.DateOffset(hours=hour-assim_hour)
    offset = 1*(hour<=assim_hour)
    # Initialization
    if NPNT is None: NPNT = patch_frac.shape[1]
    data2read = pd.DataFrame(index=out_dates,columns=pl.arange(NPNT),dtype='f4')
    # Loop over dates
    for i in trange(offset,len(out_dates)):
        data2read.ix[out_dates[i]] = readModOutDaily(mod_var,patch_frac,mod_out_dir,mod_dates[i-offset],\
                                              out_date=out_dates[i],patch=patch,new_miss_val=pl.nan,NPNT=NPNT)
    return data2read


# --------------------------------------------------------------
# PUBLIC FUNCTIONS: MISC
# --------------------------------------------------------------

def memmapOnPickle(fname):
    '''
    Works for 2D DataFrame
    '''
    def memmapVariable(fid,var,memo):
      found_var_def = False
      found_binstring = False
      found_mark = False
      found_shp = False
      found_fmt = False
      found_endian = False
      found_variable = False
      shp = []
      add_ns = ''
      order = 'F'
      while not found_variable:
        o,a,p = pkt.genops(fid).next()
        if o.name=='GLOBAL':
          pos = fid.tell()
          o1,a1,p1 = pkt.genops(fid).next()
          if o1.name=='BINPUT': memo[a1] = a
          fid.seek(pos)
        if not found_var_def:
          if o.name=='GLOBAL' and var in a: found_var_def = True
          elif o.name=='BINGET' and a in memo.keys() and var in memo[a]: found_var_def = True
        elif not found_binstring:
          if o.name=='SHORT_BINSTRING' and a=='b': found_binstring = True
        elif not found_mark:
          if o.name=='MARK': found_mark = True
        elif not found_shp:
          if o.name[:6]=='BININT': shp.append(a)
          if 'TUPLE' in o.name: shp,found_shp = shp[-int(o.name[5:]):],True
        elif not found_fmt:
          if 'SHORT_BINSTRING' in o.name: fmt,found_fmt = a,True
        elif not found_endian:
          if o.name=='SHORT_BINSTRING': endian,found_endian = a,True
        elif o.name=='SHORT_BINSTRING' and 'ns' in a: add_ns = '[ns]'
        else:
          if o.name == 'NEWTRUE': order = 'C'
          if fid.read(1)=='T': found_variable = True
          fid.seek(-1,1)
      offset = fid.tell()+5
      while 1 in shp: shp.remove(1)
      values = pl.memmap(fname,endian+fmt+add_ns,mode='r',offset=offset,shape=tuple(shp[::-1]),order=order)
      fid.seek(5+pl.dtype(endian+fmt+add_ns).itemsize*reduce(lambda x, y: x*y, shp),1)
      return values
    fid = open(fname,'rb')
    memo = dict()
    dat = memmapVariable(fid,'DatetimeIndex',memo).copy()
    val = memmapVariable(fid,'numpy ndarray',memo)
    fid.close()
    return dat,val

def findPDataFiles(obs_var, period, out_dir) :
    '''
    Returns a list of booleans to indicate if PData file already exists?
    Arguments:
        obs_var (str): name of the observed variable
        period (list): start and end dates
        out_dir (str): path to where the PData file are saved
    Returns:
        exist_pathPickles (list of bools): list of booleans to indicate if PData file already exists
        list_pathPickles (list of str): list of path filenames
    '''
    exist_pathPickles = [] 
    list_pathPickles = []
    years = range(pd.to_datetime(period[0]).year,pd.to_datetime(period[1]).year+1)
    for y in years:
        year_period = ['{0}-01-01'.format(str(y)),'{0}-12-31'.format(str(y))]
        pathPickle  = os.path.normpath('{0}/{1}_{2}_{3}.PData'.format(out_dir, obs_var, year_period[0], year_period[1]))
        exist_pathPickles.append(os.path.isfile(pathPickle))
        list_pathPickles.append(pathPickle)
    return exist_pathPickles, list_pathPickles
    
def readPDataFiles(list_pathPickles,data2read,IPNT=None):
    '''
    Returns the data saved in Pickle files.
    Arguments:
        list_pathPickles (list of str): list of path filenames
        data2read (pandas.DataFrame): data read from PData files
        IPNT (array of bool): mask for points to read
    Returns:
        read_dates (bool array): dates that have been read
    '''
    read_dates = pl.zeros(data2read.shape[0],dtype=bool)
    for pathPickle in tqdm(list_pathPickles):
        try:
            dat,values = memmapOnPickle(pathPickle)
        except:
            tmp = pd.read_pickle(pathPickle)
            dat = tmp.index.values
            values = tmp.values
        datIn = [d in data2read.index.values for d in dat]
        datOut = [d in dat for d in data2read.index.values]
        if IPNT is None: data2read.ix[datOut,:] = values[pl.ix_(datIn)]
        else           : data2read.ix[datOut,:] = values[pl.ix_(datIn,IPNT)]
        read_dates[pl.argwhere(datOut)] = True
    return read_dates

def writePDataFiles(obs_var,mod_dates,obs_values,out_dir):
    '''
    Write PData files per year of obs_values.
    Arguments:
        obs_var (str): name of the observed variable
        mod_dates (datetime array): dates of the model
        obs_values (pandas.DataFrame)
        out_dir (str): path to where the PData file are saved
    '''
    if not os.path.isdir(out_dir): os.makedirs(out_dir)
    IFILL = obs_values==FILL
    obs_values[obs_values==FILL] = pl.nan
    years = range(mod_dates[0].year,(mod_dates[-1]+pd.DateOffset(days=360)).year)
    for y in years:
        year_period = ['{0}-01-01'.format(str(y)),'{0}-12-31'.format(str(y))]
        pickle_path = os.path.normpath('{0}/{1}_{2}_{3}.PData'.format(out_dir, obs_var, year_period[0], year_period[1]))
        obs_values.ix[mod_dates.year==y].to_pickle(pickle_path)
        print('{0} from {1} to {2} saved in {3}'.format(obs_var,'{0}-01-01'.format(str(y)),'{0}-12-31'.format(str(y)),pickle_path))
    obs_values[IFILL] = FILL

def createPDataFromModel(var,period,assim_hour,patch_frac,mod_dir,out_dir,patch=None,NPNT=None):
    '''
    Creates PData files from model outputs
    Arguments:
        var (str): name of the model variable
        period (list): start and end dates (str: yyyy-mm-dd)
        assim_hour (int): hour of assimilation (hour of the start of each daily simulation)
        patch_frac (numpy.ndarray): patch fractions of each model grid point; shape = n_patches,n_points
        mod_dir (str): directory of model outputs (model or analysis)
        out_dir (str): path to where the PData file are saved
        patch (array): patches to read (None if aggregated)
        NPNT (int): number of points (required if it differs from patch_frac.shape[1], as with TRIP variables)
    '''
    mod_dates = pd.date_range(pd.to_datetime(period[0]),pd.to_datetime(period[1]),freq='D')+pd.DateOffset(hours=assim_hour)
    data_values = readModOut(var, patch_frac, mod_dir, mod_dates=mod_dates, patch=patch, NPNT=NPNT)
    pkl_name = var if patch is None else var+'_{0:02d}_{1}'.format(patch+1,PATCH_NAMES[patch])
    writePDataFiles(pkl_name,mod_dates,data_values,out_dir)

def readFromFile(file_name,var_names,row_slice=None,col_slice=None,row_invert=False,curr_date=None,usr_slice=None):
    '''
    Read values from netcdf or hdf5 file
    Arguments:
        file_name (str): full path to the file to read
        var_names (list): names of variables to read (can be of type str for single variable)
        row_slice/col_slice (slices): indices to slice the variables to the output region
        row_invert (bool): if True, row order is inverted
        curr_date (str or pandas.DateTime): date at which data must be read
        usr_slice (dict of slices): indices to slice the variables over any dimension (given by its name)
    Returns:
        obs_values (list): values read from file
    '''
    # determine format
    if file_name[-2:] not in ('h5','nc'):
        raise Exception('filename extension must be h5 of nc')
    file_format = file_name[-2:]
    if isinstance(var_names,str): var_names = [var_names]
    obs_values = []
    if file_format == 'nc':
        fid = Dataset(file_name)
        # loop over variables
        for i in range(len(var_names)):
            v = fid.variables[var_names[i]]
            # get slices over each dimension
            shp = v.shape
            sl = []
            for s in shp: sl = sl+[slice(0,s)]
            if type(usr_slice) is dict:
                for nusr,susr in usr_slice.iteritems():
                    if all([nusr not in d.lower() for d in v.dimensions]) or susr is None: continue
                    iusr = pl.argmax([nusr in d.lower() for d in v.dimensions])
                    sl[iusr] = susr if type(susr) is slice else slice(susr,susr+1)
            ilon = pl.argmax(['lon' in d.lower() for d in v.dimensions])
            ilat = pl.argmax(['lat' in d.lower() for d in v.dimensions])
            sl_lon = sl[ilon] if col_slice is None else col_slice
            sl_lat = sl[ilat] if row_slice is None else row_slice
            if row_invert: sl_lat = slice(-sl_lat.start-1,-sl_lat.stop-1,-1)
            sl[ilon],sl[ilat] = sl_lon,sl_lat
            # get slices over the time dimension if curr_date is given
            if 'time' in v.dimensions and curr_date is not None:
                if isinstance(curr_date,str): curr_date = pd.to_datetime(curr_date)
                obs_dates = num2date(fid.variables['time'][:],fid.variables['time'].units)
                idat = [i for i in range(len(obs_dates)) if obs_dates[i].strftime('%Y-%m-%d')==curr_date.strftime('%Y-%m-%d')]
                if len(idat)==0:
                    obs_values.append(FILL*pl.ones((pl.absolute(sl_lat.stop-sl_lat.start),pl.absolute(sl_lon.stop-sl_lon.start))))
                    continue
                sl[pl.argmax(['time' in d.lower() for d in v.dimensions])] = slice(idat[0],idat[0]+1)
            # extract variable
            tmp = v[sl]
            tmp = tmp.astype(float)
            # handle missing values
            if not isinstance(tmp,pl.ma.core.MaskedArray): tmp = pl.ma.masked_array(tmp)
            for a in v.ncattrs():
                if a.lower() in ['missing_value','_fillvalue']:
                    tmp.mask = pl.logical_or(tmp.mask,(tmp==float(v.getncattr(a))))
                    break
            tmp = tmp.filled(FILL)
            # add variable to output
            obs_values.append(tmp)
        fid.close()
    elif file_format == 'h5':
        fid = h5File(file_name,'r')
        for i in range(len(var_names)):
            v = fid.get(var_names[i])
            tmp = v[:].squeeze()
            if row_invert: tmp = tmp[::-1,:]
            if row_slice is not None and col_slice is not None: tmp = tmp[row_slice,col_slice]
            tmp = tmp.astype(float)
            if not type(tmp) == pl.ma.core.MaskedArray: tmp = pl.ma.masked_array(tmp)
            for a in v.attrs.keys():
                if a.lower() in ['missing_value','_fillvalue']: tmp.mask = pl.logical_or(tmp.mask,(tmp==float(v.attrs.get(a))))
                if 'scal' in a.lower(): tmp *= float(v.attrs.get(a))
                if 'offset' in a.lower(): tmp += float(v.attrs.get(a))
            tmp = tmp.filled(FILL)
            obs_values.append(tmp)
        fid.close()
    return obs_values

def res2lonlat(res,zone,pix_center=0,cut_buffer=0):
    '''
    Returns longitude and latitude vectors inside zone with specific resolution
    Arguments:
        res (float): spatial resolution
        zone (2 by 2 float array): [[ul_lon,ul_lat],[lr_lon,lr_lat]]
        pix_center (bool): if true, the ul coord are the coord of the center of the pixel
        cut_buffer (float): buffer zone (in degrees)
    Returns:
        lon/lat (np.ndarray): longitude/latitude vectors
    '''
    #lon = pl.arange(zone[0][0]-cut_buffer+(1-pix_center)*res/2,zone[1][0]+cut_buffer+pix_center*res/2,res)
    #lat = pl.arange(zone[1][1]-cut_buffer+(1-pix_center)*res/2,zone[0][1]+cut_buffer+pix_center*res/2,res)
    lon = pl.arange(zone[0][0]-cut_buffer+(1-pix_center)*res/2,zone[1][0]+cut_buffer,res)
    lat = pl.arange(zone[1][1]-cut_buffer+(1-pix_center)*res/2,zone[0][1]+cut_buffer,res)
    return lon,lat

def makeDirIfNeeded(folder):
    '''
    Create directory (full path) if not already existing.
    '''
    if not os.path.isdir(folder):
        os.makedirs(folder)
        print('\nDirectory {0} created.'.format(folder))

def extractGZ(file_path):
    '''
    Extracts file (file_path) from its g-zipped version (file_path+'.gz')
    '''
    f = open(file_path,'wb')
    f.write(gzOpen(file_path+'.gz','r').read())
    f.close()

def importLocals(f):
    '''
    Imports local variables from the calling function
    Arguments:
        f (inspect.frame): frame of the current function from the inspect module (f = currentframe())
    '''
    for key in [k for k in f.f_back.f_locals.keys() if k not in getargvalues(f).locals.keys()]:
        f.f_globals[key] = f.f_back.f_locals[key]

# --------------------------------------------------------------
# OBSOLETE FUNCTIONS
# --------------------------------------------------------------

def showProgress(iter_ix,num_iters,bar_len=49):
    '''
    Show progress bar with percentage.
    Arguments:
        iter_ix: iteration index (zero-based).
        num_iters: total number of iterations.
        bar_len (int): bar length.
    '''
    progress = float(iter_ix+1)/num_iters
    arrow_len = int(round((bar_len*progress)))
    stdout.write('\r[{0}>{1}] {2}%'.format('-'*arrow_len,' '*(bar_len-arrow_len),str(int(round(100*progress))).zfill(2)))
    stdout.flush()
    if iter_ix+1 == num_iters:
        print('\n')

def getDimName(list_keys,start_str):
    '''
    Get the dimension name starting with start_str from the list list_keys 
    (useful to get the right name of a variable in a netcdf file)
    '''
    start_str = start_str.lower()
    matching = [s for s in list_keys if start_str in s.lower()]
    true_match = str([s for s in matching if s.lower().startswith(start_str)][0])
    return true_match

