# --------------------------------------------------------------------------------------------
# IMPORT PYTHON LIBRARIES
# --------------------------------------------------------------------------------------------
import locale
from matplotlib.ticker import FormatStrFormatter
if not 'argv' in locals(): from sys import argv

# --------------------------------------------------------------------------------------------
# IMPORT DEVELOPED LIBRARIES
# --------------------------------------------------------------------------------------------
from ldasModule import *
from ldasMapSet import mapSet, uniformTicks, Basemap, Figure

# --------------------------------------------------------------------------------------------
# SETTINGS
# --------------------------------------------------------------------------------------------
# Matplotlib settings
pl.ioff()
pl.rc('mathtext',default='regular') # use normal font for math expressions in rendered text
# Dictionary mapping observation names to model output names
OBS_MOD = {'FAPAR':'FAPAR', 'FAPAR_V1':'FAPAR', 'FAPAR_V2':'FAPAR', 'FAPAR300':'FAPAR',
           'LAI':'LAI', 'LAI_V1':'LAI', 'LAI_V2':'LAI', 'LAI300':'LAI',
           'LST':'TS', 'LST-6':'TS', 'LST-12':'TS',
           'SA':'TALB',
           'SWI':'WG2', 'SSM':'WG2', 'SSM_COMBINED':'WG2', 'SSM_ACTIVE':'WG2', 'SSM_PASSIVE':'WG2',
           'SIF':'GPP', 'SIF-L3':'GPP', 'SIF-F':'GPP', 'FLCGPP':'GPPC',
           'SIF-L3_V27':'GPPC','SIF-L3_V27_d':'GPPC','SIF-L3_V27_p':'GPPC',
           'E_ACTUAL':'EVAPC', 'E_t':'LETRC', 'E_b':'LEGC', 'E_i':'LERC', 'FLLE':'EVAPC', 'FLGPP':'GPPC'}
OBS_MOD_sameUnit = {'FAPAR':True, 'FAPAR_V1':True, 'FAPAR_V2':True, 'FAPAR300':True,
           'LAI':True, 'LAI_V1':True, 'LAI_V2':True, 'LAI300':True,
           'LST':True, 'LST-6':True, 'LST-12':True,
           'SA':True,
           'SWI':True, 'SSM':True, 'SSM_COMBINED':True, 'SSM_ACTIVE':True, 'SSM_PASSIVE':True,
           'SIF':False, 'SIF-L3':False, 'SIF-F':False, 'FLCGPP':True,
           'SIF-L3_V27':False,'SIF-L3_V27_d':False,'SIF-L3_V27_p':False,
           'E_ACTUAL':True, 'E_t':True, 'E_b':True, 'E_i':True, 'FLLE':True, 'FLGPP':True}
# Dictionary mapping user time window to Pandas resampling frequency, and to figure prefix
TW_USER_PD       = {'day':'D'       , 'month':'M'      , 'season':'M'       , 'year':'A'     }
TW_USER_FIG      = {'day':'daily'   , 'month':'monthly', 'season':'seasonal', 'year':'annual', 'all':'all_dates_ave'} 
TW_USER_STRFTIME = {'day':'%Y-%m'   , 'month':'%Y-%m'  , 'season':'%b'      , 'year':'%Y'    , 'all':'%Y'}
# Dictionary mapping variable names to y-axis labels for graphical output
VAR_YLAB = {'FAPAR': r'FAPAR', 'LAI': r'LAI', 'SA': r'Surface albedo', 'SWI': r'Soil moisture', 'SSM':r'Surface Soil Moisture',\
            'WG1': r'Surface soil moisture', 'WG2': r'Root zone soil moisture', 'WG3': r'Soil liquid layer 3', 'WR': r'Interception Reservoir',\
            'SIF': r'Sun Induced Fluorescence', 'SIF-F':r'Sun Induced Fluorescence', 'SIF-L3':r'Sun Induced Fluorescence', 'GPP': r'Gross primary production',\
            'SIF-L3_V27': r'Sun Induced Fluorescence','SIF-L3_V27_d': r'Sun Induced Fluorescence','SIF-L3_V27_p': r'Sun Induced Fluorescence',\
            'LST': r'Surface temperature', 'LST-6': r'Surface temperature (6 am)','LST-12': r'Surface temperature (12 pm)',\
            'DRAINC': r'Daily cumulated surface drainage', 'RUNOFFC': r'Daily cumulated surface runoff',\
            'EVAPC': r'Daily cumulated evapotranspiration', 'GPPC': r'Daily cumulated gross primary production',\
            'LETRC': r'Daily cumulated transpiration', 'RC_ECO': r'Daily cumulated ecosystem respiration',\
            'LERC':r'Daily cumulated canopy direct evaporation','LEGC':r'Daily cumulated ground evaporation',\
            'NEE': r'Net ecosystem exchange', 'NEEC': r'Daily cumulated net ecosystem exchange'}
# Dictionary mapping variable names to units for graphical output
VAR_UNIT = {'FAPAR': r'[-]', 'LAI': r'[$m^{2}m^{-2}$]', 'SA': r'[-]', 'SWI': r'$[m^{3}m^{-3}]$', 'SSM': r'$[m^{3}m^{-3}]$',\
            'WG1': r'$[m^{3}m^{-3}]$', 'WG2': r'$[m^{3}m^{-3}]$', 'WG3': r'$[m^{3}m^{-3}]$', 'WR': r'[xxx]',\
            'SIF': r'$[mW.m^{-2}sr^{-1}nm^{-1}]$', 'SIF-F':r'$[mW.m^{-2}sr^{-1}nm^{-1}]$', 'SIF-L3':r'$[mW.m^{-2}sr^{-1}nm^{-1}]$', 'GPP': r'$[mg(C)m^{-2}s^{-1}]$',\
            'SIF-L3_V27': r'$[mW.m^{-2}sr^{-1}nm^{-1}]$','SIF-L3_V27_d': r'$[mW.m^{-2}sr^{-1}nm^{-1}]$','SIF-L3_V27_p': r'$[mW.m^{-2}sr^{-1}nm^{-1}]$',\
            'LST': r'[$^{\circ}C$]' , 'LST-6': r'[$^{\circ}C$]', 'LST-12': r'[$^{\circ}C$]',\
            'DRAINC': r'$kg/m^{2}/day$', 'RUNOFFC': r'$kg/m^{2}/day$', 'EVAPC': r'$kg/m^{2}/day$', 'GPPC': r'$g(C)/m^{2}/day$',\
            'E_ACTUAL':r'$kg/m2/day$', 'E_t':r'$kg/m2/day$', 'E_i':r'$kg/m2/day$', 'E_b':r'$kg/m2/day$',\
            'LETRC': r'$W/m^{2}/day$', 'RC_ECO': r'$g(C)/m^{2}/day$', 'NEE': r'$mg(C)/m^{2}/day$', 'NEEC': r'$mg(C)/m^{2}/day$', 'QDIS': r'$m^{3}/s$',\
            'LERC': r'$W/m^{2}/day$', 'LEGC': r'$W/m^{2}/day$', 'FLGPP':r'$kg/m^{2}/day$', 'FLCGPP':r'$kg/m^{2}/day$'}
# Dictionary mapping variable names to colormap names for graphical output
VAR_CMAP = {'FAPAR': 'rainbow', 'LAI': 'RdYlGn', 'SA': 'spectral', 'SWI': 'Spectral', 'SSM': 'Spectral', 'QDIS':'Spectral_r',\
            'WG1': 'Spectral', 'WG2': 'Spectral', 'WG3': 'Spectral', 'WR': 'Spectral',\
            'SIF': 'RdYlGn', 'SIF-F':'RdYlGn', 'SIF-L3':'RdYlGn', 'GPP': 'RdYlGn', 'FLGPP': 'RdYlGn',\
            'SIF-L3_V27':'RdYlGn','SIF-L3_V27_d':'RdYlGn','SIF-L3_V27_p':'RdYlGn',\
            'LST': 'jet', 'LST-6': 'jet', 'LST-12': 'jet',\
            'DRAINC': 'jet', 'RUNOFFC': 'jet', 'EVAPC': 'jet', 'GPPC': 'RdYlGn',\
            'LETRC': 'jet', 'RC_ECO': 'jet', 'NEE': 'RdYlBu', 'NEEC': 'RdYlBu',\
            'ANA_INCR': 'bwr_r', 'analysis-model': 'bwr_r'}
# Default panel title
DEF_NAMES = ['Obs','Model','Analysis']
# Default figure DPI
DEF_FIG_DPI = 300
# Dictionary mapping data type to line style
D_LINE = {'Model': 'b', 'Analysis': 'r', 'Obs': 'g:.', 'Analysis-Model':'k'}#, 'ANAL_INCR1': 'b', 'ANAL_INCR2':'r', 'ANAL_INCR3':'g', 'ANAL_INCR4':'c', 'ANAL_INCR5':'m', 'ANAL_INCR6':'y', 'ANAL_INCR7':'k', 'ANAL_INCR8':'r', 'ANAL_INCR9':'g', 'ANAL_INCR10':'b'}
# Graphics
SCORES = ('Bias','Correlation','RMSD','SDD','Nb_pts') # names
SCORES_CMAP = {'Bias':'bwr', 'Correlation':'jet', 'RMSD':'jet', 'SDD':'jet', 'Nb_pts':'jet'}
CBAR_MINMAX = {'SWI':[0.,0.3], 'LAI':[0.,3.], 'FAPAR':[0.,1.], 'SA':[0.,0.5], 'SIF':[0.,2.], 'SIF-F':[0.,2.], 'SIF-L3':[0.,2.], 'SIF-L3_V27':[0.,2.], 'SIF-L3_V27_d':[0.,2.], 'SIF-L3_V27_p':[0.,2.], 'LST-6':[-5.,30.], 'LST-12':[-5.,30.], 'SSM':[0.,0.3], 'NEEC':[-2000,2000]}
# Minimum number of observations (strictly minimum) during a month to perform a comparison with the model
MIN_OBS_MONTH = {'FAPAR':1, 'LAI':1, 'LST':1, 'LST-6':1, 'LST-12':1, 'SA':1, 'SWI':2, 'SIF':1, 'SIF-F':1, 'SIF-L3':1, 'SSM':1, 'SIF-L3_V27':1, 'SIF-L3_V27_d':1, 'SIF-L3_V27_p':1}
# Miscellaneous
# GPP is in kg(CO2)/m2/s and needs to be converted in mg(C)/m2/s (M(CO2)=44g/mol,M(C)=12g/mol,1kg=1e3g, or 1kg=1e6mg)
VAR_CONV_UNIT = {'GPP': 44/12e6, 'GPPC': 44/12e3, 'R_ECO': 44/12e6, 'RC_ECO': 44/12e3, 'LETRC':2.5008*1e6, 'LERC':2.5008*1e6, 'LEGC':2.5008*1e6, 'FLLE':2.5008}
VAR_OFFSET = {'LST': -273.15, 'TS': -273.15}
VAR2FILTER_VEG = ('LAI','FAPAR','SIF','SIF-F','SIF-L3','GPPC','RC_ECO','NEEC','NEE','ANAL_INCR1','ANAL_INCR2','ANAL_INCR3','ANAL_INCR4','ANAL_INCR5', \
                  'ANAL_INCR6','ANAL_INCR7','ANAL_INCR8','INNOV_1','INNOV_2','RESID_1','RESID_2','E_ACTUAL','FLGPP','FLCGPP','E_t','E_i','FLLE', \
                  'SIF-L3_V27','SIF-L3_V27_d','SIF-L3_V27_p')
VAR2FILTER_COAST = {'SIF-F':0.3,'SIF-L3':0.3,'E_ACTUAL':0.3,'FLGPP':0.3,'E_i':0.3,'E_t':0.3,'E_b':0.3,'FLLE':0.3,'FLCGPP':0.3,'SIF-L3_V27':0.3,'SIF-L3_V27_d':0.3,'SIF-L3_V27_p':0.3}
# For date names
locale.setlocale(locale.LC_TIME,locale='en_US')





# --------------------------------------------------------------------------------------------
# FUNCTIONS: DATA MANIPULATION
# --------------------------------------------------------------------------------------------

def readAllData(obs_var,period,assim_hour,patch_frac,out_all_dir,post_from_pickle,mod_ana_dir, \
                to_read=DEF_NAMES,patch=None,IPNT=None,filter_nan=True,map_set=None):
    '''
    Reads observations, as well as model output of openloop and analysis runs, for a variable and a period.
    Arguments:
        obs_var (str): observation variable name.
        period (list): start and end dates (str: yyyy-mm-dd).
        assim_hour (int in range(24)): hour of assimilation.
        patch_frac (numpy.ndarray): patch fractions of each model grid point; shape = n_patches,n_points.
        out_all_dir (dict of str): folder where observations, model and analysis outputs are saved as a Pickle file.
        post_from_pickle (dict of bool): if observations/model/analysis can be read in Pickle file
        mod_ana_dir (dict): directories of model openloop and analysis outputs.
        to_read (list): data to be read; in D_LINE.keys()
        patch (int): number id of patch to read (None if aggregated variable)
        IPNT (numpy.ndarray): boolean mask vector (of length equals number of points)
        filter_nan (bool): True to remove dates and points with nan values
    Returns:
        data (pandas.Panel): items are observations, model, analysis; major_axis are time indices; minor_axis are grid-points.
    '''
    # If only one experiment is to be read
    if isinstance(out_all_dir,str) and isinstance(post_from_pickle,bool):
        if len(to_read)>1: to_read = ['Model']
        out_all_dir = {'Model':out_all_dir}
        post_from_pickle = {'Model':post_from_pickle}
        mod_ana_dir = {'Model':mod_ana_dir}
    # By default hour = assim_hour
    hour = assim_hour
    # Special handling for temperature (hour can be specified in obs name)
    if 'LST' in obs_var:
        hour = int(obs_var[4:]) if obs_var > 'LST' else None
        obs_var = obs_var[0:3]
    # Model dates
    out_dates = pd.date_range(pd.to_datetime(period[0]),pd.to_datetime(period[1]),freq='D')+pd.DateOffset(hours=hour)
    # Points that are not sea
    if IPNT is None: IPNT = pl.ones(patch_frac.shape[1],bool)
    print('  Only keep points where patch_frac is defined: {0} points kept over {1} points'.format(pl.sum(IPNT),len(IPNT)))
    # For vegetation variables (e.g. LAI), discard areas with no vegetation
    if obs_var in VAR2FILTER_VEG:
        patch_frac_tmp = pl.where(patch_frac==FILL,0,patch_frac)
        IPNT = pl.logical_and(IPNT,patch_frac_tmp[0,:]+patch_frac_tmp[1,:]<0.90)
        print('  {0} is a vegetation parameter: only keep points that are vegetated (NO+ROCK<0.10): {1} points kept over {2} points'.format(obs_var,pl.sum(IPNT),len(IPNT)))
    if obs_var in VAR2FILTER_COAST and map_set is not None:
        IPNT = pl.logical_and(IPNT,coast_filter(map_set,VAR2FILTER_COAST[obs_var]))
        print('  {0} is a parameter where points too close to coastline are screened out (dist<{1}deg): {2} points kept over {3} points'.format(obs_var,VAR2FILTER_COAST[obs_var],pl.sum(IPNT),len(IPNT)))
    data = pd.Panel(items=to_read,major_axis=out_dates,minor_axis=pl.arange(len(IPNT))[IPNT],dtype='f4')
    #
    for ds in to_read:
        print('  ## '+ds+' ##')
        out_dir = out_all_dir[ds.replace('raw','')]
        new_obs_var = obs_var
        if ds in ['Model','Analysis']:
            if not obs_var in OBS_MOD.keys(): OBS_MOD.update({obs_var:obs_var})
            new_obs_var = OBS_MOD[obs_var]
        if ds=='rawObs': new_obs_var = 'raw_'+obs_var
        # find Pickle files and return boolean if they all exist
        pkl_name = new_obs_var if patch is None or ds=='Obs' else new_obs_var+'_{0:02d}_{1}'.format(patch+1,PATCH_NAMES[patch]) # [add] or ds=='Obs'
        if obs_var=='LST': pkl_name = pkl_name+'_{0}'.format(hour) # [add] LST-hour
        exist_pathPickles, list_pathPickles = findPDataFiles(pkl_name, period, out_dir)
        if ds=='Obs' and post_from_pickle[ds] and any(exist_pathPickles):
            # it can happen that observations are not available for the whole period, but we can still use the available period
            real_list_pathPickles = [list_pathPickles[i] for i in range(len(list_pathPickles)) if exist_pathPickles[i]]
            print('  Reading {0} {1} from PData files in {2}'.format(pkl_name,ds,out_dir))
            readPDataFiles(real_list_pathPickles,data[ds],IPNT)
        elif post_from_pickle[ds] and all(exist_pathPickles):
            # if post_from_pickle and all Pickle files are present: read PData files
            print('  Reading {0} {1} from PData files in {2}'.format(pkl_name,ds,out_dir))
            readPDataFiles(list_pathPickles,data[ds],IPNT)
        elif ds in ['Model','Analysis']:
            # else: read data and create PData file if post_from_pickle==True
            print('  Reading {0} {1} from SFX outputs in {1}'.format(pkl_name,ds,mod_ana_dir[ds]))
            data_values = readModOut(new_obs_var, patch_frac, mod_ana_dir[ds], \
                assim_hour=assim_hour, period=period, hour=hour, patch=patch, NPNT=len(IPNT))
            if post_from_pickle[ds]: writePDataFiles(pkl_name,out_dates,data_values,out_dir)
            data[ds].values[:] = data_values.values[:,IPNT]
        else:
            # read obs from scratch as in ldasPre (readObs function)
            raise Exception('Not implemented yet: {0} should be read from scratch as in ldasPre'.format(ds))
        if 'Obs' in ds and obs_var in MIN_OBS_MONTH:
            # if not enough data during the month, then discard the data (defined in MIN_OBS_MONTH)
            grouped = data[ds].groupby((data[ds].index.year, data[ds].index.month))
            for name, group in grouped: data[ds].ix[grouped.indices[name], pl.logical_not(pl.isnan(group.values)).sum(0) < MIN_OBS_MONTH[obs_var]] = pl.nan
        # Offset and scale factor
        new_obs_var = new_obs_var.replace('raw_','')
        if new_obs_var in VAR_CONV_UNIT: data[ds] = data[ds]/VAR_CONV_UNIT[new_obs_var]
        if new_obs_var in VAR_OFFSET: data[ds] = data[ds]+VAR_OFFSET[new_obs_var]
        #if 'Obs' in ds and 'FLLE' in obs_var : data[ds]=data[ds]/2.5008
    #
    # Trick for LST: start at 2010-01-01 and put NaN before (only data from 2009-08-10 so it messes with the statistics and maps)
    #if 'LST' in obs_var: data[:,data.major_axis<pd.to_datetime('20100701',format='%Y%m%d'),:] = pl.nan
    #
    # Remove nan values
    if filter_nan: data = remove_nan(data)
    # Print percentage of NaN values
    print('  --> {:.0f} % of NaN values in read data.'.format(pl.sum(pl.isnan(data.values))*100./pl.prod(data.shape)))
    return data
#
def readAllAnalysis(var_names, analysis_period, assim_hour, patch_frac, out_all_dir, post_from_pickle, ana_info, IPNT=None, filter_nan=True):
    '''
    Reads outputs from ISBA_ANALYSIS files and create a single Pandas Panel
    Arguments:
        var_names (str): variable names.
        analysis_period (list): start and end dates (str: yyyy-mm-dd).
        assim_hour (int in range(24)): hour of assimilation.
        patch_frac (numpy.ndarray): patch fractions of each model grid point; shape = n_patches,n_points.
        out_all_dir (dict of str): folder where observations, model and analysis outputs are saved as a Pickle file.
        post_from_pickle (dict of bool): if observations/model/analysis can be read in Pickle file
        ana_info (dict): information about model analysis outputs.
        IPNT (numpy.ndarray): boolean mask vector (of length equals number of points)
        filter_nan (bool): True to remove dates and points with nan values
    Returns:
        var (pandas.Panel): Panel of data
    '''
    var1 = readAllData(var_names[0], analysis_period, assim_hour, patch_frac, \
                        out_all_dir, post_from_pickle, ana_info, ['Analysis'], IPNT=IPNT, filter_nan=False)
    var = pd.Panel(pl.nan*pl.empty((len(var_names), len(var1.major_axis), len(var1.minor_axis))),\
                    var_names, var1.major_axis, var1.minor_axis)
    var[var_names[0]].values[:,:] = var1['Analysis'].values[:,:]
    for i in range(1,len(var_names)):
        varx =  readAllData(var_names[i], analysis_period, assim_hour, patch_frac, \
                             out_all_dir, post_from_pickle, ana_info, ['Analysis'], IPNT=IPNT, filter_nan=False)
        var[var_names[i]].values[:,:] = varx['Analysis'].values[:,:]
    #incr.values[pl.where(incr.values==0)] = pl.nan
    if filter_nan: var = remove_nan(var)
    return var
#
def coast_filter(map_set,dist_to_coast):
    '''
    '''
    lonmin,lonmax = map_set.projection.lonmin,map_set.projection.lonmax
    latmin,latmax = map_set.projection.latmin,map_set.projection.latmax
    mres = 'i' if lonmax-lonmin < 20. and latmax-latmin <20. else 'l'
    m = Basemap(projection='cyl',llcrnrlat=latmin-1.,llcrnrlon=lonmin-1.,urcrnrlat=latmax+1.,urcrnrlon=lonmax+1.,resolution=mres)
    coasts_lon = pl.hstack(pl.array(m.coastpolygons[i][0]) for i in range(len(m.coastpolygontypes)) if m.coastpolygontypes[i]==1) # 2==lakes
    coasts_lat = pl.hstack(pl.array(m.coastpolygons[i][1]) for i in range(len(m.coastpolygontypes)) if m.coastpolygontypes[i]==1) 
    # get lat/lon of model
    mod_lon,mod_lat = map_set.mod_lon.ravel(),map_set.mod_lat.ravel()
    if map_set.mod_grid_type == 'points':
        mod_lon = mod_lon[map_set.remap]
        mod_lat = mod_lat[map_set.remap]
    bool_dist_deg = pl.ones((len(mod_lon),),dtype=bool)
    for i in range(len(mod_lon)):
        dists_deg = pl.sqrt((coasts_lon-mod_lon[i])**2+(coasts_lat-mod_lat[i])**2)
        if min(dists_deg)<dist_to_coast: bool_dist_deg[i]=False
    return bool_dist_deg
#
def remove_nan(data):
    '''
    Removes nan values.
    Arguments:
        data (pandas.Panel): input data
    Returns:
        data (pandas.Panel): output data
    '''
    print('  Only keep points and time steps where all data sources are available ')
    if isinstance(data,pd.core.panel.Panel):
        # Only retain points where all the data sources are available 
        I = pl.any(pl.isnan(data),axis=0)
        data.values[:].reshape((len(data.items),-1))[:,I.flatten()] = pl.nan
        #for it in data.items: data[it].values[I] = pl.nan # if the above line does not work properly
        # Remove dates with only NaN-value data
        IDAT = pl.logical_not(pl.isnan(data.values).all((0,2)))
        data = data.ix[:,IDAT,:]
    elif isinstance(data,pd.core.frame.DataFrame):
        IDAT = pl.logical_not(pl.isnan(data.values).any((1,)))
        data = data.ix[IDAT,:]
    return data
#
def applyTimeWindow(data, time_window):
    '''
    Applies a time window to average data.
    Arguments:
        data (pandas.Panel): data to average
        time_window (str): time window used for averaging
    Returns:
        data_tw (pandas.Panel)
    '''
    if time_window=='day': # and data.major_axis.freq.freqstr=='D':
        data_tw = data.copy()
        #data_tw.major_axis = data_tw.major_axis.to_period()
    elif time_window in TW_USER_PD.keys():
        if len(data.items) == 1:
            data_tw = pd.Panel({data.items[0]: data[data.items[0]].resample(TW_USER_PD[time_window]).mean()})
        else:
            data_tw = data.resample(TW_USER_PD[time_window], axis=1).mean()
            #data_tw = pd.Panel({i: data[i].resample(TW_USER_PD[time_window]).mean() for i in data.items})
        data_tw.major_axis = data_tw.major_axis.to_period()
        if time_window=='season':    # season==by month
            data_tw_copy = data_tw.copy()
            data_tw = pd.Panel(pl.ones((len(data_tw.items), 12, len(data_tw.minor_axis)))*pl.nan, data_tw.items, \
                               pd.datetools.MONTHS, data_tw.minor_axis)
            for month in range(1,13):
                bool_in_month = data_tw_copy.major_axis.month==month
                data_tw.ix[:,month-1,:] = data_tw_copy.ix[:,bool_in_month,:].mean(axis=1)
    elif time_window=='all':
        data_tw = pd.Panel(pl.ones((len(data.items), 1, len(data.minor_axis)))*pl.nan, data.items, \
          ['{0}-{1}'.format(data.major_axis[0].strftime('%Y%m'), data.major_axis[-1].strftime('%Y%m'))] , data.minor_axis)
        data_tw.ix[:,0,:] = data.mean(axis=1)
    else: 
        raise Exception('Cannot perform time resampling because {0} is not defined as an implemented option '+\
                        'in TW_USER_PD. Please choose between: {1}'.format(time_window,TW_USER_PD.keys()))
    # Remove NaN-value data
    #data_tw = data_tw.ix[:,pl.logical_not(pl.isnan(data_tw.values).all((0,2))),:]
    return data_tw
#
def compute_scores_per_point(data, time_window):
    '''
    Compute scores of Model_vs_Obs, and Analysis_vs_Obs per point.
    Arguments:
        data (pandas.Panel): data to use to compute the statistics (must contain 'Obs' and at least another item)
        time_window (str): time window used for averaging (used for figure name only here)
    Returns:
        scores (pandas.Panel4D): scores (bias, RMSD, SDD, correlation) between the model/analysis and the obs
    '''
    # Find the diagnostics to run
    diagno_var_obs = []
    if not 'Obs'  in data.items: raise Exception('There is no observation available for computing statistical scores')
    if 'Model'    in data.items: diagno_var_obs.append('Model')
    if 'Analysis' in data.items: diagno_var_obs.append('Analysis')
    if 'Model'    in data.items and 'Analysis' in data.items: diagno_diff_bool = True
    diagnosToCompute = diagno_var_obs+['Analysis-Model']*diagno_diff_bool
    # Initialization of the scores 4D panel
    if time_window=='season':
        periods = pd.datetools.MONTHS
        id_tmp  = pl.array([pl.where(data.major_axis.month==month) for month in range(1,13)])
    elif time_window=='all':
        #periods = [data.major_axis[0], data.major_axis[-1]]
        #id_tmp = [pl.where(pl.logical_and(data.major_axis>=periods[0], data.major_axis<=periods[1]))]
        periods = ['{0}-{1}'.format(data.major_axis[0].strftime('%Y%m'), data.major_axis[-1].strftime('%Y%m'))]
        id_tmp  = pl.array([[range(len(data.major_axis))]])
    else:
        periods = pd.period_range(data.major_axis[0], data.major_axis[-1], freq=TW_USER_PD[time_window])
        id_tmp  = pl.array([pl.where(pl.logical_and(data.major_axis>=periods[i].start_time,data.major_axis<=periods[i].end_time)) for i in range(len(periods))])
    scores = pd.Panel4D(pl.empty((len(SCORES),len(diagnosToCompute),len(periods),data.minor_axis.size))*pl.nan,SCORES,diagnosToCompute,periods,data.minor_axis.values)
    # Computation of the statistics
    for id_time,i in zip(id_tmp,range(len(id_tmp))):
        if id_time[0].any(): # check if not empty
            for diagno in diagno_var_obs:
                differences = data[diagno].ix[id_time[0],:]-data['Obs'].ix[id_time[0],:]
                #scores['Nb_pts'][diagno].values[i,:] = differences.values[pl.where(pl.isnan(differences.values)==False)].size
                for ca in range(0,differences.shape[1]):
                    scores['Nb_pts'][diagno].values[i,ca] = differences.values[pl.where(pl.isnan(differences.values[:,ca])==False),ca].size
                scores['Nb_pts'][diagno].values[pl.where(scores['Nb_pts'][diagno].values==0.)] = pl.nan
                scores['Bias'][diagno].values[i,:] = differences.mean(axis=0)
                scores['RMSD'][diagno].values[i,:] = pl.sqrt((differences**2).mean(axis=0))
                scores['SDD'][diagno].values[i,:]  = pl.sqrt(((differences-differences.mean(axis=0))**2).mean(axis=0))
                corr_tmp = data[diagno].ix[id_time[0],:].corrwith(data['Obs'].ix[id_time[0],:],axis=0)
                corr_tmp.values[pl.where(pl.isinf(corr_tmp.values))] = pl.nan
                corr_tmp.values[pl.where(corr_tmp.values>=1. )] =  1.
                corr_tmp.values[pl.where(corr_tmp.values<=-1.)] = -1.
                scores['Correlation'][diagno].values[i,:]  = corr_tmp
            if diagno_diff_bool:
                for j in range(len(SCORES)):
                  scores.ix[j,'Analysis-Model',i,:] = scores.ix[j,'Analysis',i,:]-scores.ix[j,'Model',i,:]
                scores['Bias']['Analysis-Model'].values[pl.where(scores['Bias']['Analysis-Model'].values==0.)] = pl.nan
                scores['RMSD']['Analysis-Model'].values[pl.where(scores['RMSD']['Analysis-Model'].values==0.)] = pl.nan
                scores['SDD']['Analysis-Model'].values[pl.where(scores['SDD']['Analysis-Model'].values==0.)] = pl.nan
                scores['Nb_pts']['Analysis-Model'].values[pl.where(scores['Nb_pts']['Analysis-Model'].values==0.)] = pl.nan
                scores['Correlation']['Analysis-Model'].values[pl.where(scores['Correlation']['Analysis-Model'].values==0.)] = pl.nan
    return scores
#
def compute_scores_for_all_points(data, time_window):
    '''
    Compute scores of Model_vs_Obs, and Analysis_vs_Obs for all points.
    Arguments:
        data (pandas.Panel): data to use to compute the statistics (must contain 'Obs' and at least another item)
        time_window (str): time window used for averaging (used for figure name only here)
    Returns:
        scores (pandas.Panel4D): scores (bias, RMSD, SDD, correlation) between the model/analysis and the obs
    '''
    # Find the diagnostics to run
    diagno_var_obs = []
    if not 'Obs'  in data.items: raise Exception('There is no observation available for computing statistical scores')
    if 'Model'    in data.items: diagno_var_obs.append('Model')
    if 'Analysis' in data.items: diagno_var_obs.append('Analysis')
    if 'Model'    in data.items and 'Analysis' in data.items: diagno_diff_bool = True
    diagnosToCompute = diagno_var_obs+['Analysis-Model']*diagno_diff_bool
    # Initialization of the scores 4D panel
    if time_window=='season':
        periods = pd.datetools.MONTHS
        id_tmp  = pl.array([pl.where(data.major_axis.month==month) for month in range(1,13)])
    elif time_window=='all':
        periods = ['{0}-{1}'.format(data.major_axis[0].strftime('%Y%m'), data.major_axis[-1].strftime('%Y%m'))]
        id_tmp  = pl.array([[range(len(data.major_axis))]])
    else:
        periods = pd.period_range(data.major_axis[0], data.major_axis[-1], freq=TW_USER_PD[time_window])
        id_tmp  = pl.array([pl.where(pl.logical_and(data.major_axis>=periods[i].start_time,data.major_axis<=periods[i].end_time)) for i in range(len(periods))])
    scores = pd.Panel(pl.empty((len(SCORES),len(diagnosToCompute),len(periods)))*pl.nan,SCORES,diagnosToCompute,periods)
    # Computation of the statistics
    for id_time,i in zip(id_tmp,range(len(id_tmp))):
        if id_time[0].any(): # check if not empty 
            for diagno in diagno_var_obs:
                sim = data[diagno].values[id_time[0],:].reshape((len(id_time[0])*data.shape[2]))
                obs = data['Obs'].values[id_time[0],:].reshape((len(id_time[0])*data.shape[2]))
                sim,obs = sim[pl.where(pl.logical_and(pl.isnan(sim)==False,pl.isnan(obs)==False))],\
                          obs[pl.where(pl.logical_and(pl.isnan(sim)==False,pl.isnan(obs)==False))]
                differences = sim - obs
                scores['Nb_pts'].ix[diagno].values[i]      = differences.size
                scores['Bias'].ix[diagno].values[i]        = differences.mean()
                scores['RMSD'].ix[diagno].values[i]        = pl.sqrt((differences**2).mean())
                scores['SDD'].ix[diagno].values[i]         = pl.sqrt(((differences-differences.mean())**2).mean())
                scores['Correlation'].ix[diagno].values[i] = pl.corrcoef(pl.array([sim,obs]))[0,1]
            if diagno_diff_bool:
                scores.ix[:,'Analysis-Model',i] = scores.ix[:,'Analysis',i]-scores.ix[:,'Model',i]
                if scores['Bias'].ix['Analysis-Model'].values[i]==0.: scores['Bias'].ix['Analysis-Model'].values[i] = pl.nan
                if scores['RMSD'].ix['Analysis-Model'].values[i]==0.: scores['RMSD'].ix['Analysis-Model'].values[i] = pl.nan
                if scores['SDD'].ix['Analysis-Model'].values[i]==0.: scores['SDD'].ix['Analysis-Model'].values[i] = pl.nan
                if scores['Nb_pts'].ix['Analysis-Model'].values[i]==0.: scores['Nb_pts'].ix['Analysis-Model'].values[i] = pl.nan
                if scores['Correlation'].ix['Analysis-Model'].values[i]==0.: scores['Correlation'].ix['Analysis-Model'].values[i] = pl.nan
    return scores

# --------------------------------------------------------------------------------------------
# FUNCTIONS: MAPS DRAWING
# --------------------------------------------------------------------------------------------

def getSlices4MapArray(n_rows,row_thre=6):
    '''
    Compute the best split of rows for drawing an array of maps,
    Arguments:
        n_rows (int): total number of rows.
        row_thre (int): maximum number of rows in a map array.
    Returns:
        slices (list): slices for splitting the rows.
    '''
    n_figs  = pl.ceil(n_rows/float(row_thre))
    d       = pl.ceil(n_rows/n_figs)
    start   = pl.linspace(0,d*(n_figs-1),n_figs).astype(int)
    end     = pl.linspace(d,d*n_figs,n_figs).astype(int)
    end[-1] = min(end[-1],n_rows)
    return [slice(s,e) for s,e in zip(start,end)]
#
def draw_map_vars(map_set, obs_var, data, graphics_dir=None, time_window=None, fig_name_default=None, \
                  fig_dpi=DEF_FIG_DPI, **kwargs):
    '''
    Draws maps of variables contained in the data. Creates multiple if too many maps to draw.
    Arguments:
        map_set (class mapSet)
        obs_var (str): name of the observed variable
        data (pandas.Panel): data to plot
        graphics_dir (str): path to where the figures are saved
        time_window (str): time window used for averaging (used for figure name only here)
        fig_name_default (str): path+filename to save the figure
        fig_dpi (int): DPI quality of the output figure
    kwargs passed to map_set.matrixMaps
        minim,maxim (float): minimum and maximum vaues of the colorbar
        col_labels (str list): labels of columns
    '''
    if graphics_dir is not None and not os.path.isdir(graphics_dir): os.makedirs(graphics_dir)
    if not obs_var in OBS_MOD_sameUnit.keys(): OBS_MOD_sameUnit[obs_var]=True
    data2draw = data.copy()
    col_order = range(len(data2draw.items))
    if len(data2draw.items)==3 and 'Model' in data2draw.items and 'Analysis' in data2draw.items: 
        col_order[0] = pl.where(data2draw.items=='Model')[0][0]
        col_order[1] = pl.where(data2draw.items=='Analysis')[0][0]
        col_order[2] = pl.where(pl.logical_and(data2draw.items!='Model',data2draw.items!='Analysis'))[0][0] #col_order=[1,2,0]
    if len(data2draw.items)==4 and 'Model' in data2draw.items and 'Analysis' in data2draw.items and 'Obs' in data2draw.items: 
        col_order[0] = pl.where(data2draw.items=='Model')[0][0]
        col_order[1] = pl.where(data2draw.items=='Obs')[0][0]
        col_order[2] = pl.where(data2draw.items=='Analysis')[0][0]
        col_order[3] = pl.where(pl.logical_and(data2draw.items!='Model',pl.logical_and(data2draw.items!='Analysis',data2draw.items!='Obs')))[0][0] #col_order=[1,2,0,3]
    # Number of maps to draw
    n_maps = data2draw.shape[1]
    # Draw and save each figure of maps
    if map_set.fGUI: fig = dict()
    for m_sl in getSlices4MapArray(n_maps):
        obs_var_short = obs_var
        while obs_var_short.rfind('_')!=-1 and obs_var_short not in VAR_CMAP.keys():
            obs_var_short = obs_var_short[0:obs_var_short.rfind('_')]
        if not obs_var_short in VAR_CMAP.keys(): VAR_CMAP.update({obs_var_short:'Spectral'})
        if not ('minim_diff' in kwargs.keys() and 'maxim_diff' in kwargs.keys()):
            minim_diff, maxim_diff = 0.,0.
            if time_window=='month' or time_window=='year':
                if time_window=='month': id_same_minmax = pl.find(data2draw.major_axis.year==data2draw[:,m_sl,:].major_axis.year[0])
                if time_window=='year' : id_same_minmax = pl.arange(len(data2draw.major_axis))
                minim_diff = pl.nanmean(data2draw.ix[:,id_same_minmax,:][data2draw.items[-1]].values) - \
                           2*pl.nanstd( data2draw.ix[:,id_same_minmax,:][data2draw.items[-1]].values) 
                maxim_diff = pl.nanmean(data2draw.ix[:,id_same_minmax,:][data2draw.items[-1]].values) + \
                           2*pl.nanstd( data2draw.ix[:,id_same_minmax,:][data2draw.items[-1]].values) 
                minim_diff,maxim_diff = min(minim_diff,-1*maxim_diff),max(maxim_diff,-1*minim_diff)
            kwargs['minim_diff'],kwargs['maxim_diff'] = minim_diff,maxim_diff
        map_array = map_set.matrixMaps(data2draw.ix[:,m_sl,:], VAR_CMAP[obs_var_short], \
                                       to_draw='countries',col_order=col_order, obs_mod_same_unit=OBS_MOD_sameUnit[obs_var],\
                                       **kwargs)
        #
        if map_set.fGUI:
            fig_name = '{0} - {1}'.format(obs_var,TW_USER_FIG[time_window])
            fig[fig_name] = map_array
        if fig_name_default is None and (graphics_dir is None or time_window is None): continue
        if fig_name_default is None:
            fig_name = os.path.normpath('{0}/{1}_{2}_maps_{3}_to_{4}.png'.\
                                    format(graphics_dir,obs_var,TW_USER_FIG[time_window], \
                                    data2draw.ix[:,m_sl,:].major_axis[0], data2draw.ix[:,m_sl,:].major_axis[-1])).replace(' ','_')
        else: fig_name = fig_name_default
        #
        map_array.savefig(fig_name, dpi=fig_dpi, bbox_inches='tight')
        print(fig_name)
        pl.close()
    if map_set.fGUI: return fig
#
def draw_map_analysis(map_set, data, diag_names, graphics_dir=None, diag=None, time_window=None, fig_name_default=None, fig_dpi=DEF_FIG_DPI, **kwargs):
    '''
    Draws maps for the analysis diagnostics.
    Arguments:
        map_set (class mapSet)
        data (pandas.Panel): data to plot
        diag_names (array of str): name of the diagnostic variables
        graphics_dir (str): path to where the figures are saved
        diag (str): name of the dignostic (increments, jacobian, innovation or residual), for the figure name
        time_window (str): time window used for averaging (used for figure name only here)
        fig_name_default (str): path+filename to save the figure
        fig_dpi (int): DPI quality of the output figure
        kwargs passed to map_set.matrixMaps
    '''
    if graphics_dir is not None and not os.path.isdir(graphics_dir): os.makedirs(graphics_dir)
    # Number of maps to draw
    n_maps = data.shape[1]
    # Column labels
    col_labels = [diag_names[i] for i in range(data.shape[0])]
    # Draw and save each figure of maps
    for m_sl in getSlices4MapArray(n_maps):
        map_array = map_set.matrixIncrMaps(data.ix[:,m_sl,:], VAR_CMAP['ANA_INCR'], \
                                           to_draw='countries', col_labels=col_labels, \
                                           **kwargs)
        if fig_name_default is None and (graphics_dir is None or diag is None or time_window is None): continue
        if fig_name_default is None:
            fig_name = os.path.normpath('{0}/{1}_{2}_maps_{3}_to_{4}.png'.\
                                        format(graphics_dir,diag,TW_USER_FIG[time_window], \
                                        data.ix[:,m_sl,:].major_axis[0], data.ix[:,m_sl,:].major_axis[-1])).replace(' ','_')
        else: fig_name = fig_name_default
        #
        map_array.savefig(fig_name, dpi=fig_dpi, bbox_inches='tight')
        print(fig_name)
        pl.close()
#
def draw_map_stats(map_set, scores, graphics_dir=None, obs_var=None, time_window=None, fig_name_default=None, fig_dpi=DEF_FIG_DPI, col_labels=None):
    '''
    Draws maps of scores. Creates multiple files if needed.
    Arguments:
        map_set (class mapSet)
        scores (pandas.Panel4D): scores to plot, should contain Correlation, Bias, RMSD, SDD
        graphics_dir (str): path to where the figures are saved
        obs_var (str): name of the observed variable
        time_window (str): time window used for averaging (used for figure name only here)
        fig_name_default (str): path+filename to save the figure
        fig_dpi (int): DPI quality of the output figure
    kwargs passed to map_set.matrixMaps
        col_labels (str list): labels of columns
    '''
    if graphics_dir is not None and not os.path.isdir(graphics_dir): os.makedirs(graphics_dir)
    if map_set.fGUI: fig = dict()
    for score2draw in SCORES: #list(set(SCORES)-set(['Nb_pts'])):
        data2draw = scores[score2draw]
        if col_labels is None: col_labels = data2draw.items
        col_labels = [l for l in col_labels if l!='Obs']
        # Number of maps to draw
        n_maps = data2draw.shape[1]
        # If Correlation, impose limits to maps
        if score2draw=='Correlation':
            if obs_var=='SWI': minim,maxim=-1.,1. 
            else:              minim,maxim= 0.,1. 
        if score2draw=='Bias': 
            minim,maxim=pl.nanmean(data2draw.values[0,:,:].ravel())-2*pl.nanstd(data2draw.values[0,:,:].ravel()),\
                        pl.nanmean(data2draw.values[0,:,:].ravel())+2*pl.nanstd(data2draw.values[0,:,:].ravel())
            minim,maxim=min(minim,-1*maxim),max(maxim,-1*minim)
        if score2draw=='RMSD': minim,maxim=0.,pl.nanmean(data2draw.values[0,:,:].ravel())+2*pl.nanstd(data2draw.values[0,:,:].ravel())
        if score2draw=='SDD':  minim,maxim=0.,pl.nanmean(data2draw.values[0,:,:].ravel())+2*pl.nanstd(data2draw.values[0,:,:].ravel())
        if score2draw=='Nb_pts':
            if   time_window=='month': minim,maxim=0,30
            elif time_window=='year' : minim,maxim=0,365
            else: minim,maxim=0,pl.nanmean(data2draw.values[0,:,:].ravel())+2*pl.nanstd(data2draw.values[0,:,:].ravel())
        # Draw and save each figure of maps
        for m_sl in getSlices4MapArray(n_maps):
            if score2draw=='Nb_pts': 
                map_array = map_set.matrixMaps(data2draw.ix[0:1,m_sl,:], SCORES_CMAP[score2draw], \
                                           to_draw='countries',minim=minim, maxim=maxim, col_labels=col_labels[0:1])
            else:
                minim_diff, maxim_diff = 0.,0.
                if time_window=='month' or time_window=='year':
                    if time_window=='month': id_same_minmax = pl.find(data2draw.major_axis.year==data2draw[:,m_sl,:].major_axis.year[0])
                    if time_window=='year' : id_same_minmax = pl.arange(len(data2draw.major_axis))
                    minim_diff = pl.nanmean(data2draw.ix[:,id_same_minmax,:][data2draw.items[-1]].values) - \
                               2*pl.nanstd( data2draw.ix[:,id_same_minmax,:][data2draw.items[-1]].values) 
                    maxim_diff = pl.nanmean(data2draw.ix[:,id_same_minmax,:][data2draw.items[-1]].values) + \
                               2*pl.nanstd( data2draw.ix[:,id_same_minmax,:][data2draw.items[-1]].values) 
                    minim_diff,maxim_diff = min(minim_diff,-1*maxim_diff),max(maxim_diff,-1*minim_diff)
                map_array = map_set.matrixMaps(data2draw.ix[:,m_sl,:], SCORES_CMAP[score2draw], \
                                               to_draw='countries',minim=minim, maxim=maxim, col_labels=col_labels, \
                                               minim_diff=minim_diff, maxim_diff=maxim_diff)
            #
            if map_set.fGUI:
                fig_name = '{0} {1} - {2}'.format(obs_var,score2draw,TW_USER_FIG[time_window])
                fig[fig_name] = map_array
            if fig_name_default is None and (graphics_dir is None or obs_var is None or time_window is None): continue
            if fig_name_default is None:
                fig_name = os.path.normpath('{0}/{1}_{2}_maps_{3}_to_{4}.png'.\
                                            format(graphics_dir,obs_var+'_'+score2draw,TW_USER_FIG[time_window], \
                                            data2draw.ix[:,m_sl,:].major_axis[0], data2draw.ix[:,m_sl,:].major_axis[-1])).replace(' ','_')
            else: fig_name = fig_name_default
            #
            map_array.savefig(fig_name, dpi=fig_dpi, bbox_inches='tight')
            print(fig_name)
            pl.close()
    if map_set.fGUI: return fig

# --------------------------------------------------------------------------------------------
# FUNCTIONS: TIME SERIES DRAWING
# --------------------------------------------------------------------------------------------

def draw_series(data_ts, var, graphics_dir=None, time_window=None, fig_name=None, col_labels=None, fig_dpi=300, \
                ylabel_format='%g', ylabel=None, minim=0., maxim=0., \
                right_axis=None, right_ylabel=None, right_minim=0., right_maxim=0., \
                style=None, fill_between=None, grid=False, legend_loc=3, legend_ncol=None, \
                fGUI=False):
    '''
    Draws time series, possibly with left/right y-axes
    Arguments:
        data_ts (pandas.DataFrame): time series data
        var (str): name of the variable (for figure name and labels)
    Optional arguments:
        graphics_dir (str): path to where the figures are saved
        time_window (str): time window used for averaging (used for figure name only here)
        fig_name (str): path+filename to save the figure
        col_labels (str list): labels of columns
        fig_dpi (int): DPI quality of the output figure
        ylabel_format (str): format of the y-axis ticks
        ylabel (str): label of the y-axis
        minim/maxim (float): limits of the y-axis
        right_axis (str list): names of the items to plot on the right y-axis
        right_ylabel (str): label of the right y-axis
        right_minim/right_maxim (float): limits of the right y-axis
        style (str list): lines style
        fill_between (str list): names of the items used by fill_between ploting function (usually ['Model','Analysis'])
        grid (bool): plot grid or not
        legend_loc (int): legend location
        legend_ncol (int): number of columns in the legend
    '''
    col_names = data_ts.columns
    if col_labels is None: col_labels = col_names
    # Reference var name
    var_short = var
    while var_short.rfind('_')!=-1 and var_short not in VAR_YLAB.keys()+VAR_UNIT.keys():
        var_short = var_short[0:var_short.rfind('_')]
    # Special case for seasonal time series
    if isinstance(data_ts.index[0],str): #time_window=='season':
        t = range(0,data_ts.index.size)
        xlim = [-.25, data_ts.index.size-.75]
        xtickpos = range(data_ts.index.size)
        xticklbl = [m[0:3] for m in data_ts.index]
    else:
        t = data_ts.index.to_datetime()
        xlim = [t[0]-pd.DateOffset(months=1), t[-1]+pd.DateOffset(months=1)]
        xtickpos = pd.date_range(t[0]-pd.DateOffset(months=1),t[-1]+pd.DateOffset(months=1),freq='A')+pd.DateOffset(days=1)
        xticklbl = [xtickpos[i].strftime('%Y') for i in range(len(xtickpos))]
    # Plot
    if fGUI:
      fig = Figure(pl.figaspect(.5))
      ax = fig.add_axes((0.06,0.07,0.93,0.84))
    else:
      fig = pl.figure(figsize=pl.figaspect(.5))
      ax = fig.add_subplot(111)
    if right_axis is not None:
      #ToPlotOnFirstAxis,ToPlotOnSecondAxis = list(set(col_names)-set(right_axis)),right_axis
      ToPlotOnFirstAxis,ToPlotOnSecondAxis = col_names-right_axis,right_axis
    else:
      ToPlotOnFirstAxis,ToPlotOnSecondAxis = col_names,[]
    if style is None and all([c in D_LINE for c in col_names]):
        style = [D_LINE[c] for c in col_names]
    if style is not None:
        hp = data_ts.ix[:,col_names].plot(ax=ax,secondary_y=ToPlotOnSecondAxis,style=style,legend=False)
    else:
        hp = data_ts.ix[:,col_names].plot(ax=ax,secondary_y=ToPlotOnSecondAxis,legend=False)
    # Left ax
    ax.set_xlim(xlim)
    ax.set_xticks(xtickpos)
    ax.set_xticklabels(xticklbl)
    if minim!=maxim: ax.set_ylim([minim,maxim])
    if ylabel is None: ylabel = ' '.join((VAR_YLAB[var_short],VAR_UNIT[var_short]))
    ax.set_ylabel(ylabel)
    ax.yaxis.set_major_formatter(FormatStrFormatter(ylabel_format))
    if grid: ax.grid()
    # Right ax
    if len(ToPlotOnSecondAxis)>0:
        #ax.right_ax.set_xlim(xlim)
        #ax.right_ax.set_xticks(xtickpos)
        #ax.right_ax.set_xticklabels(xticklbl)
        if right_minim!=right_maxim: ax.right_ax.set_ylim([right_minim,right_maxim])
        if right_ylabel is None: right_ylabel = ' '.join((VAR_YLAB[var_short],VAR_UNIT[var_short]))
        ax.right_ax.set_ylabel(right_ylabel)
        ax.right_ax.yaxis.set_major_formatter(FormatStrFormatter(ylabel_format))
        #if grid: ax.right_ax.grid()
    # Fill the gap between the lines in red/blue if analysis is > or < than model
    if fill_between is not None and all([c in col_names for c in fill_between]):
        ax.fill_between(t,data_ts[fill_between[0]],data_ts[fill_between[1]], \
                        where=data_ts[fill_between[0]]<=data_ts[fill_between[1]],facecolor='red',alpha=0.1,interpolate=True)
        ax.fill_between(t,data_ts[fill_between[0]],data_ts[fill_between[1]], \
                        where=data_ts[fill_between[0]]>=data_ts[fill_between[1]],facecolor='blue',alpha=0.1,interpolate=True)
    # Legend
    if legend_ncol is None: legend_ncol = len(col_names)
    lines = ax.get_lines() + (ax.right_ax.get_lines() if len(ToPlotOnSecondAxis)>0 else [])
    if legend_loc == 'best':
      ax.legend(lines,col_labels,ncol=legend_ncol,loc='best')
    else:
      ax.legend(lines,col_labels,ncol=legend_ncol,loc=legend_loc,bbox_to_anchor=(0., 1.02, 1., .102),borderaxespad=0.,mode='expand')
    # Save figure
    if not (fig_name is None and (graphics_dir is None or time_window is None)):
        if graphics_dir is not None and not os.path.isdir(graphics_dir): os.makedirs(graphics_dir)
        if fig_name is None:
            if isinstance(data_ts.index[0],str): #time_window=='season':
                fig_name = os.path.normpath('{0}/{1}_{2}_series_{3}_to_{4}.png'.\
                                        format(graphics_dir, var, TW_USER_FIG[time_window], \
                                        data_ts.index[0],data_ts.index[-1]))
            else:
                fig_name = os.path.normpath('{0}/{1}_{2}_series_{3}_to_{4}.png'.\
                                        format(graphics_dir, var, TW_USER_FIG[time_window], \
                                        data_ts.index[0].strftime(TW_USER_STRFTIME[time_window]),\
                                        data_ts.index[-1].strftime(TW_USER_STRFTIME[time_window])))
        fig.savefig(fig_name, dpi=fig_dpi, bbox_inches='tight')
        #bbox_extra_artists=(lgd,)
        print(fig_name)
        pl.close()
    if fGUI: return {'{0} - {1}'.format(var,TW_USER_FIG[time_window]):fig}
#
def draw_series_vars(data_ts, obs, fGUI=False, **kwargs):
    '''
    Draws time series of variables contained in the data. Should be averaged over the whole domain first.
    Arguments:
        data_ts (pandas.DataFrame): time series data
        obs (str): name of the variable (for figure name and labels)
    '''
    # Reference var name
    obs_short = obs
    while obs_short.rfind('_')!=-1 and obs_short not in VAR_YLAB.keys()+VAR_UNIT.keys():
        obs_short = obs_short[0:obs_short.rfind('_')]
    if not obs_short in OBS_MOD_sameUnit.keys(): OBS_MOD_sameUnit[obs_short] = True
    bool_obs_mod_sameunit = OBS_MOD_sameUnit[obs_short]
    if not obs_short in VAR_YLAB: VAR_YLAB.update({obs_short:obs_short})
    if not obs_short in VAR_UNIT: VAR_UNIT.update({obs_short:'[xxx]'})
    col_names = [n for n in data_ts.columns if n != 'Analysis-Model']
    if OBS_MOD_sameUnit[obs_short]:
      ylabel,right_axis = None,None
    else:
      ylabel = ' '.join((VAR_YLAB[OBS_MOD[obs_short]], VAR_UNIT[OBS_MOD[obs_short]]))
      right_axis = ['Obs']
    if fGUI: 
      fig = draw_series(data_ts[col_names],obs,fill_between=['Model','Analysis'], \
                        ylabel=ylabel,right_axis=right_axis,fGUI=fGUI,**kwargs)
      return fig
    else:
      draw_series(data_ts[col_names],obs,fill_between=['Model','Analysis'], \
                  ylabel=ylabel,right_axis=right_axis,**kwargs)
#
def draw_series_analysis(data_tw, col_labels, minim=0., maxim=0., \
                         right_axis=None, ylabel=None, right_ylabel=None, **kwargs):
    '''
    Draws time series of increments.
    Arguments:
        data_tw (pandas.DataFrame): time series data
        col_labels (str list): labels of columns
    '''
    data_ts = data_tw.mean(axis=2)
    col_names = data_ts.columns
    ncols = len(col_names)
    # Manage left/right axes for simple cases (eg LAI/WG)
    if right_axis is None:
      ILAI = ['LAI' in c for c in col_labels]
      IWG  = ['WG'  in c for c in col_labels]
      if any(ILAI) and any(IWG):
        right_axis = col_names[IWG].tolist()
        Ileft,Iright = ILAI,IWG
        ylabel,right_ylabel = 'LAI increments','WG increments'
      elif any(ILAI):
        ylabel = 'LAI increments'
      elif any(IWG):
        ylabel = 'WG increments'
    # Define line styles (plain for left, dashed for right)
    allStyles = ['r-','g-','b-','m-','c-','k-','y-','r-']
    allStyles = [(allStyles[i]+('-' if col_names[i] in right_axis else '')) for i in range(len(col_names))]
    # Set minim/maxim
    if not isinstance(minim,pl.ndarray): minim = pl.array(minim)
    if not isinstance(maxim,pl.ndarray): maxim = pl.array(maxim)
    if len(minim.shape)==0: minim = pl.ones((ncols,))*minim
    if len(maxim.shape)==0: maxim = pl.ones((ncols,))*maxim
    left_minim,left_maxim = tuple(uniformTicks(data_ts[col_names-right_axis], num_ticks=2, zero_centred=True, \
                  minim=pl.nanmin(minim[pl.ix_(Ileft)]), maxim=pl.nanmax(maxim[pl.ix_(Ileft)])))
    if right_axis is not None:
      right_minim,right_maxim = tuple(uniformTicks(data_ts[right_axis], num_ticks=2, zero_centred=True, \
                    minim=pl.nanmin(minim[pl.ix_(Iright)]), maxim=pl.nanmax(maxim[pl.ix_(Iright)])))
    else: right_minim,right_maxim = left_minim,left_maxim
    # Draw figure
    draw_series(data_ts,'Increments',minim=left_minim,maxim=left_maxim, \
                col_labels=col_labels,grid=True,style=allStyles,ylabel=ylabel, \
                right_axis=right_axis,right_ylabel=right_ylabel,right_minim=right_minim,right_maxim=right_maxim, \
                legend_loc='best',legend_ncol=2,ylabel_format='%.1e',**kwargs)
#
def draw_series_analysis_min_max(data_tw, diag_names, graphics_dir=None, diag=None, time_window=None, fig_name=None, fig_dpi=DEF_FIG_DPI, minim=0., maxim=0.):
    '''
    Draws time series of increments.
    Only for daily.
    '''
    # Size of subplots
    nb_subp = len(data_tw.items)
    max_col = 3
    if nb_subp<=max_col: n_row,n_col=1,nb_subp
    if nb_subp>max_col : n_row,n_col=int(pl.ceil(len(data_tw.items)/pl.float32(max_col))),max_col
    # Common time
    t = data_tw.major_axis.to_datetime() 
    # Min/Max
    if not isinstance(minim, pl.ndarray): minim = pl.ones((nb_subp,))*minim
    if not isinstance(maxim, pl.ndarray): maxim = pl.ones((nb_subp,))*maxim
    # Figure
    fig = pl.figure(figsize=pl.figaspect(.5))
    for subp in range(nb_subp):
        ax = fig.add_subplot(n_row,n_col,subp+1)
        ax.plot(t, data_tw[data_tw.items[subp]].max(axis=1).values , 'b-')
        ax.plot(t, data_tw[data_tw.items[subp]].min(axis=1).values , 'g-')
        ax.plot(t, data_tw[data_tw.items[subp]].mean(axis=1).values, 'r-')
        xlim = [t[0]-pd.DateOffset(months=1), t[-1]+pd.DateOffset(months=1)]
        xtickpos = pd.date_range(t[0]-pd.DateOffset(months=1),t[-1]+pd.DateOffset(months=1),freq='A')+pd.DateOffset(days=1)
        xticklbl = [xtickpos[i].strftime('%y') for i in range(len(xtickpos))]
        ax.set_title(diag_names[subp], fontsize=8)
        ax.set_xlim(xlim)
        if minim[subp]==0. and maxim[subp]==0.:
            minim[subp] = pl.nanmin(data_tw[data_tw.items[subp]].min(axis=1).values)
            maxim[subp] = pl.nanmax(data_tw[data_tw.items[subp]].max(axis=1).values)
        ax.set_ylim([minim[subp],maxim[subp]])
        ax.set_xticks(xtickpos)
        ax.set_xticklabels(xticklbl)
        ax.grid()
        for item in ([ax.xaxis.label, ax.yaxis.label] + ax.get_xticklabels() + ax.get_yticklabels()): item.set_fontsize(6)
        #pl.legend(('max','min','mean'), ncol=3, loc='best')
    if not (fig_name is None and (graphics_dir is None or diag is None or time_window is None)):
        if graphics_dir is not None and not os.path.isdir(graphics_dir): os.makedirs(graphics_dir)
        if fig_name is None:
            fig_name = os.path.normpath('{0}/{1}_MINMAX_{2}_series_{3}_to_{4}.png'.\
                                        format(graphics_dir, diag, TW_USER_FIG[time_window], \
                                        data_tw.major_axis[0].strftime(TW_USER_STRFTIME[time_window]),\
                                        data_tw.major_axis[-1].strftime(TW_USER_STRFTIME[time_window])))
        # Save figure
        fig.savefig(fig_name, dpi=fig_dpi, bbox_inches='tight')
        print(fig_name)
        pl.close()
#
def draw_series_analysis_histo(data_tw, diag_names, graphics_dir=None, diag=None, time_window=None, fig_name=None, fig_dpi=DEF_FIG_DPI, num_bins=None):
    '''
    Draws histograms of increments/Jacobians.
    It is supposed here that 1st analysis increment is LAI and the others are WG.
    Only for daily.
    '''
    # Size of subplots
    nb_subp = len(data_tw.items)
    max_col = 4
    if nb_subp<=max_col: n_row,n_col=1,nb_subp
    if nb_subp>max_col : n_row,n_col=int(pl.ceil(len(data_tw.items)/pl.float32(max_col))),max_col
    # Common time
    t = data_tw.major_axis.to_datetime()
    # Figure
    fig = pl.figure(figsize=pl.figaspect(.5))
    for subp in range(nb_subp):
        pl.subplot(n_row,n_col,subp+1)
        data_n = pl.array(data_tw[data_tw.items[subp]].mean(axis=1).values)
        id_noNaN = pl.find(~pl.isnan(data_n))
        mu     = pl.mean(  data_n[id_noNaN])
        sigma  = pl.std(   data_n[id_noNaN])
        median = pl.median(data_n[id_noNaN])
        #print diag_names[subp],"mean %.6f median %.6f sdv %.6f min %.6f max %.6f N %.6f" %(mu, median, sigma,pl.min(data_n[id_noNaN]),pl.max(data_n[id_noNaN]),len(data_n[id_noNaN]))
        #the histogram of the data
        if num_bins is None:
            bin_size = 0.001
            min_edge = pl.mean(data_n[id_noNaN])-pl.std(data_n[id_noNaN])*1.96
            max_edge = pl.mean(data_n[id_noNaN])+pl.std(data_n[id_noNaN])*1.96
            N = (max_edge-min_edge)/bin_size + 1
            num_bins = pl.linspace(min_edge, max_edge, N)
        n, bins, patches = pl.hist(data_n[id_noNaN], num_bins, normed=1, facecolor='green', alpha=0.9)
        y = pl.normpdf(bins, mu, sigma)
        pl.tick_params(axis='both',labelsize=7)
        pl.plot(bins, y,'r-',lw=3)
        pl.axvline(data_n[id_noNaN].mean(), color='b', linestyle='dashed', linewidth=2)
        pl.axvline(0, color='k',linewidth=1)
        pl.ylabel('Probability',fontsize=7)
        pl.title(diag_names[subp], fontsize=7)
        # Tweak spacing to prevent clipping of ylabel
        pl.subplots_adjust(left=0.05)
    if not (fig_name is None and (graphics_dir is None or diag is None or time_window is None)):
        if graphics_dir is not None and not os.path.isdir(graphics_dir): os.makedirs(graphics_dir)
        if fig_name is None:
            fig_name = os.path.normpath('{0}/{1}_{2}_Histo_{3}_to_{4}.png'.\
                                        format(graphics_dir, diag, TW_USER_FIG[time_window], \
                                        data_tw.major_axis[0].strftime(TW_USER_STRFTIME[time_window]),\
                                        data_tw.major_axis[-1].strftime(TW_USER_STRFTIME[time_window])))
        # Save figure
        fig.savefig(fig_name, dpi=fig_dpi, bbox_inches='tight')
        print(fig_name)
        pl.close()
#
def draw_series_raw(data_ts, obs, graphics_dir=None, time_window=None, fig_name=None, **kwargs):
    '''
    Draws time series of observations before and after rescaling. Should be averaged over the whole domain first.
    Arguments:
        data_ts (pandas.DataFrame): time series data (which must include 'Obs','rawObs' and 'Model')
        obs (str): name of the variable (for figure name and labels)
        graphics_dir (str): path to where the figures are saved
        time_window (str): time window used for averaging (used for figure name only here)
        fig_name (str): path+filename to save the figure
    '''
    if not 'Obs' in data_ts.columns or not 'rawObs' in data_ts.columns: 
        raise Exception('The first argument does not contain Obs and rawObs to produce the plot')
    if 'raw_'+obs in VAR_UNIT.keys(): ylabel = 'Original '+obs+' ['+VAR_UNIT['raw_'+obs]+']'
    else:                             ylabel = 'Original '+obs+' [-]'
    right_ylabel = 'Model and Rescaled '+obs+' '+VAR_UNIT[obs]
    #lgd = fig.legend((ax.lines[0],ax.right_ax.lines[0]),('Raw','Rescaled'),ncol=2,loc=8,bbox_to_anchor=(.0725,0,.75,.95),borderaxespad=0.)
    # Figure name
    if fig_name is None and not (graphics_dir is None or time_window is None):
        fig_name = os.path.normpath('{0}/{1}_rescaled_vs_raw_{2}_series_{3}_to_{4}.png'.\
                                    format(graphics_dir, obs, TW_USER_FIG[time_window], \
                                    data_ts.index[0].strftime(TW_USER_STRFTIME[time_window]),\
                                    data_ts.index[-1].strftime(TW_USER_STRFTIME[time_window])))
    # Draw figure
    draw_series(data_ts[['rawObs','Model','Obs']], obs, fig_name=fig_name, \
                col_labels=['Raw Obs.','Model','Rescaled Obs.'],grid=True,style=['r','b','g'],ylabel=ylabel, \
                right_axis=['Obs','Model'],right_ylabel=right_ylabel, **kwargs)
#
def draw_series_stats(scores_ts, obs, plot_nbpts=True, graphics_dir=None, time_window=None, fig_name=None, fig_dpi=DEF_FIG_DPI, col_labels=None, fGUI=False):
    '''
    Draws a figure with 4 subplots (1 for each score) to compare openloop/analysis
    Arguments:
        scores_ts (pandas.Panel)
        obs (str)
        graphics_dir (str)
        time_window (str)
        fig_name (str)
        fig_dpi (int)
    '''
    # Reference var name
    obs_short = obs
    while obs_short.rfind('_')!=-1 and obs_short not in VAR_UNIT.keys():
        obs_short = obs_short[0:obs_short.rfind('_')]
    # To display
    diagnos2show = scores_ts.major_axis
    if 'Analysis-Model' in diagnos2show: diagnos2show = diagnos2show.drop('Analysis-Model')
    if col_labels is None: col_labels = diagnos2show
    col_labels = [col_labels[i] for i in range(len(diagnos2show)) if diagnos2show[i]!='Analysis-Model']
    # 2 cases:
    if isinstance(scores_ts.minor_axis[0],str): #time_window=='season'
        t = range(0, scores_ts.minor_axis.size)
        xlim = [-.25, scores_ts.minor_axis.size-.75]
        xtickpos = range(scores_ts.minor_axis.size)
        xticklbl = [m[0:3] for m in scores_ts.minor_axis]
    else:
        t = scores_ts.minor_axis.to_datetime()
        xlim = [t[0]-pd.DateOffset(months=1), t[-1]+pd.DateOffset(months=1)]
        xtickpos = pd.date_range(t[0]-pd.DateOffset(months=1),t[-1]+pd.DateOffset(months=1),freq='A')+pd.DateOffset(days=1)
        xticklbl = [xtickpos[i].strftime('%Y') for i in range(len(xtickpos))]
    #
    # Plot
    if fGUI: fig = Figure(figsize=[l/3.*4.5 for l in pl.rcParams['figure.figsize']])
    else   : fig = pl.figure(figsize=[l/3.*4.5 for l in pl.rcParams['figure.figsize']])
    for sco, i in zip(list(set(SCORES)-set(['Nb_pts'])),range(len(list(set(SCORES)-set(['Nb_pts']))))):
        ax = fig.add_subplot(2,2,i+1)
        title_str = sco if sco == 'Correlation' else ' '.join((sco,VAR_UNIT[obs_short]))
        ph = []
        for diagno,label in zip(diagnos2show,col_labels):
            p_tmp = ax.plot(t, scores_ts[sco].ix[diagno], D_LINE[diagno], label=label)
            ph.append(p_tmp)
        labs = [ph[i][0].get_label() for i in range(len(diagnos2show))]
        #p_tot = [ph[i] for i in range(len(diagnos2show))]
        ax.set_title(title_str)
        ax.set_xlim(xlim)
        ax.set_xticks(xtickpos)
        ax.set_xticklabels(xticklbl)
        #if sco=='SDD': ax.set_ylim([0,1])
        #elif sco=='Bias': ax.set_ylim([-1,1])
        #elif sco=='RMSD': ax.set_ylim([0,1.5])
        #elif sco=='Correlation': ax.set_ylim([0.5,1])
        ax.grid()
        # Fill the gap between the lines in red/blue if analysis is > or < than model
        if 'Analysis-Model' in scores_ts.major_axis:
            Ibest = pl.absolute(scores_ts[sco].ix['Analysis'])<=pl.absolute(scores_ts[sco].ix['Model'])
            if sco=='Correlation': Ibest = pl.logical_not(Ibest)
            ax.fill_between(t, scores_ts[sco].ix['Analysis'], scores_ts[sco].ix['Model'],\
                         where=Ibest, facecolor='red' , alpha=0.1, interpolate=True)
            ax.fill_between(t, scores_ts[sco].ix['Analysis'], scores_ts[sco].ix['Model'],\
                         where=pl.logical_not(Ibest), facecolor='blue', alpha=0.1, interpolate=True)
    fig.legend(ax.lines, labs, ncol=len(diagnos2show), loc=8, bbox_to_anchor=(.0725, 0, .75, .95), borderaxespad=0.)
    if not (fig_name is None and (graphics_dir is None or time_window is None)):
        if graphics_dir is not None and not os.path.isdir(graphics_dir): os.makedirs(graphics_dir)
        if fig_name is None:
            if isinstance(scores_ts.minor_axis[0],str): #time_window=='season':
                fig_name_tmp = os.path.normpath('{0}/{1}_{2}_scores_{3}_to_{4}.png'.\
                                            format(graphics_dir, obs, TW_USER_FIG[time_window], \
                                            scores_ts.minor_axis[0], scores_ts.minor_axis[-1]))
            else:
                fig_name_tmp = os.path.normpath('{0}/{1}_{2}_scores_{3}_to_{4}.png'.\
                                            format(graphics_dir, obs, TW_USER_FIG[time_window], \
                                            scores_ts.minor_axis[0].strftime(TW_USER_STRFTIME[time_window]), \
                                            scores_ts.minor_axis[-1].strftime(TW_USER_STRFTIME[time_window])))
        else: fig_name_tmp = fig_name
        # Save figure
        fig.savefig(fig_name_tmp, dpi=fig_dpi, bbox_inches='tight')
        print(fig_name_tmp)
        pl.close()
    if fGUI: fig_out = {'{0} - {1}'.format(obs,TW_USER_FIG[time_window]):fig}
    #
    # Only the number of points
    if plot_nbpts:
        sco = 'Nb_pts'
        if fGUI: fig = Figure(figsize=[l/3.*3. for l in pl.rcParams['figure.figsize']])
        else   : fig = pl.figure(figsize=[l/3.*3. for l in pl.rcParams['figure.figsize']])
        ax = fig.add_subplot(1,1,1)
        title_str = 'Number of points'
        ph = []
        for diagno,label in zip(diagnos2show,col_labels):
            p_tmp = ax.plot(t, scores_ts[sco].ix[diagno], D_LINE[diagno], label=label)
            ph.append(p_tmp)
        labs = [ph[i][0].get_label() for i in range(len(diagnos2show))]
        ax.set_title(title_str)
        ax.set_xlim(xlim)
        ax.set_xticks(xtickpos)
        ax.set_xticklabels(xticklbl)
        ax.grid()
        fig.legend(ax.lines, labs, ncol=len(diagnos2show), loc=8, bbox_to_anchor=(.0725, 0, .75, .95), borderaxespad=0.)
        if not (fig_name is None and (graphics_dir is None or time_window is None)):
            if graphics_dir is not None and not os.path.isdir(graphics_dir): os.makedirs(graphics_dir)
            if fig_name is None:
                if isinstance(scores_ts.minor_axis[0],str): #time_window=='season':
                    fig_name_tmp = os.path.normpath('{0}/{1}_{2}_scores_Nb_pts_{3}_to_{4}.png'.\
                                                format(graphics_dir, obs, TW_USER_FIG[time_window], \
                                                scores_ts.minor_axis[0], scores_ts.minor_axis[-1]))
                else:
                    fig_name_tmp = os.path.normpath('{0}/{1}_{2}_scores_Nb_pts_{3}_to_{4}.png'.\
                                                format(graphics_dir, obs, TW_USER_FIG[time_window], \
                                                scores_ts.minor_axis[0].strftime(TW_USER_STRFTIME[time_window]), \
                                                scores_ts.minor_axis[-1].strftime(TW_USER_STRFTIME[time_window])))
            else:
                filename,ext = os.path.splitext(fig_name)
                fig_name_tmp = filename+'_Nb_pts'+ext
            # Save figure
            fig.savefig(fig_name_tmp, dpi=fig_dpi, bbox_inches='tight')
            print(fig_name_tmp)
            pl.close()
    if fGUI:
        fig_out['{0} Number of points - {1}'.format(obs,TW_USER_FIG[time_window])] = fig
        return fig_out
#
def drawScoresTable(scores_ts, obs=None, graphics_dir=None, time_window=None, fig_name=None, fig_dpi=DEF_FIG_DPI):
    '''
    Draws tables of scores for each period of time, for 'Model' and 'Analysis'
    '''
    if graphics_dir is not None and not os.path.isdir(graphics_dir): os.makedirs(graphics_dir)
    # Construct the table with the scores to save
    n_rows = 2*len(scores_ts.minor_axis)
    n_cols = 2+len(scores_ts.items)       # period, version, SCORES
    col_labels  = pl.hstack((['Period','Version'],scores_ts.items))
    col_period  = scores_ts.minor_axis.repeat(2)
    #col_version = scores_ts.major_axis[0:2].repeat(n_rows/2).reshape((2,n_rows/2)).T.ravel()
    col_version = [scores_ts.major_axis[pl.mod(i,2)] for i in range(n_rows)]
    table2draw  = n_rows*[[]]
    for r in range(n_rows):
        table2draw[r] = [col_period[r], col_version[r]] + ['{:.3f}'.format(scores_ts.values[i,int(pl.ceil(r/2.))-int(pl.floor(r/2.)),int(pl.floor(r/2.))]) for i in range(len(scores_ts.items)-1)] \
                        + ['{:.0f}'.format(scores_ts['Nb_pts'].values[int(pl.ceil(r/2.))-int(pl.floor(r/2.)),int(pl.floor(r/2.))])]
    # Figure
    fig = pl.figure(figsize=(1.*n_cols, .3*(n_rows+1)))
    ax = fig.add_subplot(111)
    ax.axis('off')
    the_table = ax.table(cellText=table2draw, colLabels=col_labels, loc='center')
    #for r in range(n_rows+1):
    #    for c in range(n_cols):
    #            cell = the_table.properties()['celld'][0,c]
    #            cell.set_height(2.5*cell.get_height())
    #            cell = the_table.properties()['celld'][r,c]
    #            cell.set_fontsize(20)
    if not (fig_name is None and (graphics_dir is None or obs is None or time_window is None)):
        if fig_name is None:
            if isinstance(scores_ts.minor_axis[0],str):   #time_window=='season'
                fig_name_tmp = os.path.normpath('{0}/{1}_{2}_table_scores_{3}_to_{4}.png'.\
                                            format(graphics_dir, obs, TW_USER_FIG[time_window], \
                                            scores_ts.minor_axis[0], scores_ts.minor_axis[-1]))
            else:
                fig_name_tmp = os.path.normpath('{0}/{1}_{2}_table_scores_{3}_to_{4}.png'.\
                                            format(graphics_dir, obs, TW_USER_FIG[time_window], \
                                            scores_ts.minor_axis[0].strftime(TW_USER_STRFTIME[time_window]), \
                                            scores_ts.minor_axis[-1].strftime(TW_USER_STRFTIME[time_window])))
        else: fig_name_tmp = fig_name
        fig.savefig(fig_name_tmp, dpi=fig_dpi, bbox_inches='tight')
        print(fig_name_tmp)
        pl.close()
    #
    # Draw table of the number of points only
    n_rows = 2+len(scores_ts.minor_axis)
    n_cols = 2
    col_period  = scores_ts.minor_axis
    table2draw  = n_rows*[[]]
    for r in range(n_rows-2):
        table2draw[r] = [str(col_period[r]), '{:.0f}'.format(scores_ts['Nb_pts'].values[0,r])]
    table2draw[-2] = ['min', '{:.0f}'.format(min(scores_ts['Nb_pts'].values[0,:]))]
    table2draw[-1] = ['max', '{:.0f}'.format(max(scores_ts['Nb_pts'].values[0,:]))]
    fig = pl.figure(figsize=(2.*n_cols, .3*(n_rows+1)))
    ax = fig.add_subplot(111)
    ax.axis('off')
    the_table = ax.table(cellText=table2draw, loc='center')
    for r in range(n_rows):
        for c in range(n_cols):
            cell = the_table.properties()['celld'][r,c]
            cell.set_height(0.1)
    if not (fig_name is None and (graphics_dir is None or obs is None or time_window is None)):
        if fig_name is None:
            if isinstance(scores_ts.minor_axis[0],str):   #time_window=='season'
                fig_name_tmp = os.path.normpath('{0}/{1}_{2}_table_nb_pts_{3}_to_{4}.png'.\
                                            format(graphics_dir, obs, TW_USER_FIG[time_window], \
                                            scores_ts.minor_axis[0], scores_ts.minor_axis[-1]))
            else:
                fig_name_tmp = os.path.normpath('{0}/{1}_{2}_table_nb_pts_{3}_to_{4}.png'.\
                                            format(graphics_dir, obs, TW_USER_FIG[time_window], \
                                            scores_ts.minor_axis[0].strftime(TW_USER_STRFTIME[time_window]), \
                                            scores_ts.minor_axis[-1].strftime(TW_USER_STRFTIME[time_window])))
        else: fig_name_tmp = fig_name
        fig.savefig(fig_name_tmp, dpi=fig_dpi, bbox_inches='tight')
        print(fig_name_tmp)
        pl.close()


# --------------------------------------------------------------------------------------------
# FUNCTIONS: INITIALIZATION
# --------------------------------------------------------------------------------------------

def check_post_zones(post_zones):
    '''
    Function used to check and format the post_zones dict.
    '''
    tmp_zones = dict()
    for zname,zone in post_zones.iteritems():
        if zname=='def':
            tmp_zones['model zone'] = None
        elif zname=='point':
            for i,xy in zip(range(len(zone)),zone):
                pnt = 'pnt_{:02d}'.format(i+1)
                tmp_zones[pnt] = [[xy[0],xy[1]],[xy[0],xy[1]]]
        elif isinstance(zone,str): # If a netcdf file is given
            if not os.path.isfile(zone):
                raise Exception('\nNetcdf file given for mask does not exist: {0}'.format(zone))
            with Dataset(zone) as nc:
                if zname=='all':
                    for name in nc.variables['Name']: tmp_zones[name] = zone
                elif zname not in nc.variables['Name']:
                    raise Exception('Mask name {0} not in {1}'.format(zname,zone))
                else: tmp_zones[zname] = zone
        else:
            tmp_zones[zname] = zone
    return tmp_zones
#
def prepareZone(zname,zone,map_set,graphics_dir):
    '''
    Prepares variables for the selected zone
    '''
    if zname=='model zone': # By default, zone = model grid
        map_set.zone = None
        graphics_dir_zone = graphics_dir
        IPNT = None
    else:
        graphics_dir_zone = graphics_dir+zname+'/'
        if isinstance(zone,str): # If a netcdf file is given
            nc = Dataset(zone)
            zone_id = nc.variables['Id'][pl.argwhere(nc.variables['Name'][:]==zname)[0][0]]
            IPNT = pl.where(nc.variables['Mask'][::-1,:]==zone_id,True,False).ravel()
            map_set.zone = [[min(map_set.X.ravel()[IPNT]),max(map_set.Y.ravel()[IPNT])],
                            [max(map_set.X.ravel()[IPNT]),min(map_set.Y.ravel()[IPNT])]]
            nc.close()
        else:                    # If zone is a rectangular area [[lon_min,lat_max],[lon_max,lat_min]]
            map_set.zone = zone
            IPNT = pl.logical_and(map_set.mod_lon.ravel()>=zone[0][0],map_set.mod_lon.ravel()<=zone[1][0])
            IPNT = pl.logical_and(IPNT,\
                   pl.logical_and(map_set.mod_lat.ravel()<=zone[0][1],map_set.mod_lat.ravel()>=zone[1][1]))
        if map_set.mod_grid_type!='uniform': IPNT = IPNT[map_set.remap]
    # Directories for the selected zone
    graphics_dir_zone = graphics_dir_zone.replace(' ','_')
    makeDirIfNeeded(graphics_dir_zone)
    return map_set,graphics_dir_zone,IPNT
#
def initPost(options_file=None):
    # Get local variables from caller namespace
    #local_vars = currentframe(1).f_globals
    #
    # Read variables from options_file
    if options_file is not None:
        local_vars = dict()
        try: execfile(options_file,local_vars)
        except: raise Exception('\nError in the options file!')
    else:
        local_vars = currentframe(1).f_globals
    # Update local variables and check options
    for key,value in local_vars.iteritems(): exec(key+"=value")
    checkOptions(locals(),options_file)
    #
    print('\nInitializing LDAS post-processing ({0} - {1}).\n'.format(analysis_period[0],analysis_period[1]))
    #
    # Make initializations
    # If LST is in the obs list, then LST is analyzed at 6am and 12pm
    if 'LST' in obs_names: 
        id_LST = obs_names.index('LST')
        del to_assim[id_LST]
        del rescale_calibs[id_LST]
        del rescale_calib_periods[id_LST]
        obs_names = filter(lambda x: x != 'LST',obs_names) + ['LST-6', 'LST-12']
        to_assim  = to_assim + [False, False]
        rescale_calibs = rescale_calibs + [None,None]
        rescale_calib_periods = rescale_calib_periods+[[],[]]
    #
    # Main parameters definition
    out_all_dir = {'Obs': out_obs_dir, 'Model': out_mod_dir, 'Analysis': out_ana_dir}
    post_from_pickle = {'Obs': post_obs_from_pickle, 'rawObs': post_obs_from_pickle, \
                        'Model': post_mod_from_pickle, 'Analysis': post_ana_from_pickle}
    obs_assim   = [obs_names[i] for i in range(len(obs_names)) if to_assim[i]==True ]
    obs_NOassim = [obs_names[i] for i in range(len(obs_names)) if to_assim[i]==False]
    obs_assim_rescaled = [obs_names[i] for i in range(len(obs_names)) if to_assim[i] and rescale_calibs[i]]
    patch_frac = getModelPatchFractions(mod_pgd_path)
    mod_grid,trip_grid = parseOptionsNam(options_path)
    mod_ana_dir = {'Model': openloop_dir, 'Analysis': analysis_dir}
    if modexp_names=='def': modexp_names = {'Model':'Model', 'Analysis':'Analysis'}
    else: modexp_names = {'Model':modexp_names[0], 'Analysis':modexp_names[1]}
    patch_post = pl.array(patch_out)-1
    #
    # Create instance of mapSet
    map_set = mapSet(mod_grid,pgd=mod_pgd_path)
    if len(trip_vars)>0:
        map_set_trip = mapSet(trip_grid)
    #
    # Check for netcdf files given as mask
    post_zones = check_post_zones(post_zones)
    #
    # Initialize loop over zones
    zname,zone = post_zones.iteritems().next()
    map_set,graphics_dir_zone,IPNT = prepareZone(zname,zone,map_set,graphics_dir)
    #
    # Update caller local variables
    local_vars = currentframe(1).f_globals
    for key,value in locals().iteritems(): local_vars[key] = value


# ============================================================================================
# MAIN PLOTTING ROUTINES
# ============================================================================================

def checkObs(obs, minim=None, maxim=None):
    importLocals(currentframe())
    print('===================================================')
    print('Checking observation: {0}'.format(obs, OBS_MOD[obs]))
    # Read all data
    data = readAllData(obs, analysis_period, assim_hour, patch_frac, out_all_dir, \
                        post_from_pickle, mod_ana_dir, ['Obs'], IPNT=IPNT)
    if 0 in data.shape:
        print('No dataset or no time step or no point. Continuing with next obs...')
        return
    for time_window,t in zip(list_time_windows,range(len(list_time_windows))): 
        if any([var_spatial_maps_out[t],var_time_series_out[t]]):
            print('Time window applied for {0}: {1}'.format(obs, time_window))
            # Apply time window
            data_tw = applyTimeWindow(data, time_window)
            col_labels = [modexp_names[n] if n in modexp_names.keys() else n for n in data_tw.items]
            #
            # Draw maps of variables: model, obs, analysis, diff(analysis-model)
            if minim is None and obs in CBAR_MINMAX: minim = CBAR_MINMAX[obs][0]
            if maxim is None and obs in CBAR_MINMAX: maxim = CBAR_MINMAX[obs][1]
            if var_spatial_maps_out[t]: draw_map_vars(map_set, obs, data_tw, graphics_dir=graphics_dir_zone, time_window=time_window, fig_dpi=fig_dpi, \
                                                      minim=minim, maxim=maxim, col_labels=col_labels)
            # Draw time series of variables (spatial average over the whole domain)
            if var_time_series_out[t]: draw_series_vars(data_tw.mean(axis=2), obs, graphics_dir=graphics_dir_zone, time_window=time_window, fig_dpi=fig_dpi, col_labels=col_labels)  #, vmin[obs_var], vmax[obs_var])
            #
    if obs in obs_assim_rescaled:
        print('Observations {0} have been rescaled, plotting raw vs. rescaled observations, monthly averaged only'.format(obs))
        # read obs_raw
        data_raw = readAllData(obs, analysis_period, assim_hour, patch_frac, \
                                out_all_dir, post_from_pickle, mod_ana_dir, ['Obs','Model','rawObs'], IPNT=IPNT)
        data_raw_tw = applyTimeWindow(data_raw, 'month')
        # draw_timeseries ('rescaled vs. raw')
        draw_series_raw(data_raw_tw.mean(axis=2), obs, graphics_dir=graphics_dir_zone, time_window='month')


def postObs(obs, data=None, minim=None, maxim=None):
    importLocals(currentframe())
    print('===================================================')
    print('Post-processing for observation: {0}'.format(obs)+(', model variable: {1}'.format(obs, OBS_MOD[obs]) if obs in OBS_MOD.keys() else ''))
    # Read all data
    if data is None:
        data = readAllData(obs, analysis_period, assim_hour, patch_frac, out_all_dir, \
                            post_from_pickle, mod_ana_dir, ['Model','Analysis','Obs'], \
                            IPNT=IPNT, map_set=map_set, filter_nan=not('SIF' in obs or 'E_ACTUAL' in obs or 'FLGPP' in obs))
    if 0 in data.shape:
        print('No dataset or no time step or no point. Continuing with next obs...')
        return
    col_labels = [modexp_names[n] if n in modexp_names.keys() else n for n in data.items]
    if 'Analysis' in data.items and 'Model' in data.items:
        col_labels.append(col_labels[pl.find(data.items=='Analysis')[0]]+'-'+col_labels[pl.find(data.items=='Model')[0]])
    for time_window,t in zip(list_time_windows,range(len(list_time_windows))): 
        if any([var_spatial_maps_out[t],var_time_series_out[t]]):
            print('Time window applied for {0}: {1}'.format(obs, time_window))
            # Apply time window
            data_tw = applyTimeWindow(data, time_window)
            # Additional diagnostis: analysis-model
            if 'Analysis' in data_tw.items and 'Model' in data_tw.items:
                data_tw['Analysis-Model'] = data_tw['Analysis']-data_tw['Model']
                data_tw['Analysis-Model'].values[pl.where(data_tw['Analysis-Model'].values==0.)]=pl.nan
            #
            # Draw maps of variables: model, obs, analysis, diff(analysis-model)
            if minim is None and obs in CBAR_MINMAX: minim = CBAR_MINMAX[obs][0]
            if maxim is None and obs in CBAR_MINMAX: maxim = CBAR_MINMAX[obs][1]
            if var_spatial_maps_out[t]: draw_map_vars(map_set,obs,data_tw,graphics_dir=graphics_dir_zone,time_window=time_window,fig_dpi=fig_dpi,minim=minim,maxim=maxim,col_labels=col_labels)
            # Draw time series of variables (spatial average over the whole domain)
            if var_time_series_out[t]: draw_series_vars(data_tw.mean(axis=2), obs, graphics_dir=graphics_dir_zone, time_window=time_window, fig_dpi=fig_dpi, col_labels=col_labels)
            #
        # Compute statistics ('R','bias','std','rmse') and show
        if 'Obs' in data.items and any([stats_spatial_maps_out[t],stats_time_series_out[t],stats_table_out[t]]):
            # Draw maps of statistic scores: model vs obs, analysis vs obs, diff(analysis-model)
            if 'SIF' in obs or 'E_ACTUAL' in obs or 'FLGPP' in obs:
                scores_per_point = compute_scores_per_point(applyTimeWindow(data,'month'), time_window)
            else:
                scores_per_point = compute_scores_per_point(data, time_window)
            if stats_spatial_maps_out[t]: draw_map_stats(map_set, scores_per_point, graphics_dir=graphics_dir_zone, obs_var=obs, time_window=time_window, fig_dpi=fig_dpi, col_labels=col_labels)
            # Draw timeseries of statistic scores (spatial average over the whole domain)
            if 'SIF' in obs: scores_for_all_points = compute_scores_for_all_points(applyTimeWindow(data,'month'), time_window)
            else           : scores_for_all_points = compute_scores_for_all_points(data, time_window)
            if stats_time_series_out[t]: draw_series_stats(scores_for_all_points, obs, graphics_dir=graphics_dir_zone, time_window=time_window, fig_dpi=fig_dpi, col_labels=col_labels)
            if stats_table_out[t]: drawScoresTable(scores_for_all_points, obs=obs, graphics_dir=graphics_dir_zone, time_window=time_window, fig_dpi=fig_dpi)


def postExtraVar(var, data=None, minim=None, maxim=None):
    importLocals(currentframe())
    print('===================================================')
    print('Post-processing for model variable {0}'.format(var))
    # Read all data
    if data is None:
        if var in ['RC_ECO','NEEC']:
            # NEEC = RC_ECO - GPPC (RC_ECO needs to be corrected also)
            data_RC_ECO = readAllData('RC_ECO', analysis_period, assim_hour, patch_frac, \
                            out_all_dir, post_from_pickle, mod_ana_dir, ['Model','Analysis'], IPNT=IPNT)
            data_GPPC   = readAllData('GPPC'  , analysis_period, assim_hour, patch_frac, \
                            out_all_dir, post_from_pickle, mod_ana_dir, ['Model','Analysis'], IPNT=IPNT)
            # needs to apply a correction for the NIT option (*0.72*(sum_time(GPPC)/sum_time(RC_ECO)))
            for i in range(2): data_RC_ECO.values[i,:] = 0.72*data_RC_ECO.values[i,:]*pl.nansum(data_GPPC.values[i,:],axis=0)/pl.nansum(data_RC_ECO.values[i,:],axis=0)
            #
            data = data_RC_ECO.copy()
            if var=='NEEC': data.values[:] = data.values[:] - data_GPPC.values[:]
            del data_RC_ECO, data_GPPC
        else:
            data = readAllData(var, analysis_period, assim_hour, patch_frac, \
                           out_all_dir, post_from_pickle, mod_ana_dir, ['Model','Analysis'], IPNT=IPNT)
    #
    if 0 in data.shape:
        print('No dataset or no time step or no point. Continuing with next obs...')
        return
    for time_window,t in zip(list_time_windows,range(len(list_time_windows))):
        print('Time window applied for {0}: {1}'.format(var, time_window))
        # Apply temporal window 
        data_tw = applyTimeWindow(data, time_window)
        col_labels = [modexp_names[n] if n in modexp_names.keys() else n for n in data_tw.items]
        # Additional diagnostis: analysis-model
        if 'Analysis' in data_tw.items and 'Model' in data_tw.items:
            data_tw['Analysis-Model'] = data_tw['Analysis']-data_tw['Model']
            data_tw['Analysis-Model'].values[pl.where(data_tw['Analysis-Model'].values==0.)]=pl.nan
            col_labels.append(col_labels[1]+'-'+col_labels[0])
        #
        # Draw maps of variables: model, analysis, diff(analysis-model)
        if minim is None and var in CBAR_MINMAX: minim = CBAR_MINMAX[var][0]
        if maxim is None and var in CBAR_MINMAX: maxim = CBAR_MINMAX[var][1]
        if var_spatial_maps_out[t]: draw_map_vars(map_set, var, data_tw, graphics_dir=graphics_dir_zone, time_window=time_window, fig_dpi=fig_dpi, minim=minim, maxim=maxim, col_labels=col_labels)
        # Draw time series of variables (spatial average over the whole domain)
        if var_time_series_out[t]: draw_series_vars(data_tw.mean(2), var, graphics_dir=graphics_dir_zone, time_window=time_window, fig_dpi=fig_dpi, col_labels=col_labels)


def postAnalysis(assim_diag=None):
    importLocals(currentframe())
    print('===================================================')
    print('Reading and plotting assimilation diagnostics')
    if assim_diag==None: assim_diag = ['Increments', 'Jacobian', 'Innovation', 'Residual']
    for diag in assim_diag:
        if diag.lower()+'_assim' not in extra_diagnostics.keys(): continue
        print('  Processing '+diag)
        if diag == 'Jacobian': patch_frac_tmp = pl.ones(patch_frac.shape)
        else                 : patch_frac_tmp = patch_frac
        data = readAllAnalysis(extra_diagnostics[diag.lower()+'_assim'], analysis_period, assim_hour, \
                patch_frac_tmp, out_all_dir, post_from_pickle, mod_ana_dir, IPNT=IPNT, filter_nan=False)
        if 0 in data.shape:
            print('No dataset or no time step or no point. Continuing with next obs...')
            return
        for time_window,t in zip(list_time_windows,range(len(list_time_windows))):
            print('Time window applied for {0}: {1}'.format(diag,time_window))
            # Apply time window
            data_tw = applyTimeWindow(data, time_window)
            col_labels = [modexp_names[n] if n in modexp_names.keys() else n for n in data_tw.items]
            diag_names = extra_diagnostics[diag.lower()+'_names']
            minim,maxim = 0,0
            if diag.lower()+'_limits' in extra_diagnostics.keys():
                minim = [extra_diagnostics[diag.lower()+'_limits'][k][0] for k in range(len(diag_names))]
                maxim = [extra_diagnostics[diag.lower()+'_limits'][k][1] for k in range(len(diag_names))]
            # Draw maps
            if var_spatial_maps_out[t]: draw_map_analysis(map_set, data_tw, diag_names, graphics_dir=graphics_dir_zone, time_window=time_window, \
                                                          diag=diag, fig_dpi=fig_dpi, minim=minim, maxim=maxim)
            # Draw time series (spatial average over the whole domain)
            if var_time_series_out[t]:
                draw_series_analysis(data_tw, diag_names, graphics_dir=graphics_dir_zone, time_window=time_window, minim=minim, maxim=maxim, fig_dpi=fig_dpi)
            # Draw daily min/max
            if var_time_series_out[t] and time_window=='day':
                # Currently spatial average
                # Future development: min,max,histo  on all dates and all points
                draw_series_analysis_min_max(data_tw, diag_names, graphics_dir=graphics_dir_zone, diag=diag, time_window=time_window, fig_dpi=fig_dpi, minim=minim, maxim=maxim)
                draw_series_analysis_histo  (data_tw, diag_names, graphics_dir=graphics_dir_zone, diag=diag, time_window=time_window, fig_dpi=fig_dpi, num_bins=None)


def postTripVar(var,local_vars):
    from ldasPre import getLookups,applyLookup
    #importLocals(currentframe())
    for key,value in local_vars.iteritems(): exec(key+"=value")
    print('===================================================')
    print('Post-processing for model variable {0}'.format(var))
    if IPNT is not None:
        mod_lon,mod_lat   = res2lonlat(mod_grid['res'],mod_grid['zone'])
        trip_lon,trip_lat = res2lonlat(trip_grid['res'],trip_grid['zone'])
        trip_lon,trip_lat = pl.meshgrid(trip_lon,trip_lat)
        trip_lon,trip_lat = trip_lon.ravel(),trip_lat.ravel()
        grid_transf_info  = {'method':'upscale_average','parameters':{'half_pix':mod_grid['res']/2.,'min_obs':1}}
        trip_lookup,_,_   = getLookups(trip_lon,trip_lat,mod_lon,mod_lat,grid_transf_info)
        trip_mask,_       = applyLookup(trip_lookup,IPNT,1)
        IPNT              = pl.array(trip_mask,dtype=bool)
    map_set_sfx = map_set; map_set = map_set_trip
    map_set.zone = map_set_sfx.zone
    postExtraVar(var)
    map_set = map_set_sfx


def postPatch(var):
    importLocals(currentframe())
    print('===================================================')
    print('Diagnostics run by patch for model variable {0}'.format(var))
    # Read all data per patch
    for patch in patch_post:
        print('Patch {0} {1} for var {2} over {3}'.format(patch+1,PATCH_NAMES[patch],var,zname))
        to_read_data = ['Model','Analysis']
        data_patch = readAllData(var, analysis_period, assim_hour, patch_frac, out_all_dir, \
                                  post_from_pickle, mod_ana_dir, to_read_data, patch=patch, IPNT=IPNT)
        patch_obs_name = var+'_{0:02d}_{1}'.format(patch+1,PATCH_NAMES[patch])
        postObs(patch_obs_name,data_patch,minim=0,maxim=0)



# ============================================================================================
# MAIN PROGRAM
# ============================================================================================


if __name__ == '__main__':
    # Check if the options file is passed correctly, read it, and check for missing options
    if len(argv) >= 2: options_file = getOptionsFile(argv)
    else: raise Exception('No option file given. Exiting')
    #
    # Initialization
    initPost(options_file)
    #
    # Loop over zones
    #
    for zname,zone in post_zones.iteritems():
        if zname in ['Greenland','Oceans','Antarctic']: continue
        #
        print('\n*********************************************')
        print('  Postprocessing over {0}'.format(zname))
        print('*********************************************\n')
        #
        # Prepare zone selection
        map_set,graphics_dir_zone,IPNT = prepareZone(zname,zone,map_set,graphics_dir)
        #
        #checkObs('LAI')
        for obs in obs_names: postObs(obs)
        #
        postAnalysis()
        #
        for var in extra_out_vars: postExtraVar(var)
        #
        for var in trip_vars: postTripVar(var,locals())
        #
        if 'patch' in extra_diagnostics:
          for var in extra_diagnostics['patch']: postPatch(var)
        #
    print('Post-processing is complete. Thank you for watching.\n')

