# ----------------------------------------------
# IMPORT LIBRARIES
# ----------------------------------------------
# Python modules
from imp import load_dynamic
if not 'argv' in locals(): from sys import argv
# LDAS modules
from ldasModule import *
from obsFunctions import OBS_VAR, flagScreen, updateObsDates
# Compiled Fortran modules
from socket import gethostname
host = gethostname()
F90_DIR = os.path.dirname(os.path.realpath(__file__))+'/F2Py/'
applyLookup = load_dynamic('apply_lookup',F90_DIR+'apply_lookup_'+host+'.so').apply_lookup
calibrateCdfMatching = load_dynamic('calculate_cdf_parameters_SWI',F90_DIR+'calculate_cdf_parameters_SWI_'+host+'.so').calculate_cdf_parameters
seasonalCdfMatching = load_dynamic('seasonal_cdf_matching_SWI',F90_DIR+'/seasonal_cdf_matching_SWI_'+host+'.so').seasonal_cdf_matching

# --------------------------------------------------------------
# PUBLIC FUNCTIONS (CALLED BY MAIN OR IMPORTED BY OTHER MODULES)
# --------------------------------------------------------------

def getObsGrid(grid_transform_info,zone):
    '''
    Build the observation variable grid and resize it according to the model grid extension.
    Arguments:
        grid_transform_info (dict): parsed grid transformation information for the observation variable.
        zone (numpy.ndarray): zone where observations are to be extracted ([[lon_min,lat_max],[lon_max,lat_min]])
    Returns:
        obs_lon/obs_lat (numpy.ndarray): longitude/latitude values of observation grid points.
        obs_slices (dict): row/column indices for slicing raw observation data for each region
    '''
    # compute buffer around model grid
    _, cut_buffer = _getGridTransformInfo(grid_transform_info)
    cut_buffer *= 2
    # buffer zone for model grid
    zbuf = pl.array(zone)+cut_buffer*pl.array([[-1,1],[1,-1]])
    # lon/lat of final observation grid
    regions = grid_transform_info['regions']
    r = regions[regions.keys()[0]]
    obs_res = grid_transform_info['grid']['res']
    #obs_res = (r['lr_lon']-r['ul_lon'])/(r['nlon']-r['pix_center'])
    obs_lon_min = r['ul_lon']+obs_res/2*(1-r['pix_center'])+obs_res*pl.ceil((zbuf[0][0]-r['ul_lon'])/obs_res-(1-r['pix_center'])/2)
    obs_lat_min = r['lr_lat']+obs_res/2*(1-r['pix_center'])+obs_res*pl.ceil((zbuf[1][1]-r['lr_lat'])/obs_res-(1-r['pix_center'])/2)
    obs_lon,obs_lat = res2lonlat(obs_res,[[obs_lon_min,zbuf[0][1]],[zbuf[1][0],obs_lat_min]],pix_center=1)
    # slice observation grid for each region
    eps = obs_res*1e-4
    obs_slices = dict()
    for r_name,r in regions.iteritems():
        # does this region intersect the model zone?
        zobs = pl.array([[r['ul_lon'],r['ul_lat']],[r['lr_lon'],r['lr_lat']]])
        to_read = (zobs[0][1]>=zbuf[1][1] and zobs[0][0]<=zbuf[1][0] and zobs[1][1]<=zbuf[0][1] and zobs[1][0]>=zbuf[0][0])
        # observation slices for this region
        lon,lat = res2lonlat(obs_res,zobs,pix_center=r['pix_center'])
        I = (lat>=zbuf[1][1]-eps)*(lat<=zbuf[0][1]+eps)*(lat>=obs_lat[0]-eps)*(lat<=obs_lat[-1]+eps)
        row_slice_in  = slice(I.argmax(),len(I)-I[::-1].argmax())
        I = (lon>=zbuf[0][0]-eps)*(lon<=zbuf[1][0]+eps)*(lon>=obs_lon[0]-eps)*(lon<=obs_lon[-1]+eps)
        col_slice_in  = slice(I.argmax(),len(I)-I[::-1].argmax())
        I = (obs_lat>=zobs[1][1]-eps)*(obs_lat<=zobs[0][1]+eps)*(obs_lat>=lat[0]-eps)*(obs_lat<=lat[-1]+eps)
        row_slice_out  = slice(I.argmax(),len(I)-I[::-1].argmax())
        I = (obs_lon>=zobs[0][0]-eps)*(obs_lon<=zobs[1][0]+eps)*(obs_lon>=lon[0]-eps)*(obs_lon<=lon[-1]+eps)
        col_slice_out  = slice(I.argmax(),len(I)-I[::-1].argmax())
        obs_slices[r_name] = {'to_read': to_read, 'lat_invert': r['lat_invert'],\
            'row_slice_in': row_slice_in, 'col_slice_in': col_slice_in, 'row_slice_out': row_slice_out, 'col_slice_out': col_slice_out}
    return obs_lon,obs_lat,obs_slices

def getAssimDates(period,assim_hour,obs_root_dir,obs_dataset,obs_var):
    '''
    Build the range of assimilation dates for an observation variable and return corresponding
    files name
    Arguments:
        period (list): start and end dates for assimilation (str: yyyy-mm-dd).
        assim_hour (int in range(24)): hour of assimilation.
        obs_root_dir (str): root directory of observations
        obs_dataset (str): observation dataset
        obs_var (str): observation variable name.
    Returns:
        assim_dates (numpy.ndarray): array of assimilation dates for the observation variable (datetime.datetime).
        files_name (str): complete path of file to read
    '''
    var_date_file = os.path.join(obs_root_dir,obs_dataset,obs_var,obs_dataset+'_'+obs_var+'_DATES.csv')
    if not os.path.isfile(var_date_file):
        ans = raw_input('{0} does not exist. Would you like to create it (can take a while)? (y/n) '.format(var_date_file))
        if ans[0].lower()=='y':
            updateObsDates(obs_root_dir,obs_dataset,obs_var)
        else:
            print('ldasPre canceled by user.')
            exit()
    period_str = [period[0].replace('-','')+'00',period[1].replace('-','')+'24']
    assim_hour_str = '{:02d}'.format(assim_hour)
    tlines = pd.read_csv(var_date_file,names=('date','file'),index_col=0)
    if obs_var == 'LST':
        dat = tlines.index.astype(str)
        i_I = [i for i in range(len(dat)) if dat[i]>=period_str[0] and dat[i]<=period_str[1] and dat.values[i][-2:]==assim_hour_str]
        I = pl.zeros((len(tlines),),dtype=bool)
        I[i_I] = True
    else:
        dat = tlines.index.map(lambda s: s.replace('XX',assim_hour_str))
        I = pl.logical_and(dat>=period_str[0],dat<=period_str[1])
    assim_dates = pd.to_datetime(dat[I],format='%Y%m%d%H')
    files_name = tlines.ix[I].values.reshape(-1)
    files_name = [os.path.join(obs_root_dir,obs_dataset,obs_var,'XXXREGIONXXX',f) for f in files_name]
    return assim_dates,files_name

def readObsDaily(obs_file,obs_dataset,obs_var,obs_slices,obs_shape,out_nonflag=False,dtype='f4',**kwargs):
    '''
    Read observation data and quality flags for the input date
    Arguments:
        obs_file (str) : root file name for observation
        obs_dataset (str): observation dataset
        obs_var (str): observation variable name.
        obs_slices (dict): row/column indices for slicing raw observation data mesh
        obs_shape (tuple): shape of observation matrix
        out_nonflag (bool): True if original observation (non flagged) should be also output
    Returns:
        obs_values (numpy.ndarray): observation values for one daily time step.
    '''
    # Variable names to read
    var_names = OBS_VAR[obs_dataset][obs_var]
    # Initialization
    obs_values = []
    for i in range(len(var_names)):
        obs_values.append(FILL*pl.ones(obs_shape,dtype=dtype))
    # Loop over regions
    for region,S in obs_slices.iteritems():
        if not S['to_read']: continue
        file_name = obs_file.replace('XXXREGIONXXX',region)
        if file_name is None: continue
        values = readFromFile(file_name,var_names,S['row_slice_in'],S['col_slice_in'],S['lat_invert'],**kwargs)
        for i in range(len(var_names)):
            obs_values[i][S['row_slice_out'],S['col_slice_out']] = values[i]
    # Apply quality flag (if required)
    obs_flagged = flagScreen(obs_dataset,obs_var,obs_values[0],obs_values[1:],file_name)
    # Return read values
    if out_nonflag: return [obs_flagged.ravel()]+[o.ravel() for o in obs_values[1:]]
    else       : return obs_flagged.ravel()

def readObs(obs_dataset, obs_var, period, \
            obs_root_dir=None, mod_res=None, mod_zone=[[-180,90],[180,-90]], hour=0, \
            dataframe=True, dtype='f4', **kwargs):
    '''
    Read observation data for external use
    Arguments:
        obs_dataset (str): observation dataset
        obs_var (str): observation variable name.
        period (str or list): date or period
    Returns:
        obs_values (pandas.DataFrame): observation values
    '''
    # Check inputs
    if not isinstance(period,list): period = [period,period]
    if obs_root_dir is None: obs_root_dir = '/cnrm/vegeo/LDAS/OBSERVATIONS'
    regions = getObsRegions(obs_root_dir,obs_dataset,obs_var)
    if mod_res  is None: mod_res  = regions.ix['reslon'][0]
    if mod_zone is None: mod_zone = [[regions.ix['ul_lon'].values.min(),regions.ix['ul_lat'].values.max()],\
                                    [regions.ix['ul_lon'].values.max(),regions.ix['ul_lat'].values.min()]]
    mod_zone = pl.array(mod_zone)
    if mod_zone.shape == (2,): mod_zone = pl.vstack((mod_zone+[-mod_res/2,mod_res/2],mod_zone+[mod_res/2,-mod_res/2]))
    # Define model grid
    mod_grid = {'type':'uniform','res':mod_res,'zone':mod_zone}
    mod_lon, mod_lat = getModelGrid(mod_grid)
    # Define grid transformation
    grid_transform_info = parseGridTransformInfo([obs_var],['def'],['def'],['def'], \
        obs_root_dir,[obs_dataset],mod_grid)[obs_var]
    # Define observation grid and lookup table
    obs_lon,obs_lat,obs_slices = getObsGrid(grid_transform_info,mod_zone)
    obs_shape = (len(obs_lat),len(obs_lon))
    mod_lookup,_,_ = getLookups(mod_lon,mod_lat,obs_lon,obs_lat,grid_transform_info)
    # Get observation file
    obs_dates, obs_files = getAssimDates(period,hour,obs_root_dir,obs_dataset,obs_var)
    obs_values = pd.DataFrame(pl.empty((len(obs_dates),len(mod_lon)),dtype=dtype),obs_dates)
    # Loop over dates
    for curr_date, curr_file in tqdm(zip(obs_dates,obs_files)):
        # Read observation
        obs_daily = readObsDaily(curr_file,obs_dataset,obs_var,obs_slices,obs_shape,curr_date=curr_date,dtype=dtype,\
                                 **kwargs)
        # Project to model grid
        obs_values.ix[curr_date],_ = projectObs2ModGrid(mod_lookup,obs_daily,grid_transform_info)
    # Return resampled observation
    if dataframe == False:
        # Reshape outputs
        nlon,nlat = len(pl.unique(mod_lon)),len(pl.unique(mod_lat))
        mod_lon,mod_lat = mod_lon[:nlon],mod_lat[::nlon]
        obs_values = obs_values.values.reshape((-1,nlat,nlon))
        if nlon==1 and nlat==1: obs_values = obs_values.flatten()
        return obs_dates,mod_lon,mod_lat,obs_values
    else:
        return obs_values

def getLookups(mod_lon,mod_lat,obs_lon,obs_lat,grid_transform_info):
    '''
    Build lookup tables mapping model grid points to observation grid points and viceversa.
    Arguments:
        mod_lon/mod_lat (numpy.ndarray): longitude/latitude values of model grid points (as from meshgrid)
        obs_lon/obs_lat (numpy.ndarray): longitude/latitude values of observation grid points.
        grid_transform_info (dict): parsed grid transformation information for the observation variable.
    Returns:
        LUT (numpy.ndarray): row j contains the indices of observation points associated to model point j.
        empty_val (int): value assigned to empty slots in lookup tables.
        Npnt (int): maximum number of points in lookup table (LUT.shape[1]
    '''
    build_lookup_fun, assim_dist = _getGridTransformInfo(grid_transform_info)
    return build_lookup_fun(mod_lon,mod_lat,obs_lon,obs_lat,assim_dist)

def projectObs2ModGrid(mod_lookup,obs_values,grid_transf_info):
    '''
    Project screened observation values and error flag for one daily time step onto model grid.
    Arguments:
        mod_lookup (numpy.ndarray): row j contains the indices of observation points associated to model point j.
        obs_values (numpy.ndarray): observation values for one daily time step.
        grid_transf_info (dict): parsed grid transformation information for the observation variable.
    Returns:
        obs_values (numpy.ndarray): observation values for one daily time step projected on model grid.
        std_obs (numpy.ndarray): standard deviation of observation values assigned to each model point.
    '''
    min_num_obs = grid_transf_info['parameters']['min_obs']
    return applyLookup(mod_lookup,obs_values,min_num_obs)

def screenFromModel(mod_dates,obs_values,mod_patch_frac,mod_out_dir,screen_mod_info):
    '''
    Screen the values of an observation variable for one day using model output, if needed.
    Arguments:
        mod_dates (datetime.datetime): date for which model values are read.
        obs_values (numpy.ndarray): observation values for one daily time step.
        mod_patch_frac (numpy.ndarray): patch fractions of each model grid point; shape = n_points,n_patches.
        mod_out_dir (str): directory of model outputs.
        screen_mod_info (dict): information about the use of model output to screen an observation variable.
    Returns:
        obs_values (numpy.ndarray): screened observation values for one daily time step.
    '''
    if screen_mod_info:
        to_discard = pl.zeros(obs_values.shape,bool)
        for mod_var, thre in zip(screen_mod_info['mod_out'],screen_mod_info['thres']):
            mod_values = readModOut(mod_var,mod_patch_frac,mod_out_dir,mod_dates=mod_dates)
            to_discard = pl.logical_or(to_discard, mod_values > thre)
        obs_values[to_discard] = FILL
    return obs_values

def screenFromSurface(obs_values,surf_screen_info):
    '''
    Screen out observation grid points that correspond to model grid points with unsatisfactory surface properties.
    Arguments:
        obs_values (pandas.core.frame.DataFrame): observation variable values for all points and times; shape = n_times, n_points.
        surf_screen_info (dict): parsed surface screening information for an observation variable.
    Returns:
        obs_values (pandas.core.frame.DataFrame): observation variable values for all points and times; shape = n_times, n_points.
    '''
    if surf_screen_info:
        to_discard = pl.zeros(obs_values.shape[1],bool)
        for feature_info in surf_screen_info.values():
            feature_values = readFromFile(feature_info['path'],[feature_info['ncname']])[0].ravel()
            to_discard = pl.logical_or(to_discard, feature_values > feature_info['threshold'])
        obs_values.ix[:,to_discard] = FILL
    return obs_values

def getRescaleParameters(mod_dates,obs_values,mod_patch_frac,mod_out_dir,rescale_info):
    '''
    Get (estimate or read from file) the rescaling parameters for an observation variable using model output.
    Arguments:
        mod_dates (datetime.datetime): date for which model values are read
        obs_values (pandas.core.frame.DataFrame): observation variable values for all points and times; shape = n_times, n_points.
        mod_patch_frac (numpy.ndarray): patch fractions of each model grid point; shape = n_points,n_patches.
        mod_out_dir (str): directory of model outputs.
        rescale_info (dict): information about the use of model output to rescale an observation variable.
    Returns:
        rescale_parameters (numpy.ndarray): rescaling parameters; shape = n_parameters, n_seasons, n_points.
    '''
    if not rescale_info['calibrate']:
        rescale_parameters = pl.load(rescale_info['parameter_path'])
    else:
        if rescale_info['method'] == 'cdf_matching':
            # Read model output for rescaling
            mod_values = readModOut(rescale_info['mod_out'],mod_patch_frac,mod_out_dir,mod_dates=mod_dates)
            mod_values[pl.isnan(mod_values)] = FILL
            # Compute rescaling parameters
            rescale_parameters = pl.empty((4,12,mod_values.shape[1])) # DIMENSIONS: 1: a, b, Npnt, Residual Sum Squares. 2: months. 3: model points.
            for month in range(1,13):
                three_months = pl.mod(month+pl.arange(-1,2)-1,12)+1
                #in_window = mod_values.index.map(lambda d: pl.logical_and((d.month==three_months).any(),\
                #  pl.logical_and(d.year>=rescale_info['period'][0],d.year<=rescale_info['period'][1])))
                in_window = mod_values.index.map(lambda d: pl.logical_and((d.month==three_months).any(),\
                  pl.logical_and(d>=pd.to_datetime(rescale_info['period'][0]),d<=pd.to_datetime(rescale_info['period'][1]))))
                rescale_parameters[0,month-1,:], rescale_parameters[1,month-1,:], \
                  rescale_parameters[2,month-1,:], rescale_parameters[3,month-1,:] = \
                  calibrateCdfMatching(obs_values.ix[in_window].values.T,mod_values.ix[in_window].values.T)
        else:
            raise Exception('Rescaling method must be \'cdf matching\', if defined')
        pl.save(rescale_info['parameter_path'],rescale_parameters)
    return rescale_parameters

def applyRescaleParameters(obs_values,rescale_parameters,rescale_info):
    '''
    Apply rescale parameters on observation variable.
    Arguments:
        obs_values (pandas.core.frame.DataFrame): observation variable values for all points and times; shape = n_times, n_points.
        rescale_parameters (numpy.ndarray): rescale parameters; shape = n_parameters, n_seasons, n_points.
        rescale_info (dict): information about the use of model output to rescale an observation variable.
    Returns:
        obs_values (numpy.ndarray): rescaled observation variable values for all points and times; shape = n_points, n_times.
    '''
    if rescale_info['method'] == 'cdf_matching':
        for m in range(1,13):
            in_month = obs_values.index.month == m
            if in_month.any():
                aux = obs_values.groupby(in_month).get_group(1).values.T
                seasonalCdfMatching(aux,rescale_parameters[0,m-1,:],rescale_parameters[1,m-1,:])
                obs_values.ix[in_month] = aux.T
        obs_values.ix[:,(obs_values == FILL).sum() >= rescale_info['max_miss']*obs_values.shape[0]] = FILL
    return obs_values

def rescaleFromModel(mod_dates,obs_values,mod_patch_frac,mod_out_dir,rescale_info):
    '''
    Rescale an observation variable using model output.
    Arguments:
        mod_dates (datetime.datetime): date for which model values are read
        obs_values (pandas.core.frame.DataFrame): observation variable values for all points and times; shape = n_times, n_points.
        mod_patch_frac (numpy.ndarray): patch fractions of each model grid point; shape = n_points,n_patches.
        mod_out_dir (str): directory of model outputs.
        rescale_info (dict): information about the use of model output to rescale an observation variable.
    Returns:
        obs_values (numpy.ndarray): rescaled observation variable values for all points and times; shape = n_points, n_times.
    '''
    if rescale_info:
        # Estimation of rescaling parameters & save them to file
        rescale_parameters = getRescaleParameters(mod_dates,obs_values,mod_patch_frac,mod_out_dir,rescale_info)
        # Rescaling
        obs_values = applyRescaleParameters(obs_values,rescale_parameters,rescale_info)
    return obs_values

def readCanariFiles(assim_obs_names,period,assim_hour,out_files_dir,file_pattern=None,usecols=None):
    '''
    Write observation variables to be assimilated to daily CANARI files.
    Arguments:
        assim_obs_names (list): names of observation variables to be assimilated (str).
        period (list): start and end dates for pre-processing (str: yyyy-mm-dd).
        assim_hour (int in range(24)): hour of assimilation.
        out_files_dir (str): folder where files are written.
        file_pattern (str): if different than default 'CANARI_NATURE_yyyymmddHhh.DAT'
        usecols (list): list of columns to read from CANARI file
    Returns:
        assim_data (dict): pre-processed observation variables (numpy.ndarray).
    '''
    assim_data = None
    if assim_obs_names:
        if usecols is not None and len(usecols)!=len(assim_obs_names):
            raise Exception('readCanariFiles: number of columns to read must match assim_obs_names')
        out_obs_dir = os.path.normpath(out_files_dir+'/CANARI')
        if file_pattern is None:
            file_pattern = ''.join((out_obs_dir,'/CANARI_NATURE_{0}H',str(assim_hour).zfill(2),'.DAT'))
        else:
            file_pattern = ''.join(out_obs_dir,file_pattern)
        print('\nReading CANARI files:')
        all_dates = pd.date_range(period[0],period[1])+pd.DateOffset(hours=assim_hour)
        tmp = pl.loadtxt(file_pattern.format(all_dates[0].strftime('%y%m%d')),usecols=[0])
        assim_data = pd.Panel(items=assim_obs_names,major_axis=all_dates,minor_axis=pl.arange(tmp.shape[0]))
        for dat in tqdm(all_dates):
            if usecols is None:
                assim_data.ix[:,dat,:] = pl.loadtxt(file_pattern.format(dat.strftime('%y%m%d'))).T
            else:
                assim_data.ix[:,dat,:] = pl.loadtxt(file_pattern.format(dat.strftime('%y%m%d')),usecols=usecols).T
        return assim_data

def writeCanariFiles(assim_data,assim_obs_names,period,assim_hour,out_files_dir,file_pattern=None):
    '''
    Write observation variables to be assimilated to daily CANARI files.
    Arguments:
        assim_data (dict): pre-processed observation variables (numpy.ndarray).
        assim_obs_names (list): names of observation variables to be assimilated (str).
        period (list): start and end dates for pre-processing (str: yyyy-mm-dd).
        assim_hour (int in range(24)): hour of assimilation.
        out_files_dir (str): folder where files are written.
        file_pattern (str): if different than default 'CANARI_NATURE_yyyymmddHhh.DAT'
    '''
    if assim_obs_names:
        out_obs_dir = os.path.normpath(out_files_dir+'/CANARI')
        makeDirIfNeeded(out_obs_dir)
        if file_pattern is None:
            file_pattern = ''.join((out_obs_dir,'/CANARI_NATURE_{0}H',str(assim_hour).zfill(2),'.DAT'))
        else:
            file_pattern = ''.join((out_obs_dir,file_pattern))
        print('\nWriting CANARI files:')
        all_dates = pd.date_range(period[0],period[1])+pd.DateOffset(hours=assim_hour)
        assim_data = pd.Panel(assim_data,items=assim_obs_names,major_axis=all_dates)
        assim_data.values[pl.isnan(assim_data.values)] = FILL
        for dat in tqdm(assim_data.major_axis):
            pl.savetxt(file_pattern.format(dat.strftime('%y%m%d')),assim_data.ix[:,dat].values,'%.10g')


# ----------------------------------------------
# PRIVATE FUNCTIONS (CALLED BY PUBLIC FUNCTIONS)
# ----------------------------------------------

def lookupUpscale(mod_lon,mod_lat,obs_lon,obs_lat,assim_dist):
    '''
    Lookup table for upscale resampling for uniform observation grids and list of model points
    '''
    missing_value = -999
    Kobs = pl.arange(len(obs_lon)*len(obs_lat),dtype='uint32').reshape(len(obs_lat),len(obs_lon))
    obs_res = float(obs_lon[1]-obs_lon[0])
    b = assim_dist/obs_res
    c_lon = (mod_lon-obs_lon[0])/obs_res
    c_lat = (mod_lat-obs_lat[0])/obs_res
    n_lon,n_lat = len(obs_lon),len(obs_lat)
    Npnt = int((2*b+1)**2)
    LUT = missing_value*pl.ones((len(mod_lon),Npnt),dtype='int32')
    for k in range(len(mod_lon)):
      I = slice(max(min(int(pl.ceil(c_lon[k]-b)),n_lon),0),\
                max(min(int(pl.floor(c_lon[k]+b))+1,n_lon),0))
      J = slice(max(min(int(pl.ceil(c_lat[k]-b)),n_lat),0),\
                max(min(int(pl.floor(c_lat[k]+b))+1,n_lat),0))
      tmp = Kobs[J,I].ravel()
      LUT[k,:len(tmp)] = tmp
    Npnt = pl.argmax(pl.all(LUT==missing_value,axis=0))
    if Npnt>0: LUT = LUT[:,:Npnt]
    return LUT,Npnt,missing_value

def lookupDownscale(mod_lon,mod_lat,obs_lon,obs_lat,assim_dist):
    '''
    Lookup table for downscale resampling for uniform observation grids and list of model points
    '''
    missing_value = -999
    n_lon,n_lat = len(obs_lon),len(obs_lat)
    obs_res = float(obs_lon[1]-obs_lon[0])
    Ilon = pl.int_(pl.round_((mod_lon-obs_lon[0])/obs_res))
    Ilat = pl.int_(pl.round_((mod_lat-obs_lat[0])/obs_res))
    r2 = assim_dist**2
    N = int(pl.ceil(assim_dist/obs_res))
    Npnt = (2*N+1)**2
    LUT = missing_value*pl.ones((len(mod_lon),Npnt),dtype='int32')
    for k in trange(len(mod_lon)):
      for j in range(-N,N+1):
        for i in range(-N,N+1):
          if Ilon[k]+i>=0 and Ilon[k]+i<n_lon and Ilat[k]+j>=0 and Ilat[k]+j<n_lat and \
            (mod_lon[k]-obs_lon[Ilon[k]+i])**2+(mod_lat[k]-obs_lat[Ilat[k]+j])**2<=r2:
            LUT[k,i+N+(j+N)*(2*N+1)] = Ilon[k]+i+(Ilat[k]+j)*n_lon
    LUT[:,pl.any(LUT!=missing_value,axis=0)]
    Npnt = LUT.shape[1]
    return LUT,Npnt,missing_value

def lookupNearest(mod_lon,mod_lat,obs_lon,obs_lat,assim_dist):
    '''
    Lookup table for resampling using the nearest neighbor of the observation grid 
    (observation grid must be uniform !)
    '''
    missing_value = -999
    n_lon,n_lat = len(obs_lon),len(obs_lat)
    Npnt = 1
    LUT = missing_value*pl.ones((len(mod_lon),Npnt),dtype='int32')
    for k in trange(len(mod_lon)):
        i_lon  = pl.find((obs_lon-mod_lon[k])**2==min((obs_lon-mod_lon[k])**2))[0]
        i_lat  = pl.find((obs_lat-mod_lat[k])**2==min((obs_lat-mod_lat[k])**2))[0]
        if pl.logical_and(pl.absolute(obs_lon[i_lon]-mod_lon[k])<=assim_dist,\
                          pl.absolute(obs_lat[i_lat]-mod_lat[k])<=assim_dist):
            LUT[k] = i_lon + i_lat*n_lon
        #if pl.sqrt((obs_lon[i_lon]-mod_lon[k])**2+(obs_lat[i_lat]-mod_lat[k])**2)<=assim_dist: LUT[k] = i_lon + i_lat*n_lon
    return LUT,Npnt,missing_value

def _getGridTransformInfo(grid_transf_info):
    '''
    Get the function and the distance parameter to build the lookup tables.
    '''
    if grid_transf_info['method'] == 'downscale_radius':
        return lookupDownscale, grid_transf_info['parameters']['radius']
    elif grid_transf_info['method'] == 'upscale_average':
        return lookupUpscale, grid_transf_info['parameters']['half_pix']
    elif grid_transf_info['method'] == 'nearest':
        return lookupNearest, grid_transf_info['parameters']['d_max']
    else:
        raise Exception('grid transformation method must be either \'upscale_average\' or \'downscale_radius\' or \'nearest\'')


# ----------------------------------------------
# MAIN PROGRAM
# ----------------------------------------------

def initPre(options_file=None):
    # Get local variables from caller namespace
    local_vars = currentframe(1).f_locals
    # Read variables from options_file
    if options_file is not None:
        try: execfile(options_file,local_vars)
        except: raise Exception('\nError in the options file!')
    # Update local variables and check options
    for key,value in local_vars.iteritems(): exec(key+"=value")
    checkOptions(locals(),options_file)
    #
    print('\nInitializing LDAS pre-processing ({0} - {1}).\n'.format(analysis_period[0],analysis_period[1]))
    #
    # Make initializations
    # Parse some input
    mod_grid,_ = parseOptionsNam(options_path)
    grid_transf_info_all = parseGridTransformInfo(obs_names,obs_grids,grid_transform_methods,grid_transform_parameters,\
        obs_root_dir,obs_datasets,mod_grid)
    surf_screen_info_all = parseSurfaceScreenInfo(obs_names,surf_screen_features,surf_screen_thresholds,mod_pgd_path)
    mod_screen_info_all = parseModelScreenInfo(obs_names,screen_mod_names,screen_mod_thres)
    rescale_info_all = parseRescaleInfo(obs_names,rescale_methods,rescale_calibs,rescale_calib_periods,analysis_period,\
        rescale_mod_names,rescale_mod_steps,rescale_max_miss,exp_name,out_obs_dir)
    #
    # Obs period: add 1 day to the analysis period (obs are assimilated at the end of the daily simulation)
    obs_period = [(pd.to_datetime(d)+pd.DateOffset(days=1)).strftime('%Y-%m-%d') for d in analysis_period]
    # Get model patch fractions and longitudes/latitudes
    mod_patch_frac = getModelPatchFractions(mod_pgd_path)
    mod_lon, mod_lat = getModelGrid(mod_grid,mod_pgd_path) # Point coordinates
    # Allocate data collections and loop through observation variables...
    assim_obs_names = list()
    assim_data = dict()
    assim_variance = dict()
    #
    # Update caller local variables
    for key,value in locals().iteritems(): local_vars[key] = value


if __name__ == '__main__':
    # Check if the options file is passed correctly, read it, and check for missing options
    if len(argv) >= 2: options_file = getOptionsFile(argv)
    else: raise Exception('No option file given. Exiting')
    #
    initPre(options_file)
    #
    # Loop over observations
    for v, obs_var in enumerate(obs_names):
        # If LST is in the observation list, then a special treatment is needed (stored every 3 hours)
        out_hour = range(0,21+1,3) if obs_var=='LST' else [assim_hour]
        patches = pl.arange(12) if obs_var=='LAI_MC' else [None]
        for hour in out_hour:
          for p in patches:
            if   obs_var=='LST'   : obs_var_tmp = '{0}_{1}'.format(obs_var,hour)
            elif obs_var=='LAI_MC': obs_var_tmp = obs_var+'.t{:02d}'.format(p+1)
            else                  : obs_var_tmp = obs_var
            print('\nPre-processing {0}'.format(obs_var_tmp))
            # Observation and model dates
            obs_dates, obs_files = getAssimDates(obs_period,hour,obs_root_dir,obs_datasets[v],obs_var)
            mod_dates = obs_dates.map(lambda d: d+pd.DateOffset(days=-1))
            # Allocate data structures (pandas.core.frame.DataFrame)
            obs_values = pd.DataFrame(pl.empty((obs_dates.size,mod_patch_frac.shape[1]),dtype='f4'),obs_dates)
            #
            # Optionaly read from pickle file
            read_dates = pl.zeros(obs_dates.size,dtype=bool)
            if obs_from_pickle[v]:
                exist_pathPickles, list_pathPickles = findPDataFiles(obs_var_tmp,analysis_period,out_obs_dir)
                if pl.all(exist_pathPickles):
                    print 'Read variable {0} from pickle'.format(obs_var_tmp)
                    read_dates = readPDataFiles([list_pathPickles[l] for l in pl.find(exist_pathPickles)],obs_values)
            #
            if not all(read_dates):
                # Read resized observation grid
                obs_lon,obs_lat,obs_slices = getObsGrid(grid_transf_info_all[obs_var],mod_grid['zone'])
                obs_shape = (len(obs_lat),len(obs_lon))
                # Model lookup table: mod_lookup[i] = obs points associated to model point i
                mod_lookup,_,_ = getLookups(mod_lon,mod_lat,obs_lon,obs_lat,grid_transf_info_all[obs_var])
                # Loop through dates...
                for curr_date, curr_file in tqdm(zip(obs_dates,obs_files)):
                    # Read observation and flag values (resized for model domain)
                    obs_daily = readObsDaily(curr_file,obs_datasets[v],obs_var,obs_slices,obs_shape,curr_date=curr_date,usr_slice={'patch':p})
                    # Project observations onto model grid
                    obs_values.ix[curr_date], _ = projectObs2ModGrid(mod_lookup,obs_daily,grid_transf_info_all[obs_var])
                # Screen projected observations with model output (if needed)
                obs_values = screenFromModel(mod_dates,obs_values,mod_patch_frac,openloop_dir,mod_screen_info_all[obs_var])
                # Screen using surface features (if needed)
                obs_values = screenFromSurface(obs_values,surf_screen_info_all[obs_var])
                # Handle NaN values
                obs_values[pl.isnan(obs_values)] = FILL
                # Store obs values before rescaling
                if rescale_info_all[obs_var]:
                    writePDataFiles('raw_'+obs_var_tmp,mod_dates,obs_values,out_obs_dir)
                    # Observation rescaling (if needed)
                    obs_values = rescaleFromModel(mod_dates,obs_values,mod_patch_frac,openloop_dir,rescale_info_all[obs_var])
                # Save data to pickle file for post-processing
                writePDataFiles(obs_var_tmp,mod_dates,obs_values,out_obs_dir)
                print('Pre-processing of {0} complete'.format(obs_var_tmp))
            #
            # Store data in dict only if it has to be assimilated
            if to_assim[v]:
                assim_data[obs_var_tmp] = obs_values
                assim_obs_names.append(obs_var_tmp)
    #
    # Write CANARI files
    writeCanariFiles(assim_data,assim_obs_names,obs_period,assim_hour,out_obs_dir)
    print('\nCANARI files written.')
    #

