# ----------------------------------------------
# IMPORT LIBRARIES
# ----------------------------------------------
import pandas as pd
import pylab as pl
import numpy as np
import os
from netCDF4 import Dataset
from sys import argv
from glob import glob
from ftplib import FTP, error_perm
from getpass import getpass
from shutil import copyfile
from subprocess import call
import calendar
# LDAS modules
from ldasModule import *

# ----------------------------------------------
# SET CONSTANTS
# ----------------------------------------------
# All SFX variables needed
SFX_VARS_UNITS = {'Tair':'K', 'Qair':'kg/kg', 'Wind':'m/s', 'DIR_SWdown':'W/m2', 'LWdown':'W/m2', 'Rainf':'kg/m2/s',\
                   'Snowf':'kg/m2/s' , 'PSurf':'Pa' ,'CO2air':'kg/m3', 'SCA_SWdown':'W/m2' ,'Wind_DIR':'deg'}
# Location of ancillary data
data_dir = os.path.realpath(os.path.split(__file__)[0])+'/data/'

# ----------------------------------
# PUBLIC FUNCTIONS
# ----------------------------------

def retrieveRawForcing(period,raw_forcing_dir,raw_forcing_format,forcing_dir,freq_forcing,nc_version,mod_pgd_path,mod_grid,FORC_VARS):
    '''
    Retrieve raw forcing files from raw_forcing_dir, select the given area and add the missing fields.
    Monthly forcing.
    Arguments:
        period (list): start and end dates for pre-processing (str: yyyy-mm-dd).
        raw_forcing_dir (str): directory from where the raw forcing data are read
        raw_forcing_format (str): either VAR_NC (one forcing file per variable in seperate folders) or SFX_NC (already in SFX format in raw_forcing_dir)
        forcing_dir (str): directory where forcing files are written
        nc_version (str): NetCDF version for output ('NETCDF4', 'NETCDF4_CLASSIC', 'NETCDF3_64BIT', 'NETCDF3_CLASSIC'), read from option file
        mod_pgd_path (str): path to the directory that contains PGD.nc
        mod_grid (dict): types and characteristics of model grid
        FORC_VARS (dict): dictionaries that contains the list of variables present in the forcing files and their correspondance with SFX forcing variables
    '''
    #
    # 0-Read PGD nc file to get information
    mod_ZS = getModelZS(mod_pgd_path)
    #
    # 1-Find the temporal range of the raw forcing data to read
    months = pd.date_range(pd.to_datetime(period[0]),pd.to_datetime(period[1]),freq='M').to_period() # +pd.DateOffset(months=1)
    #
    # 2-LAT/LON slice => nb_points
    if raw_forcing_format=='SFX_NC':
        nc_filename = os.path.normpath('{0}/{1}'.format(raw_forcing_dir,os.listdir(raw_forcing_dir)[0]))
    elif raw_forcing_format=='VAR_NC':
        #nc_filename = os.path.normpath('{0}/Tair/{1}'.format(raw_forcing_dir,os.listdir(raw_forcing_dir+'/Tair/')[0]))
        nc_filename = glob(raw_forcing_dir+'/Tair/*.nc')[0]
    # Get longitude and latitude vectors (must be named as lon* and lat* in forcing files)
    with Dataset(nc_filename) as nc:
        nam_lon = [v for v in nc.variables.keys() if 'lon' in v[:3].lower()][0]
        nam_lat = [v for v in nc.variables.keys() if 'lat' in v[:3].lower()][0]
        src_lon, src_lat = nc.variables[nam_lon][:], nc.variables[nam_lat][:]
    src_lon = np.mod(src_lon+180.,360.)-180.
    if src_lon.ndim==1: src_lon,src_lat = np.meshgrid(src_lon,src_lat)
    lon,lat = src_lon[0,:],src_lat[:,0]
    (min_lat,max_lat,min_lon,max_lon) = (mod_grid['zone'][1][1],mod_grid['zone'][0][1],mod_grid['zone'][0][0],mod_grid['zone'][1][0])
    #sl_lat, sl_lon = np.unravel_index(pl.find((src_lat >= min_lat) & (src_lat <= max_lat) & (src_lon >= min_lon) & (src_lon <= max_lon)),src_lat.shape)
    #nb_points = len(sl_lat)
    #
    # Check if resolution is ok
    res_optionFile = mod_grid['res']
    res_forcing    = pl.absolute(lat[1]-lat[0])
    if not res_optionFile==res_forcing: print('Warning: resolution of model in option file ({0}) is not the same as the forcing ({1}). Resolution of the forcing is kept here.'\
                    .format(res_optionFile,res_forcing))
    #
    #pdb.set_trace()
    #sl_lon = pl.arange(np.argwhere(lon>=min_lon)[0][0],np.argwhere(lon>max_lon)[0][0])
    sl_lon = slice(np.argwhere(lon>=min_lon)[0][0],np.argwhere(lon>max_lon)[0][0])
    if lat[1]>lat[0]:
        #sl_lat = pl.arange(np.argwhere(lat>=min_lat)[0][0],np.argwhere(lat>max_lat)[0][0])
        sl_lat = slice(np.argwhere(lat>=min_lat)[0][0],np.argwhere(lat>max_lat)[0][0])
    else:
        #sl_lat = pl.arange(np.argwhere(lat<min_lat)[0][0]-1,np.argwhere(lat<=max_lat)[0][0]-1,-1)
        sl_lat = slice(np.argwhere(lat<min_lat)[0][0]-1,np.argwhere(lat<=max_lat)[0][0]-1,-1)
    nlon,nlat = sl_lon.stop-sl_lon.start,(sl_lat.stop-sl_lat.start)*(1 if sl_lat.step is None else sl_lat.step)
    nb_points = nlon*nlat
    #
    # 3-Read CCO2 annual estimations and constants
    co2_year_ppmv_file = data_dir+'GHG_1765_2500_A2.txt'
    co2_year_ppmv = np.genfromtxt(co2_year_ppmv_file)[:,0:1+1] # years 1765-2500
    M_CO2     = 44.010e-3    # molar mass [kg/mol] of CO2
    M_air     = 28.965e-3    # molar mass [kg/mol] of air
    M_wat     = 18.015e-3    # molar mass [kg/mol] of water
    gravity   = 9.807        # gravity [m/s2]
    NA        = 6.022e23     # Avogrado number [-] (number of entities in a mole)
    kB        = 1.381e-23    # Boltzman constant [J/K] (physical constant relating energy at the individual particle level with temperature)
    r_dry_air = NA*kB/M_air  # specific gas constant [J/kg/K] for dry air
    r_wat_vap = NA*kB/M_wat  # specific gas constant [J/kg/K] for water vapor contained in the air (humidity)
    height_for_Tair = 2.0    # Tair is at 2 meters
    #
    # 4-Extract from monthly raw forcing and add missing fields
    print('*******************************')
    print('Extracting ROI from raw forcing')
    print('AND adding missing fields from')
    print('{0} to {1}'.format(analysis_period[0],analysis_period[1]))
    print('*******************************')
    print('')
    for curr_month in months:
        print(curr_month.strftime('%Y%m'))
        #
        # A - Time info
        start = curr_month.start_time
        end   = curr_month.end_time + pd.DateOffset(minutes=1)
        time  = pd.date_range(start, end, freq='{0}H'.format(freq_forcing))
        #
        # B - Init data structure
        data = pd.Panel(items = SFX_VARS_UNITS.keys() , major_axis=time, minor_axis=pl.arange(nb_points), dtype='f4')
        #
        # C - Extract forcing data
        if raw_forcing_format=='SFX_NC':
            files_date = glob('{0}/*{1}*.nc'.format(raw_forcing_dir,curr_month.strftime('%Y%m')))
            if len(files_date)==0:
                raise Exception('No monthly forcing file for date {0} in {1}'.format(curr_month.strftime('%Y%m'),raw_forcing_dir))
            file2read = os.path.normpath(files_date[0])
            for sfx_var,forc_var in FORC_VARS.iteritems():
                data[sfx_var] = Dataset(file2read).variables[forc_var][:,sl_lat,sl_lon].reshape((time.size, nb_points))
        elif raw_forcing_format=='VAR_NC':
            for sfx_var,forc_var in FORC_VARS.iteritems():
                files_date = glob('{0}/{1}/*{2}*.nc'.format(raw_forcing_dir,forc_var,curr_month.strftime('%Y%m')))
                #if sfx_var=='Rainf':
                #    files_date = glob('{0}/rainf_{1}/*{2}*.nc'.format(raw_forcing_dir,forc_var,curr_month.strftime('%Y%m')))
                #    forc_var = 'Rainf'
                if len(files_date)==0:
                    raise Exception('No monthly forcing file for date {0}, variable {1} in {2}'.format(curr_month.strftime('%Y%m'),forc_var,raw_forcing_dir))
                file2read = os.path.normpath(files_date[0])
                print file2read
                # Dimensions should be checked first instead of being hard-coded ...
                if forc_var in ['Tair','Qair','Wind']:
                    data[sfx_var] = Dataset(file2read).variables[forc_var][:,0,sl_lat,sl_lon].reshape((time.size, nb_points))
                #    # in E2O forcing file (probably all ECMWF forcing), Tair and Qair have one more dimension : (time, height, lat, lon)
                else:
                    #import pdb ; pdb.set_trace()
                    data[sfx_var] = Dataset(file2read).variables[forc_var][:,sl_lat,sl_lon].reshape((time.size, nb_points))
        #
        # D - Find the missing variables and assign them when possible
        MISS_VARS = set(SFX_VARS_UNITS.keys())-set(FORC_VARS.keys())
        if 'SCA_SWdown' in MISS_VARS:
            print('SCA_SWdown variable is missing in forcing file, assigning it to 0 W/m2')
            data['SCA_SWdown'] = np.zeros((time.size, nb_points))
        if 'Wind_DIR' in MISS_VARS:
            print('Wind_DIR variable is missing in forcing file, assigning it to 180 degrees')
            data['Wind_DIR'] =  np.ones((time.size, nb_points))*180.
        #
        # E - CALCULATION OF CO2 CONCENTRATION [ppmv to kg/m3] if needed
        if 'CO2air' in MISS_VARS:
            print('CO2air variable is missing in forcing file, calculating it using Psurf, Tair, Qair, ppmv/year')
            '''
            ## Psurf = Psoil                               +  rho_air_moist * g * height
            ##       = rho_air_moist * r_air_moist * Tair  +  rho_air_moist * g * height
            ##       = rho_air_moist * (r_air_moist * Tair + g * height)
            ## => rho_air_moist = Psurf / (r_air_moist * Tair + g * height)
            ##    with: r_air_moist = r_dry_air * (1+ ((r_wat_vap/r_dry_air) -1)*Qs ) with Qs:specific humidity
            ## C[CO2 in kg/m3] = ppmv*1e-6 * (M_CO2/M_air) * rho_air_moist 
            ##   (1st multiplication with molar masses to transform volume ratio into mass ratio)
            ##   (2nd multiplication with density to transforme air (mix with water) mass into volume 
            ##    to get C[CO2] in mass of CO2 per volume of air)
            '''
            r_air_moist     = r_dry_air * (1+ ((r_wat_vap/r_dry_air) -1)*data['Qair'].values )
            rho_air_moist   = data['PSurf'].values / (r_air_moist*data['Tair'].values + gravity * height_for_Tair)
            CCO2            = co2_year_ppmv[curr_month.year-1765,1]*1e-6 * (M_CO2/M_air) * rho_air_moist
            data['CO2air']  = CCO2
        OTHER_MISS_VARS = set(MISS_VARS)-set(['SCA_SWdown','Wind_DIR','CO2air'])
        if len(OTHER_MISS_VARS)>0:
            raise Exception('Forcing file cannot be produced because of missing variables : {0}'.format(OTHER_MISS_VARS))
        #
        # F - Init raw netcdf file to write
        if not os.path.isdir('{0}/raw_monthly/'.format(forcing_dir)): os.makedirs('{0}/raw_monthly/'.format(forcing_dir))
        raw_nc_file = '{0}/raw_monthly/FORCING_{1}.nc'.format(forcing_dir,curr_month.strftime('%Y%m'))
        nc_file = _initializeNetCDF_lonlat(raw_nc_file, nc_version, time, src_lon[sl_lat,sl_lon], src_lat[sl_lat,sl_lon], mod_ZS.reshape((nlat,nlon)))
        #
        # G - Write raw netcdf file with forcing data
        for sfx_key_var in FORC_VARS.keys() :
            nc_file.variables[sfx_key_var][:] = data[sfx_key_var].values.reshape((-1,nlat,nlon)).astype(float)
        for miss_var in MISS_VARS :
            nc_file.variables[miss_var][:] = data[miss_var].values.reshape((-1,nlat,nlon)).astype(float)
        nc_file.close()
        print('Forcing written in {0}'.format(raw_nc_file))

def retrieveRawForcing_SAFRAN_BIN(analysis_period,raw_forcing_dir,raw_forcing_format,forcing_dir,freq_forcing,nc_version,mod_pgd_path,mod_grid,FORC_VARS):
    '''
    Retrieve raw forcing files from raw_forcing_dir, select the given area and add the missing fields.
    Annual forcing. Annual for hydrology, i.e. August 1st to July 31st.
    Arguments:
        period (list): start and end dates for pre-processing (str: yyyy-mm-dd).
        raw_forcing_dir (str): directory from where the raw forcing data are read
        raw_forcing_format (str): either VAR_NC (one forcing file per variable in seperate folders) or SFX_NC (already in SFX format in raw_forcing_dir)
        forcing_dir (str): directory where forcing files are written
        nc_version (str): NetCDF version for output ('NETCDF4', 'NETCDF4_CLASSIC', 'NETCDF3_64BIT', 'NETCDF3_CLASSIC'), read from option file
        mod_pgd_path (str): path to the directory that contains PGD.nc
        mod_grid (dict): types and characteristics of model grid
        FORC_VARS (dict): dictionaries that contains the list of variables present in the forcing files and their correspondance with SFX forcing variables
    '''
    #
    raw_frc_dir = forcing_dir+'/SAFRAN_raw/'
    if not os.path.isdir(raw_frc_dir): os.makedirs(raw_frc_dir)
    # 0-Read coordinates file get information
    index_france = pl.loadtxt(data_dir+'coord_France_SAFRAN.txt',usecols=(1,),skiprows=1,dtype=int)-1 #index for Fortran (starts at 1)
    #
    # 1-Find the temporal range of the raw forcing data to read
    if pd.to_datetime(analysis_period[0])<=pd.datetime(pd.to_datetime(analysis_period[0]).year,8,1): 
        year_beg = pd.to_datetime(analysis_period[0]).year -1
    else: year_beg = pd.to_datetime(analysis_period[0]).year
    if pd.to_datetime(analysis_period[1])<=pd.datetime(pd.to_datetime(analysis_period[1]).year,8,1):
        year_end = pd.to_datetime(analysis_period[1]).year -1
    else: year_end = pd.to_datetime(analysis_period[1]).year
    # 2-Number of points
    nb_points = len(index_france)
    # 3-Read CCO2 annual estimations and constants
    co2_year_ppmv_file = data_dir+'GHG_1765_2500_A2.txt'
    co2_year_ppmv = np.genfromtxt(co2_year_ppmv_file)[:,0:1+1] # years 1765-2500
    M_CO2     = 44.010e-3    # molar mass [kg/mol] of CO2
    M_air     = 28.965e-3    # molar mass [kg/mol] of air
    M_wat     = 18.015e-3    # molar mass [kg/mol] of water
    gravity   = 9.807        # gravity [m/s2]
    NA        = 6.022e23     # Avogrado number [-] (number of entities in a mole)
    kB        = 1.381e-23    # Boltzman constant [J/K] (physical constant relating energy at the individual particle level with temperature)
    r_dry_air = NA*kB/M_air  # specific gas constant [J/kg/K] for dry air
    r_wat_vap = NA*kB/M_wat  # specific gas constant [J/kg/K] for water vapor contained in the air (humidity)
    height_for_Tair = 2.0    # Tair is at 2 meters
    #
    # 4-Extract from yearly raw forcing and add missing fields
    print('\n*****************************')
    print('Extracting ROI from raw forcing')
    print('AND adding missing fields from')
    print('{0} to {1}'.format(analysis_period[0],analysis_period[1]))
    print('*******************************')
    print('')
    # Connexion to HENDRIX
    ftp = FTP('hendrix')
    print('\nType your user name and password to connect to hendrix:')
    ftp.login(raw_input('User name: '),getpass('Password: '))
    print('\nConnected to hendrix.')
    dev_null = open('/dev/null','w')
    #
    for year_s in range(year_beg,year_end+1):
        print('\nForcing file from {0} to {1}'.format(year_s,year_s+1))
        #
        # A - Copy forcing data from FTP-Hendrix to LOCAL raw forcing directory
        # FULL YEAR HYDRO : Aug.1 to Jul.31
        try:
            ftp.cwd('{0}/CLIM-SIM/FORCAGE/{1}{2}'.format(raw_forcing_dir,str(year_s)[2:],str(year_s+1)[2:]))
            print('\nRaw annual SAFRAN files (Aug/{0}-Jul/{1}):'.format(year_s,year_s+1))
            for safran_var in FORC_VARS.itervalues():
                print('  * {0}'.format(safran_var))
                raw_frc_file = 'Forc{0}.DAT_france_{1}{2}'.format(safran_var,str(year_s)[2:],str(year_s+1)[2:])
                ftp.retrbinary('RETR '+raw_frc_file,open(raw_frc_dir+'/'+raw_frc_file,'w').write)
            print('  --> raw copy in '+raw_frc_dir)
            end_time_raw = pd.datetime(year_s+1,7,31)
            end_frc_file = str(year_s)[2:]+str(year_s+1)[2:]
        # PARTIAL YEAR: Aug.1 to OTHER DATE TBD
        except error_perm: # if reading the complete annual file fails, then look for the most recent partial file
            ftp.cwd('{0}/MENS-SIM/FORCAGE/{1}{2}'.format(raw_forcing_dir,str(year_s)[2:],str(year_s+1)[2:]))
            # Depending on the end date, the file to download is not the same
            prod_date_str = _getSafranProdDateStr(analysis_period[1])
            print('\nRaw partial mensual SAFRAN files until {0}'.format(prod_date_str))
            for safran_var in FORC_VARS.itervalues():
                print('  * {0}'.format(safran_var))
                raw_frc_file = 'Forc{0}.DAT_france_{1}{2}_{3}'.format(safran_var,str(year_s)[2:],str(year_s+1)[2:],prod_date_str)
                ftp.retrbinary('RETR '+raw_frc_file,open(raw_frc_dir+'/'+raw_frc_file,'w').write)
            print('  --> raw copy in '+raw_frc_dir)
            end_time_raw = pd.to_datetime(prod_date_str)
            end_frc_file = str(year_s)[2:]+str(year_s+1)[2:]+'_'+_getSafranProdDateStr(analysis_period[1])
        except: raise
        #
        # B - Read the downloaded files and keep only the reduced domain for France
        time = pd.date_range(pd.datetime(year_s,8,1),\
                             end_time_raw + pd.DateOffset(days=1) - pd.DateOffset(minutes=1),\
                             freq='{0}H'.format(freq_forcing))
        print('\nReduced SAFRAN files over France only')
        for safran_var in FORC_VARS.itervalues():
            print('  * {0}'.format(safran_var))
            raw_frc_file = 'Forc{0}.DAT_france_{1}'.format(safran_var,end_frc_file)
            red_frc_file = 'Forc{0}.DAT_red_france_{1}'.format(safran_var,end_frc_file)
            data_tmp = np.fromfile(raw_frc_dir+'/'+raw_frc_file,'>f4').reshape((len(time),9894))[:,1:-1] # 1st and last elements are not values
            data_keep = data_tmp[:,index_france].ravel()
            data_keep.tofile(raw_frc_dir+'/'+red_frc_file)
        print('  --> reduced to France and written in '+raw_frc_dir)
        #
        # C - Find the missing variables and assign them when possible
        MISS_VARS = set(SFX_VARS_UNITS.keys())-set(FORC_VARS.keys())
        if 'SCA_SWdown' in MISS_VARS:
            print('\nSCA_SWdown variable is missing in forcing file, assigning it to 0 W/m2')
            data_SCA_SWdown = np.zeros((len(time), nb_points)).ravel().astype(np.dtype('>f4'))
            red_frc_file    = 'Forc{0}.DAT_red_france_{1}'.format('SCA_SWdown',end_frc_file)
            data_SCA_SWdown.tofile(raw_frc_dir+'/'+red_frc_file)
            print('  --> written in '+raw_frc_dir)
        if 'Wind_DIR' in MISS_VARS:
            print('\nWind_DIR variable is missing in forcing file, assigning it to 0 degrees')
            data_Wind_DIR   = np.zeros((len(time), nb_points)).ravel().astype(np.dtype('>f4'))
            red_frc_file    = 'Forc{0}.DAT_red_france_{1}'.format('Wind_DIR',end_frc_file)
            data_Wind_DIR.tofile(raw_frc_dir+'/'+red_frc_file)
            print('  --> written in '+raw_frc_dir)
        # D - Read PSURF from FILE
        if 'PSurf' in MISS_VARS:
            print('\nPSurf variable is missing in forcing file, reading it form data/ForcPRES_France_SAFRAN.DAT')
            data_tmp        = np.fromfile(data_dir+'ForcPRES_France_SAFRAN.DAT','>f4')[1:-1] # 1st and last elements are not values
            data_PSurf      = np.tile(data_tmp,(len(time),1))
            red_frc_file    = 'Forc{0}.DAT_red_france_{1}'.format('PSurf',end_frc_file)
            data_PSurf.tofile(raw_frc_dir+'/'+red_frc_file)
            print('  --> written in '+raw_frc_dir)
        # E - Calculation of CO2 concentration [ppmv to kg/m3]
        if 'CO2air' in MISS_VARS:
            print('\nCO2air variable is missing in forcing file, calculating it using Psurf, Tair, Qair, ppmv/year')
            '''
            ## Psurf = Psoil                               +  rho_air_moist * g * height
            ##       = rho_air_moist * r_air_moist * Tair  +  rho_air_moist * g * height
            ##       = rho_air_moist * (r_air_moist * Tair + g * height)
            ## => rho_air_moist = Psurf / (r_air_moist * Tair + g * height)
            ##    with: r_air_moist = r_dry_air * (1+ ((r_wat_vap/r_dry_air) -1)*Qs ) with Qs:specific humidity
            ## C[CO2 in kg/m3] = ppmv*1e-6 * (M_CO2/M_air) * rho_air_moist 
            ##   (1st multiplication with molar masses to transform volume ratio into mass ratio)
            ##   (2nd multiplication with density to transforme air (mix with water) mass into volume 
            ##    to get C[CO2] in mass of CO2 per volume of air)
            '''
            Qair_tmp = np.fromfile(raw_frc_dir+'/Forc{0}.DAT_red_france_{1}'.format('Q'    ,end_frc_file),'>f4').reshape((len(time),nb_points))
            Tair_tmp = np.fromfile(raw_frc_dir+'/Forc{0}.DAT_red_france_{1}'.format('T'    ,end_frc_file),'>f4').reshape((len(time),nb_points))
            Psur_tmp = np.fromfile(raw_frc_dir+'/Forc{0}.DAT_red_france_{1}'.format('PSurf',end_frc_file),'>f4').reshape((len(time),nb_points))
            r_air_moist   = r_dry_air * (1+ ((r_wat_vap/r_dry_air) -1)*Qair_tmp )
            rho_air_moist = Psur_tmp / (r_air_moist*Tair_tmp + gravity * height_for_Tair)
            data_CO2air   = co2_year_ppmv[time.year-1765,1][:,None]*1e-6 * (M_CO2/M_air) * rho_air_moist
            data_CO2air   = data_CO2air.ravel().astype(np.dtype('>f4'))
            red_frc_file  = 'Forc{0}.DAT_red_france_{1}'.format('CO2air',end_frc_file)
            data_CO2air.tofile(raw_frc_dir+'/'+red_frc_file)
            print('  --> written in '+raw_frc_dir)
        #
        OTHER_MISS_VARS = set(MISS_VARS)-set(['SCA_SWdown','Wind_DIR','CO2air','PSurf'])
        if len(OTHER_MISS_VARS)>0: raise Exception('Forcing file cannot be produced because of missing variables : {0}'.format(OTHER_MISS_VARS))
    #
    dev_null.close()
    ftp.close()
    print('\nDisconnected from hendrix.\n')


def writeDailyNcForcing(period,assim_hour,nc_version,forcing_dir,freq_forcing):
    '''
    Write daily netCDF forcing files for running model in data assimilation mode (from assim_hour to assim_hour the next day).
    Arguments:
        period (list): start and end dates for pre-processing (str: yyyy-mm-dd).
        assim_hour (int in range(24)): hour of assimilation.
        nc_version (str): NetCDF version of forcing files ['NETCDF4', 'NETCDF4_CLASSIC', 'NETCDF3_64BIT', 'NETCDF3_CLASSIC'].
        forcing_dir (str): root of forcing files directory tree.
    '''
    #
    # 1 - Find the temporal range of the raw forcing data to read
    months = pd.date_range(pd.to_datetime(period[0]),pd.to_datetime(period[1]),freq='M').to_period() # +pd.DateOffset(months=1)
    nb_steps = 24/freq_forcing
    #
    # 2 - Read constants from first file
    curr_month = months[0]
    raw_nc_file = '{0}/raw_monthly/FORCING_{1}.nc'.format(forcing_dir,curr_month.strftime('%Y%m'))
    if not os.path.isfile(raw_nc_file):
        raise Exception('No raw monthly forcing for {0} in {1}'.format(curr_month.strftime('%Y%m'),forcing_dir))
    mod_lon  = Dataset(raw_nc_file).variables['LON'][:]
    mod_lat  = Dataset(raw_nc_file).variables['LAT'][:]
    mod_zs   = Dataset(raw_nc_file).variables['ZS'][:]
    nlat,nlon = mod_lon.shape
    nb_points = nlat*nlon
    #
    print('*******************************')
    print('Creating daily forcing files')
    print('*******************************')
    #
    # 3 - Loop on monthly raw forcing files to write daily netcdf
    print('Loop on monthly raw forcing')
    for curr_month in months:
        raw_nc_file = '{0}/raw_monthly/FORCING_{1}.nc'.format(forcing_dir,curr_month.strftime('%Y%m'))
        if not os.path.isfile(raw_nc_file):
            raise Exception('No raw monthly forcing for {0} in {1}'.format(curr_month.strftime('%Y%m'),forcing_dir))
        nb_days = pd.tslib.monthrange(curr_month.year,curr_month.month)[1] # nb of days in this month -> curr_month.days_in_month (but doesnot work for all pandas versions)
        #
        # A - Read monthly raw forcing
        monthly_data = pd.Panel(items=SFX_VARS_UNITS.keys(), major_axis=pd.date_range(start=pd.datetime(curr_month.year, curr_month.month, 1),\
                                periods=nb_days*nb_steps+1, freq='{0}H'.format(freq_forcing)), minor_axis=pl.arange(nb_points))
        for sfx_var in SFX_VARS_UNITS.keys():
            monthly_data[sfx_var] = Dataset(raw_nc_file).variables[sfx_var][:].reshape((-1,nb_points))
        #
        # B - Daily file starting at assim_hour to assim_hour the next day (if time step=3H, 9 values are written)
        days = pd.date_range(start=pd.datetime(curr_month.year,curr_month.month,1,assim_hour,0,0),periods=nb_days+1,freq='D')
        # Handle case where assim_hour is not 0h: creation of forcing from 0h to assim_hour
        if assim_hour>0:
            days = [pd.Timestamp(pd.datetime(curr_month.year,curr_month.month,1,0,0,0))]+days.tolist()
        else:
            days = days.tolist()
        #for curr_day,d in zip(days,range(len(days))):
        for d in range(len(days)-1):
            curr_day = days[d]
            print(curr_day)
            time = pd.date_range(start=curr_day,end=days[d+1],freq='{0}H'.format(freq_forcing))
            daily_data = pd.Panel(items=SFX_VARS_UNITS.keys(), major_axis=time, minor_axis=pl.arange(nb_points))
            # Until the end of the last day of the month
            # If last day, then need to read the first of following month
            next_day = curr_day+pd.DateOffset(days=1)
            if next_day.month==curr_day.month:
                for sfx_var in SFX_VARS_UNITS.keys():
                    daily_data[sfx_var] = monthly_data[sfx_var][daily_data.major_axis[0]:daily_data.major_axis[-1]]
            else:
                next_raw_file = '{0}/raw_monthly/FORCING_{1}.nc'.format(forcing_dir, next_day.strftime('%Y%m'))
                if not os.path.isfile(next_raw_file):
                    print('WARNING: raw monthly forcing file ({0}) does not exist, so in order to create the last date {1}, the forcing values from last day are repeated'.format(next_day.strftime('%Y%m'),curr_day))
                    next_raw_file = raw_nc_file
                for sfx_var in SFX_VARS_UNITS.keys():
                    until_hour = len(monthly_data[sfx_var][daily_data.major_axis[0]:daily_data.major_axis[-1]])
                    daily_data[sfx_var].values[0:until_hour,:] = monthly_data[sfx_var][daily_data.major_axis[0]:daily_data.major_axis[-1]]
                    daily_data[sfx_var].values[until_hour:,:]  = Dataset(next_raw_file).variables[sfx_var][:].reshape((-1,nb_points))[1:nb_steps-until_hour+2,:]
                        #next_monthly_data[sfx_var] = Dataset(next_raw_file).variables[sfx_var][:]        
            # Initialize the netCDF file to write
            nc2write = '{0}/FORCING_{1}.nc'.format(forcing_dir,curr_day.strftime('%Y%m%d%H'))
            nc_file2write = _initializeNetCDF_lonlat(nc2write, nc_version, time, mod_lon, mod_lat, mod_zs)
            # Fill it
            for sfx_var in SFX_VARS_UNITS.keys():
                nc_file2write.variables[sfx_var][:] = daily_data[sfx_var].values.reshape((-1,nlat,nlon))
            nc_file2write.close()


def writeDailyNcForcing_SAFRAN_BIN(period,assim_hour,nc_version,forcing_dir,freq_forcing,mod_grid,mod_pgd_path,FORC_VARS):
    '''
    Write daily netCDF forcing files for running model in data assimilation mode (from assim_hour to assim_hour the next day).
    Arguments:
        period (list): start and end dates for pre-processing (str: yyyy-mm-dd).
        assim_hour (int in range(24)): hour of assimilation.
        nc_version (str): NetCDF version of forcing files ['NETCDF4', 'NETCDF4_CLASSIC', 'NETCDF3_64BIT', 'NETCDF3_CLASSIC'].
        forcing_dir (str): root of forcing files directory tree.
        freq_forcing (int): time frequency of the forcing data (1 hour, 3 hours, 24 hours for example)
        mod_grid (struct): model grid type defined in the option file
        mod_pgd_path (str): path to PGD.nc, PREP.nc (defined in the option file)
    '''
    # 0 - Get lat,lon,ZS from PGD.nc
    mod_lon,mod_lat = getModelGrid(mod_grid,mod_pgd_path)
    mod_zs          = getModelZS(mod_pgd_path)
    nb_points       = len(mod_lon)
    #
    # 1 - Find the temporal range of the raw forcing data to read
    if pd.to_datetime(period[0])<=pd.datetime(pd.to_datetime(period[0]).year,8,1):
        year_beg = pd.to_datetime(period[0]).year -1
    else:
        year_beg = pd.to_datetime(period[0]).year
    if pd.to_datetime(period[1])<=pd.datetime(pd.to_datetime(period[1]).year,8,1):
        year_end = pd.to_datetime(period[1]).year -1
    else:
        year_end = pd.to_datetime(period[1]).year
    #
    print('*******************************')
    print('Creating daily forcing files')
    print('*******************************')
    #
    for year_s in range(year_beg,year_end+1):
        # 2 - Check if the files in SAFRAN_raw contain all year or are partial
        example_file = glob(forcing_dir+'/SAFRAN_raw/ForcT.DAT_france_'+str(year_s)[2:]+str(year_s+1)[2:]+'*')
        if example_file[0][-4:] == str(year_s)[2:]+str(year_s+1)[2:]: 
            end_time_raw = pd.datetime(year_s+1,7,31)
            end_frc_file = example_file[0][-4:]
        else: 
            end_time_raw  = pd.to_datetime(example_file[0][-8:])
            end_frc_file = example_file[0][-13:]
        # 3 - Time steps contained in the SAFRAN bin file and time steps to keep
        time_in_raw = pd.date_range(pd.datetime(year_s,8,1),\
                                    end_time_raw + pd.DateOffset(days=1) - pd.DateOffset(minutes=1),\
                                    freq='{0}H'.format(freq_forcing))
        times2keep_ix = pl.find(pl.logical_and(time_in_raw>=pd.to_datetime(period[0])+pd.DateOffset(hours=assim_hour),\
                                               time_in_raw<=pd.to_datetime(period[1])+pd.DateOffset(days=1,hours=assim_hour)))
        # 4 - Read yearly SAFRAN bin files (and missing variables)
        yearly_data = pd.Panel(items=SFX_VARS_UNITS.keys(), major_axis=time_in_raw[times2keep_ix], minor_axis=pl.arange(nb_points))
        for sfx_var,safran_var in zip(FORC_VARS.iterkeys(),FORC_VARS.itervalues()):
            red_frc_file = 'Forc{0}.DAT_red_france_{1}'.format(safran_var,end_frc_file)
            yearly_data[sfx_var][:]  = np.fromfile(forcing_dir+'/SAFRAN_raw/'+red_frc_file,'>f4').reshape((len(time_in_raw),nb_points))[times2keep_ix,:]
        for miss_var in set(SFX_VARS_UNITS.keys())-set(FORC_VARS.keys()):
            red_frc_file = 'Forc{0}.DAT_red_france_{1}'.format(miss_var,end_frc_file)
            yearly_data[miss_var][:] = np.fromfile(forcing_dir+'/SAFRAN_raw/'+red_frc_file,'>f4').reshape((len(time_in_raw),nb_points))[times2keep_ix,:]
        # 5 - Check previous period if it is needed to keep or not
        beforeTimes2keep = pl.find(pl.logical_and(time_in_raw>=pd.to_datetime(period[0]),\
                                                  time_in_raw<=pd.datetime(year_s,8,1)+pd.DateOffset(hours=assim_hour)))
        if len(beforeTimes2keep)!=0:
            nc_file2read = Dataset(nc2write)
            toto = pd.Panel(items=SFX_VARS_UNITS.keys(), major_axis=pl.arange(0,25), minor_axis=pl.arange(nb_points)) 
            for sfx_var in SFX_VARS_UNITS.iterkeys():
                toto[sfx_var].values[0:25-len(beforeTimes2keep),:] = nc_file2read.variables[sfx_var][0:25-len(beforeTimes2keep),:]
                toto[sfx_var].values[25-len(beforeTimes2keep):25,:] = yearly_data[sfx_var].values[beforeTimes2keep,:]
            nc_file2read.close()
            nc_file2write = _initializeNetCDF(nc2write, nc_version, time_in_day, nb_points, mod_lon, mod_lat, mod_zs)
            for sfx_var in SFX_VARS_UNITS.iterkeys():
                nc_file2write.variables[sfx_var][:,:] = toto[sfx_var].values[:,:]
            nc_file2write.close()
            times2keep_ix = times2keep_ix[len(beforeTimes2keep)-1:]
            del toto
        # 6 - One forcing file per day from D@9am to D+1@9am included
        dates_2_write = pl.unique((time_in_raw[times2keep_ix]-pd.datetools.timedelta(hours=assim_hour+1)).date)[1:]
        print('Daily forcing from {0} to {1}'.format(dates_2_write[0],dates_2_write[-1]))
        for date, i in tqdm(zip(dates_2_write,range(len(dates_2_write)))):
            id_day_start = int(    i*(24./freq_forcing)  )
            id_day_stop  = int((i+1)*(24./freq_forcing)+1)
            time_in_day  = pd.date_range(start=time_in_raw[times2keep_ix[id_day_start]],periods=(24./freq_forcing)+1,freq='{0}H'.format(freq_forcing))
            nc2write = '{0}/FORCING_{1}.nc'.format(forcing_dir,time_in_raw[times2keep_ix[id_day_start]].strftime('%Y%m%d%H'))
            nc_file2write = _initializeNetCDF(nc2write, nc_version, time_in_day, nb_points, mod_lon, mod_lat, mod_zs)
            for sfx_var in SFX_VARS_UNITS.iterkeys():
                nc_file2write.variables[sfx_var][0:min(len(times2keep_ix),id_day_stop)-id_day_start,:] = yearly_data[sfx_var].ix[time_in_day].values[0:min(len(times2keep_ix),id_day_stop)-id_day_start,:]
            nc_file2write.close()
        del yearly_data


# ----------------------------------------------
# PRIVATE FUNCTIONS (CALLED BY PUBLIC FUNCTIONS)
# ----------------------------------------------

def _initializeNetCDF_lonlat(nc_filename,nc_version,time,inLON,inLAT,inZS):
    '''
    Initialize forcing netcdf file for SFX with surface variables
    Arguments:
        nc_filename (str): name of the netcdf file to create
        nc_version (str): NetCDF version of forcing files ['NETCDF4', 'NETCDF4_CLASSIC', 'NETCDF3_64BIT', 'NETCDF3_CLASSIC']
        time (pandas.tseries.index.DatetimeIndex): pd.date_range of the time steps
        inLON,inLAT,inZS (): 
    Returns:
        nc_file ():
    '''
    #
    nlat,nlon = inLON.shape
    #
    nc_file = Dataset(nc_filename,'w',format=nc_version)
    #
    nc_file.createDimension('time', len(time))
    nc_file.createDimension('lon', nlon)
    nc_file.createDimension('lat', nlat)
    #
    nc_file.createVariable('time', float, ('time', ))
    nc_file.variables['time'].setncattr('units', 'hours since ' + time[0].strftime('%Y-%m-%d %H:%M:%S')) 
    nc_file.variables['time'][:] = pl.array([(time[i]-time[0]).total_seconds()/3600. for i in pl.arange(len(time))]).astype(float)
    #
    nc_file.createVariable('FRC_TIME_STP',float)
    nc_file.variables['FRC_TIME_STP'].setncattr('units', 's')
    nc_file.variables['FRC_TIME_STP'].assignValue(float((time[1]-time[0]).total_seconds()))
    #
    nc_file.createVariable('LON',float,('lat','lon'))
    nc_file.variables['LON'].setncattr('units','degrees')
    nc_file.variables['LON'][:] = inLON
    #
    nc_file.createVariable('LAT',float,('lat','lon'))
    nc_file.variables['LAT'].setncattr('units','degrees')
    nc_file.variables['LAT'][:] = inLAT
    #
    nc_file.createVariable('ZS',float,('lat','lon'))
    nc_file.variables['ZS'].setncattr('units','m')
    nc_file.variables['ZS'][:] = inZS
    #
    nc_file.createVariable('UREF',float,('lat','lon'))
    nc_file.variables['UREF'].setncattr('units','m')
    nc_file.variables['UREF'][:] = np.ones((nlat,nlon))*10.
    #
    nc_file.createVariable('ZREF',float,('lat','lon'))
    nc_file.variables['ZREF'].setncattr('units','m')
    nc_file.variables['ZREF'][:] = np.ones((nlat,nlon))*2.
    #
    for sfx_key_var in SFX_VARS_UNITS.keys():
        nc_file.createVariable(sfx_key_var,float,('time','lat','lon'),\
          zlib=True,complevel=6,shuffle=True)
        nc_file.variables[sfx_key_var].setncattr('units',SFX_VARS_UNITS[sfx_key_var])
    return nc_file

def _initializeNetCDF(nc_filename,nc_version,time,nb_points,inLON,inLAT,inZS):
    '''
    Initialize forcing netcdf file for SFX with surface variables
    Arguments:
        nc_filename (str): name of the netcdf file to create
        nc_version (str): NetCDF version of forcing files ['NETCDF4', 'NETCDF4_CLASSIC', 'NETCDF3_64BIT', 'NETCDF3_CLASSIC']
        time (pandas.tseries.index.DatetimeIndex): pd.date_range of the time steps
        nb_points (int): number of points for the whole domain
        inLON,inLAT,inZS (): 
    Returns:
        nc_file ():
    '''
    #
    nc_file = Dataset(nc_filename,'w',format=nc_version)
    #
    nc_file.createDimension('time', len(time))
    nc_file.createDimension('Number_of_points', nb_points)
    #
    nc_file.createVariable('time', float, ('time', ))
    nc_file.variables['time'].setncattr('units', 'hours since ' + time[0].strftime('%Y-%m-%d %H:%M:%S')) 
    nc_file.variables['time'][:] = pl.array([(time[i]-time[0]).total_seconds()/3600. for i in pl.arange(len(time))]).astype(float)
    #
    nc_file.createVariable('FRC_TIME_STP',float)
    nc_file.variables['FRC_TIME_STP'].setncattr('units', 's')
    nc_file.variables['FRC_TIME_STP'].assignValue(float((time[1]-time[0]).total_seconds()))
    #
    nc_file.createVariable('LON',float,('Number_of_points',))
    nc_file.variables['LON'].setncattr('units','degrees')
    nc_file.variables['LON'][:] = inLON
    #
    nc_file.createVariable('LAT',float,('Number_of_points',))
    nc_file.variables['LAT'].setncattr('units','degrees')
    nc_file.variables['LAT'][:] = inLAT
    #
    nc_file.createVariable('ZS',float,('Number_of_points',))
    nc_file.variables['ZS'].setncattr('units','m')
    nc_file.variables['ZS'][:] = inZS
    #
    nc_file.createVariable('UREF',float,('Number_of_points',))
    nc_file.variables['UREF'].setncattr('units','m')
    nc_file.variables['UREF'][:] = np.ones((nb_points,))*10.
    #
    nc_file.createVariable('ZREF',float,('Number_of_points',))
    nc_file.variables['ZREF'].setncattr('units','m')
    nc_file.variables['ZREF'][:] = np.ones((nb_points,))*2.
    #
    for sfx_key_var in SFX_VARS_UNITS.keys():
        nc_file.createVariable(sfx_key_var,float,('time','Number_of_points'),\
          zlib=True,complevel=6,shuffle=True)
        nc_file.variables[sfx_key_var].setncattr('units',SFX_VARS_UNITS[sfx_key_var])
    return nc_file

def _getSafranProdDateStr(last_date_str):
    '''
    Get production date of partial SAFRAN forcing file.
    Arguments:
        last_date_str (str: yyyy-mm-dd): last date string.
    Returns:
        SAFRAN formatted production date string.
    '''
    last_date = pd.datetools.parse(last_date_str)
    if last_date.day < 21 and last_date.month != 8 : return '{0}21'.format(last_date.strftime('%Y%m'))
    else: 
        if   last_date.month == 7  : return    '{0}31'.format(last_date.strftime('%Y%m'))
        elif last_date.month == 12 : return  '{0}0121'.format(last_date.year+1)
        else: return '{0}{1}21'.format(last_date.year,str(last_date.month+1).zfill(2))



# ----------------------------------------------
# MAIN PROGRAM
# ----------------------------------------------

if __name__ == '__main__':
    # Check if the options file is passed correctly, read it, and check for missing options
    options_file = getOptionsFile(argv)
    try:
        execfile(options_file)
    except:
        print('\nError in the options file!')
        raise
    checkOptions(locals(),options_file)
    mod_grid,_ = parseOptionsNam(options_path)
    # Optional retrieval of raw forcing files
    if get_raw_forcing:
        if raw_forcing_format=='SFX_NC' or raw_forcing_format=='VAR_NC' :
            retrieveRawForcing(analysis_period,raw_forcing_dir,raw_forcing_format,forcing_dir,freq_forcing,nc_version,mod_pgd_path,mod_grid,FORC_VARS)
        elif raw_forcing_format=='SAFRAN_BIN' :
            retrieveRawForcing_SAFRAN_BIN(analysis_period,raw_forcing_dir,raw_forcing_format,forcing_dir,freq_forcing,nc_version,mod_pgd_path,mod_grid,FORC_VARS)
    # Write daily netCDF forcing files
    if raw_forcing_format=='SFX_NC' or raw_forcing_format=='VAR_NC' :
        writeDailyNcForcing(analysis_period,assim_hour,nc_version,forcing_dir,freq_forcing)
    elif raw_forcing_format=='SAFRAN_BIN' :
        writeDailyNcForcing_SAFRAN_BIN(analysis_period,assim_hour,nc_version,forcing_dir,freq_forcing,mod_grid,mod_pgd_path,FORC_VARS)


